(in-package user)

;; If *load-user-epi-init* is t, and a file called epi-init.lisp is
;; present in the directory in which Epilog was started, then that
;; file will be loaded.  I (ANK) dislike this feature, because having
;; the program's behavior depend on the directory in which it was
;; started can be very confusing.  However, I 'm leaving it turned on
;; by default for backwards compatibility.
(defvar *load-user-epi-init* t)

;; *epilog-home* is the root of the epilog code tree, i.e. the
;; directory that contains this file.
(defvar *epilog-home* (make-pathname :directory
				     (pathname-directory *load-pathname*)))

;;;************************* PORTABILITY *****************************
;; This section is an attempt to centralize the changes that need to
;; be made when porting Epilog to a new lisp, or moving it to a new
;; directory.  Warning: there are still a few #+/#- conditionals
;; scattered here and there in the rest of the code.

;; These values used to be on the default *require-search-list*, and
;; we depended on them being there, but in ACL 6.0 they were removed.
(pushnew #p"" system:*require-search-list*)
(pushnew #p(:type "fasl") system:*require-search-list*)
(pushnew #p(:type "lisp") system:*require-search-list*)

#+CMU
(progn
  (setf (logical-pathname-translations "epilog")
    (list (list "**;*.*.*"
		(merge-pathnames "**/*.*.*" *epilog-home*))
	  (list "*.*.*"
		(merge-pathnames "*.*.*" *epilog-home*))
	  )
    )

  (intern "BOOLEAN" 'common-lisp)
  (export (find-symbol "BOOLEAN" 'common-lisp) 'common-lisp)
  (deftype common-lisp::boolean () '(member t nil))
  )

;; Note that in ACL (at least), logical pathname translations are
;; flushed when an image is dumped.

#+allegro
(progn
  (setf (logical-pathname-translations "epilog")
    (list (list (logical-pathname "**;*.*")
		(merge-pathnames "**/*.*" *epilog-home*))
	  ))
)

;; This is untested.  I imagine by now lucid has changed from the old
;; style of in-package (with quoted argument) to the new style
;; (non-quoted), making this unnecessary, but I don't have a lucid
;; implementation to test, so I don't know.
#+lucid
(progn
  (defvar *redefined-in-package-p* nil)
  (when (not *redefined-in-package-p*)
    (setq *redefined-in-package-p* t)
    (let (old-in-package (macro-function 'lucid-common-lisp::in-package))
      (setf (macro-function 'lucid-common-lisp::in-package)
	(lambda (form env)
	  (funcall old-in-package
		   `(quote ,(cadr form))
		   env))))))




;;;*********************************************************************

#+lucid (setq *load-if-source-newer* :load-source)
  
(with-open-file (ifile "epilog:start-up-msg"
                       :if-does-not-exist nil)
  (cond (ifile 
	 (terpri)
	 (do ((line (read-line ifile nil '|eof|) (read-line ifile nil '|eof|)))
	     ((eq line '|eof|) nil)
	     (terpri) (princ line))
	 (terpri)) ))





;;;*************************MEMORY MANAGEMENT***************************
;;;ADD MEMORY, so we don't gc as often (especially during loading):
(terpri) (princ "MEMORY MANAGEMENT .....  ") (terpri) (terpri)
#+lucid (cond ((memq :lucid *features*)
      (change-memory-management :expand 200 
				:reclamation-ratio 0.10
				:growth-rate 20 :growth-limit 400)))

#+lucid (defadvice (lucid::prompt set-package) ()
  (in-package 'epilog)
  (remove-advice 'lucid::prompt 'set-package)
  (advice-continue))


;;;*************************SET UP THE PROGRAM****************************
;;;load the program itself
(terpri) (princ "LOADING EPILOG ..... ") (terpri) (terpri)
(eval-when (compile load eval)
  (require 'epiuser "epilog:epiuser;epiuser")
  (require 'epilog "epilog:story;story")
)
(in-package epilog)
;load specialist definitions
(load "epilog:specialists")

; start up the manditory specialists
(use-spec 'type 'episode 'hier 'part 'meta 'other 'equality)

; load the required hierarchy, indicator, and lexical info
(load "epilog:epi-hier")
(load "epilog:epi-indicate")
(load "epilog:epi-lex")

; load the initialization files
; site initialization
(load "epilog:epi-init.lisp" :if-does-not-exist nil)
; user initialization
(if user::*load-user-epi-init*
    (load "epi-init.lisp" :if-does-not-exist nil))
(number-hier)

; if current process is not interactive, reset anything that will cause the
; user to be queried
(unless (check-online-status)
  (tweak '*default-lex* t)
  (tweak '*default-trans* t)
  (tweak '*prompt-if-too-complex* nil)
  (tweak '*stop-if-error* nil))
