;       Last change    August 1993
(in-package epilib)

; Wff comparison routines
;  These routines compare the wffs in the manner requested (incompatible for
; goal directed reasoning, compatible for possible factoring)
; and return comparison-info which contain unification information as
; well as which literals matched.

(defstruct comparison-info
  (wff1 nil)                    ; 1st wff compared
  (wff2 nil)                    ; 2nd wff compared
  (lit1 nil)                    ; literal from first wff
  (lit2 nil)                    ; literal from 2nd wff
  (subs1 nil)                   ; substitutions required in first
  (subs2 nil)                   ; substitutions required in second
  (residue1 nil)                ; residue (considered part of first)
  (residue2 nil)                ; residue (considered part of second)
  )

; Note - these routines are guaranteed to change!!! This is a first cut and
; is not necessarily the best (and may not even work in all cases)

;  Useful routines:
;     compare-wffs wff1 wff2 compatible trace header cleanup using-mp
;       compares the given formulas. If compatible is t, formulas are
;       compared for compatiblity, otherwise for incompatibility
;       A positive result is not necessarily a guarantee of usefulness
;       (especially when comparing conditionals).

; should only be comparing keys out of formulas, which are literals
; except for embedded formulas

(tweakable '*use-types-in-unify* t 
  "Indicates whether to use types to help decide if variables are unifiable" 
  'boolean)
(defvar *reverse-test* nil "Indicates whether or not inside K operator")

; comparison saves all substitutions up til finished, and then gets them
; this way we can ensure that the same variable doesn't get more than one
; substitution

(defun compare-wffs (wff1 wff2 compatible &key (prefer nil) 
			  (traced nil) (header "") (cleanup t) (entering nil)
			  (using-mp nil) (subnet '/) (assumption nil)
			  (types1 nil) (types2 nil)
			  &aux res subs spec-res temp useres a1 subs1 subs2
			  (p1 (getp wff1 'pred))
			  (p2 (getp wff2 'pred))
			  (type1 (getp wff1 'wff-type))
			  (type2 (getp wff2 'wff-type)))
  (if (or (null wff1) (null wff2)) (return-from compare-wffs nil))
  (setq res
  (cond ((and (or (epi-variable wff1) (epi-variable wff2))
	      (not (and p1 p2)))
	 (if compatible (compare-variables wff1 wff2 prefer)) )
	((and (epi-variable p1)
	      (= (list-length (getp wff1 'arguments)) 1)
	      (> (list-length (getp wff2 'arguments)) 1)
	      (setq temp (reverse-lambda wff2)))
	 (compare-wffs wff1 temp compatible :prefer prefer :traced traced
		       :using-mp using-mp :subnet subnet
		       :types1 types1 :types2 types2
		       :header header :cleanup nil :entering entering))
	((and (epi-variable p2)
	      (= (list-length (getp wff2 'arguments)) 1)
	      (> (list-length (getp wff1 'arguments)) 1)
	      (setq temp (reverse-lambda wff1)))
	 (compare-wffs temp wff2 compatible :prefer prefer :traced traced
		       :using-mp using-mp :subnet subnet
		       :types1 types1 :types2 types2
		       :header header :cleanup nil :entering entering))
	((and (equal (getp wff1 'pred) 'true)
	      (not (equal (getp wff2 'pred) 'true)))
	 ; compare first argument instead
	 (setq a1 (car (getp wff1 'arguments)))
	 (compare-wffs (cond ((getp a1 'quote-wff)) (t a1)) wff2 
		       (if (neg-p wff1) (not compatible) compatible)
		       :prefer prefer :traced traced
		       :using-mp using-mp :subnet subnet
		       :types1 types1 :types2 types2
		       :header header :cleanup nil :entering entering)
	 )
	((and (equal (getp wff2 'pred) 'true)
	      (not (equal (getp wff1 'pred) 'true)))
	 ; compare first argument instead
	 (setq a1 (car (getp wff2 'arguments)))
	 (compare-wffs wff1 (cond ((getp a1 'quote-wff)) (t a1)) 
		       (if (neg-p wff2) (not compatible) compatible)
		       :prefer prefer :traced traced
		       :using-mp using-mp :subnet subnet
		       :types1 types1 :types2 types2
		       :header header :cleanup nil :entering entering)
	 )
	((or (equal type1 type2)
	     (and (member type1 '(episodic causal)) (equal type2 'prefix)
		  (epi-variable-p (getp wff2 'pred)))
	     (and (member type2 '(episodic causal)) (equal type1 'prefix)
		  (epi-variable-p (getp wff1 'pred))
		  (setq type1 type2)) )
	 (case type1
	       (constant (if compatible 
			     (or (equal wff1 wff2)
				 (member wff1 (get-equivalence wff2 subnet)))))
	       (episodic
		(let ((n1 (neg-p wff1))
		      (n2 (neg-p wff2))
		      (p1 (getp wff1 'pred))
		      (p2 (getp wff2 'pred)) 
		      (a1 (getp wff1 'arguments))
		      (a2 (getp wff2 'arguments)))
		     (when (compare-items (last1 a1) (last1 a2) t 
					  :prefer prefer :subnet subnet)
			   (cond ((compare-predicates n1 p1 n2 p2 compatible 
						   prefer subnet)
					(compare-wffs (car a1) (car a2) t 
						      :using-mp using-mp
						      :types1 types1 
						      :types2 types2
						      :subnet subnet
						      :prefer prefer 
						      :entering entering
						      :cleanup nil))
				 ((and n1
				       (compare-predicates nil p1 n2 p2
						   compatible prefer subnet))
					(compare-wffs (car a1) (car a2)
						(not compatible) 
						:using-mp using-mp
						:types1 types1 
						:types2 types2
						:subnet subnet
						:prefer prefer 
						:entering entering
						:cleanup nil))
				 ((and (not n2) (not n1)
				       (compare-predicates nil p1 nil p2 t 
							   prefer subnet))
				  (compare-wffs (car a1) (car a2)
						compatible 
						:prefer prefer 
						:using-mp using-mp
						:types1 types1 
						:types2 types2
						:subnet subnet
						:entering entering
						:cleanup nil))
				 ))
		     ))
	       ((prefix causal) ;episodic
		; preds compatible, neg provides comp/incomp
		; or for incompat, preds incompat, negs same
		; compare preds and arguments, take into account neg
		; compare arguments
		; compare preds with negations
		; if predicate of one is a variable, reverse lc the other
		; and try to comapre that way?
		(let ((pres (compare-predicates (neg-p wff1) (getp wff1 'pred)
					   (neg-p wff2) (get wff2 'pred)
					   compatible prefer subnet))
		      ares minres)
		     (when pres
			   (setq ares (compare-arglists (getp wff1 'arguments)
						(getp wff2 'arguments)
						:prefer prefer :subnet subnet
						:types1 types1 :types2 types2
						:pred (getp wff1 'pred))))
		     (setq minres ares) 
		
		     (when (and minres cleanup)
			   (setq subs (get-the-subs))
			   (setq subs1 (union (car subs)
					      (if (listp ares) (car ares))
					      :test 'equal))
			   (setq subs2 (union (cdr subs)
					      (if (listp ares) (cdr ares))
					      :test 'equal))
			   (setq useres (list (make-comparison-info :lit1 wff1 
								 :lit2 wff2
					:subs1 subs1
					:subs2 subs2)))
			   (when traced
			       (print-compare wff1 wff2 compatible subs "    "))
			   )
		     ; also try specialists comparisons here
		     (unless (or (not cleanup) (not *spec-compare-lits*)
				 (if entering (not *fwd-spec-compare*))
				 (and (not assumption)
				      (not (var-present wff1))
				      (not (getp wff1 'assumption))
				      (not (var-present wff2))
				      (not (getp wff2 'assumption))
				      ))
			     (if compatible
				 (setq spec-res (spec-compatible-lits wff1 wff2 
						    *specialist-eval-effort*))
				 (setq spec-res (spec-incompatible-lits wff1 wff2 
						    *specialist-eval-effort*)) )
			     )
		     ;minres
		     (if (not cleanup)
			 minres
			 (append useres (if spec-res spec-res)))
		     )
	        )
	       (modified
		; compare operators and arguments
		(if (or (and compatible 
			     (equal (neg-p wff1) (neg-p wff2)))
			(and (not compatible) 
			     (not (equal (neg-p wff1) (neg-p wff2)))))
		    (let ((pres (compare-ops (getp wff1 'operator)
					     (get wff2 'operator) compatible
					     prefer subnet))
			  ares)
			 (when pres
			   (setq ares (compare-arglists (getp wff1 'arguments)
						(getp wff2 'arguments)
						:prefer prefer :subnet subnet)))
			 (and ares pres))
		    ))
	       (logical ;embedded only
                (let ((a1 (getp wff1 'arguments))
		      (a2 (getp wff2 'arguments))
		      (p1 (getp wff1 'pred))
		      (p2 (getp wff2 'pred))
		      (ans t))
		     (when (and (equal p1 p2) (equal (length a1) (length a2)))
			   (do ((at1 (car a1) (car a1))
				(at2 (car a2) (car a2)))
			       ((null at1))
			       (setq a1 (cdr a1))
			       (setq a2 (cdr a2))
			       (unless (compare-wffs at1 at2 compatible 
						     :prefer prefer
						     :types1 types1 
						     :types2 types2
						     :using-mp using-mp
						     :subnet subnet
						     :traced traced
						     :header header 
						     :entering entering
						     :cleanup nil)
				       (setq ans nil)
				       (return)) )
			   ans))
	        )
	       (quantified ;embedded - or mp key
			   ; if mp key keep prefer, else nil
                (if (and prefer (or using-mp (not cleanup)))
		    (setq prefer nil))
                (when (and (if (equal (neg-p wff1) (neg-p wff2))
			       compatible (not compatible))
		      (compare-items (getp wff1 'quantifier) 
				     (getp wff2 'quantifier) t :prefer prefer
				     :subnet subnet))
		      (if (and (not-disjoint-types (getp wff1 'variable) types1
						   (getp wff2 'variable) types2)
			       (compare-variables (getp wff1 'variable)
					     (getp wff2 'variable) prefer t
					     (and (equal (getp wff1 'quantifier) 'E)
						  )) ;nil
			       (if (and (getp wff1 'restrict) (getp wff2 'restrict))
				   (compare-wffs (getp wff1 'restrict) 
						 (getp wff2 'restrict) t 
						 :prefer prefer
						 :types1 types1 
						 :types2 types2
						 :using-mp using-mp
						 :subnet subnet
						 :traced traced 
						 :header header 
						 :entering entering
						 :cleanup nil)
				   (if using-mp nil t))
			       (compare-wffs (getp wff1 'main) 
						 (getp wff2 'main) compatible
						 :prefer prefer
						 :types1 types1 
						 :types2 types2
						 :using-mp using-mp
						 :subnet subnet
						 :traced traced 
						 :header header 
						 :entering entering
						 :cleanup nil)
			       )
			  t
			  nil ;unless many and A can be compared?
			  )
		      )
	        )
	       )
	 )
	((and (equal type1 'quantified)
	      (getp (getp wff1 'variable) 'matchable)
	      (not (member type2 '(quantified logical)))
	      (null (getp wff1 'restrict)))
	 (compare-wffs (getp wff1 'main) wff2 compatible 
		       :prefer prefer :traced traced :header header 
		       :types1 types1 :types2 types2
		       :using-mp using-mp
		       :subnet subnet
		       :entering entering :cleanup nil))
	((and (equal type2 'quantified)
	      (getp (getp wff2 'variable) 'matchable)
	      (not (member type1 '(quantified logical)))
	      (null (getp wff2 'restrict)))
	 (compare-wffs wff2 (getp wff2 'main) compatible 
		       :prefer prefer :traced traced :header header 
		       :types1 types1 :types2 types2
		       :using-mp using-mp
		       :subnet subnet
		       :entering entering :cleanup nil))
	((and (equal type1 'modified) (equal (getp wff1 'operator) 'nec))
	 (compare-wffs (car (getp wff1 'arguments)) wff2 compatible 
		       :using-mp using-mp
		       :types1 types1 :types2 types2
		       :subnet subnet
		       :prefer prefer :entering entering))
	((and (equal type2 'modified) (equal (getp wff2 'operator) 'nec))
	 (compare-wffs wff1 (car (getp wff2 'arguments)) compatible 
		       :using-mp using-mp
		       :types1 types1 :types2 types2
		       :subnet subnet
		       :prefer prefer :entering entering))
	(t nil)))
  ; cleanup is only set if this was the top level - so print trace info 
  ; if necessary and cleanup
  (when cleanup 
	(cond ((listp res) 
	       (setq subs (get-the-subs))
	       (dolist (r res)
		       (setf (comparison-info-subs1 r)
			     (union (comparison-info-subs1 r) (car subs)
				    :test 'equal))
		       (setf (comparison-info-subs2 r)
			     (union (comparison-info-subs2 r) (cdr subs)
				    :test 'equal))
		       (setf (comparison-info-lit1 r) wff1)
		       (setf (comparison-info-lit2 r) wff2)
		       (when traced 
			     (print-compare wff1 wff2 compatible 
					    (cons (comparison-info-subs1 r)
						  (comparison-info-subs2 r))
						  "    "))
		       ))
	      ((or useres spec-res)
	       (setq res useres)
	       (when spec-res (setq res (append res spec-res))))
	      (res
	       (setq subs (get-the-subs))
	       (setq res (list (make-comparison-info :lit1 wff1 :lit2 wff2
					:subs1 (car subs) :subs2 (cdr subs))))
	       (when res 
		 (when traced (print-compare wff1 wff2 compatible subs "    ")))
	       ))
	(compare-cleanup))
  res)

(defun pred-contains-variable (pred)
  (or (epi-variable pred)
      (case (getp pred 'pred-type)
	    (modified (or (pred-contains-variable (getp pred 'operator))
			  (pred-contains-variable (car (getp pred 'arguments)))))
	    (lambda (pred-contains-variable (getp (getp pred 'main) 'pred)))
	    ))
  )

(defun compare-items (item1 item2 compatible &key (prefer nil)
			    (match-const-fn nil) (subnet '/)
			    (types1 nil) (types2 nil)
			    &aux o1 o2 tt)
  (cond ((and (or (epi-variable item1) (epi-variable item2))
	      (not (and (getp item1 'pred) (getp item2 'pred))))
	 (if compatible 
	     (and (not-disjoint-types item1 types1 item2 types2)
		  (compare-variables item1 item2 prefer)) ))
	((and (symbolp item1) (symbolp item2))
	 (let ((type1 (getp item1 'entity-type))
	       (type2 (getp item2 'entity-type)))
	       (if (and (equal type1 'term) (equal type2 'wff)
			(equal (getp item1 'term-type) 'wff))
		   (setq type1 'wff))
	       (if (and (equal type2 'term) (equal type1 'wff)
			(equal (getp item2 'term-type) 'wff))
		   (setq type2 'wff))
	      (if (equal type1 type2)
		  (case type1
			(wff (compare-wffs item1 item2 compatible 
					   :types1 types1 :types2 types2
					   :subnet subnet
					   :prefer prefer :cleanup nil))
			(pred (compare-predicates nil item1 nil item2 compatible
						  prefer subnet))
			(term (compare-terms item1 item2 compatible 
					     :prefer prefer
					     :types1 types1 :types2 types2
					     :subnet subnet
					     :match-const-fn match-const-fn))
			(op (compare-ops item1 item2 compatible prefer subnet))
			(otherwise (compare-terms item1 item2 compatible 
						  :prefer prefer
					          :types1 types1 :types2 types2
						  :subnet subnet
					  :match-const-fn match-const-fn))
			)
		  ;; sometimes when a variable operator is involved the
		  ;; proper type is not set - compensate
		  (cond ((and (setq o1 (getp item1 'operator))
			      (setq o2 (getp item2 'operator))
			      (or (epi-variable-p o1) (epi-variable-p o2))
			      )
			 (if (epi-variable-p o1) (setq tt type2) (setq tt type1))
			 (case tt
			       (wff (compare-wffs item1 item2 compatible 
					   :subnet subnet
					   :types1 types1 :types2 types2
					   :prefer prefer :cleanup nil))
			       (pred (compare-predicates nil item1 nil item2 
						 compatible prefer subnet))
			       (term (compare-terms item1 item2 compatible 
					     :prefer prefer
					     :types1 types1 :types2 types2
					     :subnet subnet
					     :match-const-fn match-const-fn))
			       (op (compare-ops item1 item2 compatible prefer 
						subnet))
			       (otherwise (compare-terms item1 item2 compatible 
						  :prefer prefer
					          :types1 types1 :types2 types2
						  :subnet subnet
					  :match-const-fn match-const-fn))
			       ))
			)
		  )))
	((equal item1 item2) t)
	((member item1 (get-equivalence item2 subnet)) t)
	(t nil))
  )

; compare arglists is used internally, and is also available for the
; specialists to call during literal comparison
(defun compare-arglists (arglist1 arglist2 &key (match-const-fn nil) 
				  (prefer nil) (cleanup nil) (subnet '/)
				  (pred nil)
				  (types1 nil) (types2 nil)
				  &aux (res t) n-subnet a1 a2 nres)
  (unless (= (length arglist1) (length arglist2))
	  (return-from compare-arglists nil))
  (when (and pred (subnet-pred pred))
	(setq a1 (last1 arglist1))
	(setq a2 (last1 arglist2))
	(setq arglist1 (remove a1 arglist1))
	(setq arglist2 (remove a2 arglist2))
	(setq n-subnet (build-symbol (append (list subnet) arglist1
					     (list '-) (list pred))))
	)
  ;combine results
  (do ((a (car arglist1) (car arglist1))
       (b (car arglist2) (car arglist2)))
      ((or (null a) (null b)))
      (setq arglist1 (cdr arglist1))
      (setq arglist2 (cdr arglist2))
      (setq nres (compare-items a b t :prefer prefer :subnet subnet
				:types1 types1 :types2 types2
				:match-const-fn match-const-fn))
      (if (not nres) (return (setq res nil)))
      )
  (when (and res n-subnet)
	(setq res (compare-items a1 a2 t :prefer prefer :subnet n-subnet
				:types1 types1 :types2 types2
				  :match-const-fn match-const-fn)))
  (when cleanup 
	(if res (setq res (get-the-subs)))
	(compare-cleanup))
  res)

(defun compare-terms (term1 term2 compatible &key (match-const-fn nil)
			    (subnet '/)
			    (types1 nil) (types2 nil)
			    (prefer nil) &aux res)
  (cond ((or (epi-variable term1) (epi-variable term1))
	 (if compatible 
	     (and (not-disjoint-types term1 types1 term2 types2)
	      (compare-variables term1 term2 prefer)) ))
	((symbolp term1)
	 (case (getp term1 'term-type)
	       (wff
		(compare-wffs term1 term2 compatible :prefer prefer 
			      :types1 types1 :types2 types2
			      :subnet subnet
			      :cleanup nil) )
	       (modified
		; compare operator and arguments
		; if operator is K or K1 should we do something else?
		(when (member (getp term1 'operator) '(K K1))
		      (setq *reverse-test* t))
		(setq res
		(and compatible 
		   (compare-ops (getp term1 'operator) (getp term2 'operator) t
				prefer subnet)
		   (compare-arglists (getp term1 'arguments) 
				     (getp term2 'arguments) :prefer prefer
			      :types1 types1 :types2 types2
				     :subnet subnet)))
		(setq *reverse-test* nil)
		res)
	       (function
		; compare function and arguments
		(if (compare-ops (getp term1 'operator) 
				 (getp term2 'operator) t prefer subnet)
		    (compare-arglists (getp term1 'arguments) 
				      (getp term2 'arguments) :prefer prefer
			      :types1 types1 :types2 types2
				      :subnet subnet))
		)
	       ; quoted-exp, record, constant
	       (otherwise  ; if compatible and equal, no subs required
		(when compatible 
		      (or (equal term1 term2)
			  ;; TEST
			  (member term1 (equivalent-items term2 subnet))
			  (if match-const-fn
			      (apply match-const-fn (list term1 term2))))
		      ))
	       ))
	((equal term1 term2) t)
	(t nil))
  )

(defun compare-predicates (neg1 pred1 neg2 pred2 compatible &optional prefer
				(subnet '/)
				&aux cp)
  (cond ((or (epi-variable pred1) (epi-variable pred2))
	 (if (and compatible (not neg1) (not neg2))
	     (compare-variables pred1 pred2 prefer)) )
	((and (equal pred1 pred2)
	      (if compatible (equal neg1 neg2) (not (equal neg1 neg2))))
	 t)
	((and (setq cp (spec-compare-preds pred1 pred2)) (not (equal cp 'UNKNOWN)))
	 (if compatible (compatible cp neg1 neg2) (incompatible cp neg1 neg2)))
	((and (symbolp pred1) (symbolp pred2))
	 (let ((type1 (getp pred1 'pred-type))
	       (type2 (getp pred2 'pred-type)))
	      (cond ((equal type1 type2)
		     (case type1 
			   ;variables already handled above
			   (modified
			    ; compare operators and arguments
			    (and (if compatible (equal neg1 neg2)
				     (not (equal neg1 neg2)))
				 (compare-ops (getp pred1 'operator) 
					 (getp pred2 'operator) t prefer subnet)
				 (compare-arglists (getp pred1 'arguments)
						   (getp pred2 'arguments) 
						   :prefer prefer :subnet subnet))
			    )
			   (lambda
			    (if (and (getp pred1 'wff-type) 
				     (getp pred2 'wff-type)
				     (not neg1) (not neg2))
				(compare-wffs pred1 pred2 t 
					      :prefer prefer :cleanup nil
					      :subnet subnet)
				;otherwise compare preds and args
				(and (compare-predicates neg1 
					 (getp pred1 'operator) neg2 
					 (getp pred2 'operator) compatible
					 prefer subnet)
				     (compare-arglists (getp pred1 'arguments)
						       (getp pred2 'arguments)
						       :prefer prefer
						       :subnet subnet))
				)
			    )
			   ))
		    ((or (equal type1 'lambda) (equal type2 'lambda))
		     ; if lambda is squishable to another, try with that
		     (let ((np1 (if (equal type1 'lambda) (lambda-reduce pred1)
				    pred1))
			   (np2 (if (equal type2 'lambda) (lambda-reduce pred2)
				    pred2)))
			  (when (and np1 np2)
				(compare-predicates neg1 np1 neg2 np2 
						    compatible prefer subnet))
			  ))
		    (t nil))
	      ))
	(t nil))
  )

(defun compare-ops (op1 op2 compatible &optional prefer (subnet '/))
  (cond ((or (epi-variable op1) (epi-variable op2))
	 (if compatible (compare-variables op1 op2 prefer)) )
	((and (equal op1 op2) compatible) t)
	((and (symbolp op1) (symbolp op2))
	 (let ((type1 (getp op1 'op-type))
	       (type2 (getp op2 'op-type)))
	      (cond ((equal type1 type2)
		     (case type1 
			   (modified
			    (and (compare-ops (getp op1 'operator)
					      (getp op2 'operator) compatible
					      prefer subnet)
				 (compare-arglists (getp op1 'arguments)
					  (getp op2 'arguments) 
					  :prefer prefer :subnet subnet)))
			   (otherwise
			    (and compatible (equal op1 op2))) ) )
		    (t nil))
	      ))
	(t nil))
  )

; return the subs obtained through the comparison process

(defun get-the-subs (&aux subs-a subs-b)
  (mapc #'(lambda (index &aux elt subs)
     (when (and (epi-variable (uf-elt-arg (setq elt (ref-elt index))))
		(not (eq (setq subs (uf-elt-subs (uf-find-root elt))) elt)))
	   (setq subs (list (uf-elt-arg elt) (uf-elt-arg subs)
			    (if (uf-elt-keep-quant subs) t (uf-elt-clause-num subs))))
	   (if (= (uf-elt-clause-num elt) 1)
	       (push subs subs-a)
	       (push subs subs-b))
	   ))
	*uf-used-indices*)
  (cons subs-a subs-b)
  )

(defun print-compare (wff1 wff2 compatible subs &optional (header "")
			   &aux (subs1 (car subs))
			   (subs2 (cdr subs)))
  (print-line header "Found " (getp wff1 'print)
	      (if compatible " compatible" " incompatible")
	      " with " (getp wff2 'print))
  (unless (and (null subs1) (null subs2))
	  (print-line "        with substitutions: ")
	  (print-trace-subs subs1)
	  (print-it " in " (getp wff1 'print))
	  (print-line "                      and  ")
	  (print-trace-subs subs2)
	  (print-it " in " (getp wff2 'print))
	  (print-line)
	  ))

(defun compatible (predrel neg-p1 neg-p2)
  (if neg-p1
      (if neg-p2
          (member predrel (if *reverse-test* '(equivalent subsumes)
			      '(equivalent subsumed)))
          (member predrel '(disjoint))) ;nil
      (if neg-p2
          nil
          (member predrel (if *reverse-test* '(equivalent subsumed)
			      '(equivalent subsumes))))
      ))
 
(defun incompatible (predrel neg-p1 neg-p2)
  (if neg-p1
      (if neg-p2
          nil
          (member predrel (if *reverse-test* '(equivalent subsumed)
			      '(equivalent subsumes))))
      (if neg-p2
          (member predrel (if *reverse-test* '(equivalent subsumes)
			      '(equivalent subsumed)))
          (member predrel '(disjoint)))
      ))
       

	     
(defun not-disjoint-types (item1 types1 item2 types2 &aux res
				 t1 t2)
  (setq t1 (cdr (assoc item1 types1)))
  (setq t2 (cdr (assoc item2 types2)))
  (cond ((and t1 t2 *use-types-in-unify*)
	 (dolist (typ t1)
		 (dolist (typ2 t2)
			 (setq res (spec-compare-preds typ typ2))
			 (if (not (equal res 'disjoint))
			     (return-from not-disjoint-types t))
			 ))
	 (if (traced-p 'qa-test)
	     (print-line "         Could not unify " item1 " " t1 " with " item2
			 " " t2 " because they have disjoint types"))
	 nil)
	(t t))
  )
