;       Last change    August 1993
(in-package equal)   


(tweakable 'epilog::*unique-names-assumption* t
  "Indicates whether or not we are assuming that user assigned names indicate unique individuals"
  'boolean)

; evaluation functions

; evaluate tries to determine whether the  given literal is true, false
; or unknown, after first ensuring that it is in the set specialist's domain.

(defun evaluate (netname lit negp pred arglist effort &aux
		      (net (determine-net netname))
		      (a3 (caddr arglist)))
  (declare (ignore lit))
  (if (and (null a3)
	   (not (non-constant-arg arglist)))
      (eval-with-neg net negp pred arglist effort)
      'unknown)
  )

; eval-with-neg evaluates the given literal, taking into account whether it
; is negated. If so, the positive literal is evaluated, and the result reversed. 
(defun eval-with-neg (net negp reln args effort &aux
			  (posans (evaluate-equal net reln args effort)))
    
  (cond ((equal posans 'yes)
         (if negp
	     'no 
	     'yes))
	((equal posans 'no)
	 (if negp
	     'yes
	     'no))
	(t posans))
  )
 
; test-equal tests to see if two arguments are equal

(defun test-equal (net a1 a2 effort &aux
		       (sort1 (find-sort a1))
		       (sort2 (find-sort a2)))
  (declare (ignore effort))
  (cond ((equal a1 a2) 'yes)

	; check equivalence and no-equal sets as well
	; (check to see if a1 is a member of a2-equal, or of a2-non-equal)
	((check-equivalences net a1 a2) 'YES)

	((check-non-equal net a1 a2) 'NO)

	; if they have different sorts they cannot be equal
	((and sort1 sort2 (not (equal sort1 sort2))) 'NO)

	; perhaps also compare types if they are available?

	((and epilog::*unique-names-assumption*
	      (equal net (determine-net '/))
	      (not epilog::*insertion*)  ;;?
	      (not (getp a1 'epilib::indef))
	      (not (getp a2 'epilib::indef)))
	 'NO)

	; otherwise we don't know
	(t 'unknown)
	)
  )

; evaluate-equal tries to evaluate whether the given relation holds with the
; arguments.

(defun evaluate-equal (net pred arglist effort &aux
			 (a1 (car arglist))
			 (a2 (cadr arglist))
			 (res 'unknown)
			 )
  (unless (null arglist)
	  (when (traced-p 'epilog::equality-eval)
	        (print-line "    Equality Specialist: Evaluating " (print-info a1) 
			  " " (print-info pred) " " (if a2 (print-info a2) ""))
		(if (not (equal (episet-subnet-identifier net) '/))    
		    (print-it "    in subnet " (episet-subnet-identifier net)))
		)
	  (setq res (test-equal net a1 a2 effort))
	)
  ; return a yes, no or unknown answer
  (if (null res) (setq res 'no))
  (if (and res (not (member res '(unknown no)))) (setq res 'yes))
  (if (and (traced-p 'epilog::equality-eval) (not (equal res 'unknown)))
      (print-line "    Equality Specialist: Evaluated " 
	  		       (print-info a1) " " (print-info pred) " " 
			       (if a2 (print-info a2) "") " to " res))
  res)
