;       Last change    August 1993
(in-package number)
;        Routines to get and create Number nets and numbers

; Getnet returns the number net corresponding to the context given. If no
; such net exists, one is created.

(defun getnet (context)
  (let ((context-array (retrieve-specialist-info
			*environment*
			'number
			(lambda ()
			  (make-hash-table :test 'eql
					   :size '100
					   :rehash-size '10
					   :rehash-threshold '.8)))))
    (when (null (geth context context-array))
      (seth context
	    context-array
	    (newnet context)))
    (geth context context-array)))

; determine net gets the number net associated with the subnet for the
; netname given (/concpet1/concept2/...) and sets the current net in
; case any functions are used

(defun determine-net (netname &aux
			      (net nil))
  (if (null netname)
      (setq net (getnet '/))
      (setq net (getnet netname))
      )
  (setq current-net net)
  net
  )

; Getnumber returns the number for item in the number net. If there is no
; entry, once is created and returned.

(defun getnumber (net item sort &aux
		     (narray (number-subnet-numbers net)))
  (when (null (geth item narray))
	(change-hash item narray (make-epinumber :name item :ntype 
					      (if (equal sort 'number)
						  'real
						  sort)))
	)
  (geth item narray)
  )

; Findnumber returns the number in the net corresponding to item if there
; is one. If not, nil is returned

(defun findnumber (net item)
  (if (or (null net) (null item))
      nil
      (geth item (number-subnet-numbers net))
      )
  )

; Newnet creates a new number net, including the number table

(defun newnet (ident &aux

		     ; hash table for numbers

		     (nt (make-hash-table :test 'eql
					  :size '100
					  :rehash-size '10
                                          :rehash-threshold '.8))) ; changed!
                                        ;; threshold value changed from 9
                                        ;; [2/24/95 CH]

  (make-number-subnet :identifier ident
		      :numbers nt)
  )

