;       Last change    August 1993
(in-package part)


; evaluation functions

(defvar *current-part-literal* nil)

; evaluate tries to determine whether the  given literal is true, false
; or unknown, after first ensuring that it is in the part specialist's domain.

(defun evaluate (netname lit negp pred arglist effort &aux
		      (net (determine-net netname))
		      (a3 (caddr arglist)))
  (declare (ignore lit))
  (if (and (null a3)
	   (not (non-constant-arg arglist))
	   (if (equal pred 'equal)
	       (ok-part-args net (car arglist) (cadr arglist))
	       t)
	   )
      (eval-with-neg net negp pred arglist effort)
      'unknown)
  )

(defun ok-part-args (net a1 a2 &aux
			 (p1 (findpart net a1))
			 (p2 (findpart net a2)))
  (cond (p1
	 (if p2 
	     (or (part-super-parts p1) (part-super-parts p2))
	     t))
	(p2 t)
	((member-if #'(lambda (a) (part-pred-p a)) (grab-types a1)) t)
	((member-if #'(lambda (a) (part-pred-p a)) (grab-types a2)) t)
	(t nil))
  )

; eval-with-neg evaluates the given literal, taking into account whether it
; is negated. If so, the positive literal is evaluated, and the result reversed. 
(defun eval-with-neg (net negp reln args effort &aux
			  (posans (evaluate-part net reln args effort)))
    
  (cond ((equal posans 'yes)
         (if negp
	     'no 
	     'yes))
	((equal posans 'no)
	 (if negp
	     'yes
	     'no))
	(t posans))
  )
 

; evaluate-part tries to evaluate whether the given relation holds with the
; arguments.
; Can test limited equality, and membership.

(defun evaluate-part (net pred arglist effort &aux
			 (reln (find-symbol (symbol-name pred) 'part))
			 (a1 (car arglist))
			 (a2 (cadr arglist))
			 (res 'unknown))
  (unless (null arglist)
	  (when (traced-p 'epilog::part-eval)
	        (print-line "    Part Specialist: Evaluating " (print-info a1) 
			  " " (print-info pred) " " (if a2 (print-info a2) ""))
		(if (not (equal (part-subnet-identifier net) '/))     
		    (print-it "    in subnet " (part-subnet-identifier net)))
		)
	  (case reln
		(part-of (setq res (test-part net a1 a2 effort)))
		(equal (setq res (test-part-equal net a1 a2 effort)))
		(otherwise
		 (if (and (null a2) (type-pred-p pred))
		     (setq res (part-type-test net pred a1 effort))))
		)
	)
  ; return a yes, no or unknown answer
  (if (null res)
      (setq res 'no))
  (if (and res (not (member res '(unknown no))))
      (setq res 'yes))
  (if (and (traced-p 'epilog::part-eval)
	   (not (equal res 'unknown)))
      (print-line "    Part Specialist: Evaluated " (print-info a1) " " 
		  (print-info pred) " " (if a2 (print-info a2) "") 
			       " to " res))
  res)

(defun test-part-equal (net a1 a2 effort &aux
			    (p1 (findpart net a1))
			    (p2 (findpart net a2))
			    o1 o2)
  (cond ((or (null p1) (null p2)) 
	 (dolist (t1 (grab-types a1))
		 (dolist (t2 (grab-types a2))
			 (if (equal (hier-rel t1 t2) 'disjoint)
			     (return-from test-part-equal 'NO))))
	 'unknown)
	((equal (part-name p1) (part-name p2)) 'yes)
	((not (equal (part-collection p1) (part-collection p2))) 'no)
	((and (setq o1 (part-owning-entity p1))
	      (setq o2 (part-owning-entity p2))
	      (or (equal (eval-with-spec *current-netname*
					 (normalize (list o1 'equal o2)) effort)
			 'NO)
		  (incompat-types p1 o2)))
	 'NO)
	((and (part-types p1)
	      (part-types p2)
	      (incompat-types a1 a2))
	 'NO)
	(t 'unknown))
  )

(defun test-part (net part entity effort &aux
		      ppart epart oe1 oe2 res coll-part)
  (setq epart (findpart net entity))
  (if epart (setq oe2 (part-owning-entity epart))
      (if (individual entity) (setq oe2 entity)))
  (setq ppart (findpart net part))
  (when ppart (setq oe1 (part-owning-entity ppart)))
  (if (if ppart (part-collection ppart) (part-set part)) (setq coll-part t))
  (cond ((and (null epart) (null ppart))
	 (if coll-part
	     (coll-part-type-compat part (grab-types entity) 
				    (grab-types entity))
	     (case (part-type-compat (grab-types part) (grab-types entity)
			   (grab-types entity))
		   (NO 'NO) (otherwise 'unknown)) ))
	((and (null epart) (not (individual entity)))
	 ; no part info about 2nd part 
	 (if oe1
	     (let ((resp (if coll-part
			    (coll-part-type-compat part (grab-types entity) 
				    (grab-types oe1))
			    (part-type-compat (grab-types part)
				(grab-types entity)
				(grab-types oe1)))) )
		  (if (and (equal res 'unknown) (individual entity))
		      (incompat-types oe1 entity)
		      (case resp (NO 'NO) (otherwise 'unknown))))
	     'unknown) )
	((null ppart)
	 ; check to see if it is possible for entity to have a subpart of that
	 ; nature
	 (if oe2
	     (if coll-part 
		 (coll-part-type-compat part (grab-types entity) 
				    (grab-types oe2))
		 (case (part-type-compat (grab-types part)
			       (grab-types entity)
				(grab-types oe2))
		       (NO 'NO) (otherwise 'unknown)))
	     'unknown) )
	((and oe1 oe2
	      (setq res (eval-with-spec (part-subnet-identifier net)
					(normalize (list oe1 'equal oe2)) 
					effort))
	      (member res '(YES NO)))
	 res)
	((and oe1 oe2 (setq res (incompat-types oe1 oe2)) (equal res 'NO))
	 'NO)
	((check-supers net ppart entity))
	((or oe1 oe2) 
	 (if coll-part 
	     (coll-part-type-compat part (grab-types entity) 
				    (grab-types (cond (oe1) (t oe2))))
	     (case (part-type-compat (grab-types part) (grab-types entity)
			   (grab-types (cond (oe1) (t oe2))))
		   (NO 'NO) (otherwise 'unknown))))
	((individual entity) 'unknown)
	(t 'unknown)
	)
  )

(defun incompat-types (oe1 oe2 &aux
			   (tps1 (grab-types oe1))
			   (tps2 (grab-types oe2))
			   res)
  (dolist (t1 tps1)
	  (dolist (t2 tps2)
		  (setq res (hier-rel t1 t2))
		  (when (equal res 'DISJOINT) 
			(return-from incompat-types 'NO))
		  ))
  'UNKNOWN)

(defun check-supers (net ppart entity &aux
			   res npart
			   (sups (part-super-parts ppart)))
  (dolist (p sups)
	  (if (equal p entity)
	      (return-from check-supers 'YES)
	      (if (setq npart (findpart net p))
		  (setq res (check-supers net npart entity)))
	      )
	  (if res (return-from check-supers res))
	  )
  )

(defun part-type-compat (p1 p2 etypes &aux tot part-hiers res)
  ; check for disjointedness between p1 and p2 types within part hierarchies
  ; available to etypes (do only part hierarchies - parts themselves 
  ; can be disjoint)
  (dolist (tp etypes)
     (setq part-hiers (find-part-hiers tp))
     (dolist (pt1 p1)
	     (dolist (pt2 p2)
		     (setq res (hier-rel pt1 pt2 part-hiers))
		     (unless (member res '(subsumed))
	;;; CHANGE
		     (dolist (p (intersection part-hiers 
					      (intersection (getp pt2 'hier) 
							    (getp pt1 'hier))))
			     (setq res (epilib::hier-rel-in-same-hier p pt1 pt2))
			     (when (member res '(disjoint subsumed))
				   (return))) )
		     (when (member res '(disjoint))
			   (return-from part-type-compat 'NO))
		     (when (member res '(subsumed))
			   (dolist (h part-hiers)
				   (when (and (setq tot (cdr 
					   (assoc h (getp pt2 'total))))
					      (equal (cadr tot) 1))
				       (return-from part-type-compat 'YES) )
				   ))
		     ))
	  )
  ; if we get here we have no answer
  (if (equal p2 etypes) ;ie given owning entity
      (check-normal-levels p1 p2)
      'unknown)
  )

(defun coll-part-type-compat (part p2 etypes &aux fn ptypes c)
  (setq fn (make-fn-arg 'epilog::member-type-of part))
  (setq ptypes (spec-eval-fn fn))
  (if (not (equal ptypes fn))
      (setq ptypes (list ptypes))
      (setq ptypes nil))
  (if ptypes
      (case (part-type-compat ptypes p2 etypes) 
	    (NO 'NO)
	    (YES
	     ; also check for totals
	     (if (numberp (setq c (spec-eval-fn (make-fn-arg 'cardinality-of part))))
		 (let (total bounds)
		      (dolist (p ptypes) (setq bounds (part-total p etypes))
			      (when bounds (setq total bounds)))
		      (if (and c total (> c total)) 'NO 'unknown)
		      )))
	    (otherwise 'unknown))
      'unknown)
  )

(defun check-normal-levels (part-types owner-types &aux
				       norm res phiers normals)
  ; check to see if that type of part can exist on any part hierarchy
  ; accessible to owner-types
  (setq phiers (remove-duplicates 
		(apply 'append (mapcar 'find-part-hiers owner-types))))
  ; if we can show that the given part type cannot exist on any of the
  ; hierarchies, we can return 'NO
  (dolist (p part-types)
	  (setq normals (getp p 'normal-level))
	  (when (and normals (member p *familiar-parts*))
		(dolist (h phiers)
			; check only those part hierarchies with the same
			; types as the part hierarchies the part HAS been
			; defined on
			; i.e. if only exists on FORM hierarchies, don't 
			; bother checking FUNCTION or STATE hierarchies
			(if (setq norm (cdr (assoc (cond ((getp h 'part-type)) 
							 (t 'form))
				   normals)))
			    (setq res (part-in-hier p h norm)))
			(when (equal res 'unknown) (return))
		  ))
	  (when (equal res 'unknown) (return))
	  )
  (if res res 'unknown))

(defun part-in-hier (part hier normal &aux nodes)
  (when (hier-property hier 'non-exhaustive)
	(return-from part-in-hier 'unknown))
  ; look at the hierarchy, comparing each subnode with part until bottom of
  ; hierarchy, or until level normal is reached
  ; if bottom reached first, answer is unknown
  (if (getp hier 'hier-change) (number-hier hier))
  (setq nodes (list (getp hier 'root)))
  (do ((i 0 (+ i 1)))
      ((> i normal))
      (dolist (n nodes)
	      (if (member (hier-rel part n) '(equivalent subsumed subsumes))
		  (return-from part-in-hier 'YES)) ;unknown?
	      )
      (setq nodes (apply 'append 
			 (mapcar #'(lambda (a) (get-sub hier a)) nodes)))
      (when (and (null nodes) (< i normal))
	    (return-from part-in-hier 'unknown))
      )
  ; if we get here, we've searched deep enough and not found it - since it
  ; is a familiar part on an exhaustive hierarchy, we can say that it can't
  ; be there
  'NO)

(defun find-part-hiers (item)
  (cond ((getp item 'parts))
	(t (apply 'append 
		  (mapcar 'find-part-hiers (get-all-super item)))))
  )

(defun part-type-test (net pred arg effort &aux
			   (ppart (findpart net arg))
			   pp ptypes oe allowed res)
  (cond ((or (equal pred 'part)
	     (and (setq pp (pred-parts pred))
		  (equal (cadr pp) 'part)))
	 (part-part-test pred arg effort ppart pp))
	; if the arg is a part of something, and the pred is an entire entity
	; type, cannot be true - ?
	((and ppart (part-collection ppart)) 'no)
	((and ppart 
	      (setq oe (part-owning-entity ppart))
	      (setq ptypes (part-types ppart)))
	 (setq allowed (apply 'append (mapcar 'find-part-hiers 
					      (grab-types oe))))
	 (dolist (p ptypes)
		 (setq res (hier-rel pred p allowed))
		 (when (member res '(equivalent))
		       (return-from part-type-test 'YES))
		 (when (member res '(disjoint))
		       (return-from part-type-test 'NO))
		 )
	 'unknown)
	(t 'unknown))
  )

(defun part-part-test (pred arg effort ppart pp &aux 
			    ptypes allowed testpred res)
  (declare (ignore effort pred))
  (if pp 
      (setq testpred (car pp))
      (setq testpred 'entity))
  (if  nil ;(and pp (type-pred-p (car pp)) (not (part-pred-p (car pp))))
      (setq allowed (find-part-hiers (car pp)))
      (if ppart
	  (let* ((oe (part-owning-entity ppart)))
		(when oe
		      (setq allowed (apply 'append (mapcar 'find-part-hiers 
						   (grab-types oe))))
		  )) ))
  (when (and allowed 
	     (setq ptypes (if ppart 
			      (part-types ppart) 
			      (grab-types arg))))
	(dolist (p ptypes)
		(setq res (hier-rel testpred p allowed))
		(when (member res '(disjoint))
		      (return-from part-part-test 'NO))
		(when (member res '(subsumes))
		      (return-from part-part-test 'YES))
		)
	)
  'unknown)

