(in-package response)

;       Last change    August 1993

(defvar *basic-nouns* '(thing person human) 
  "Indicates which nouns can be combined with every and some") 

(defvar *said* nil)
(defvar *first* nil)
(defvar *conj* nil)
(defvar *current-sentence* nil "Contains only nps used in curent sentence")

(defun check-noun-property (noun property) 
  (and noun (symbolp noun) (member property (noun-properties noun))))

; fill-gaps fills the gaps in the top level sentence

(defun fill-gaps (frag &aux newfrag)
  ; reset sentential flags
  (setq *current-sentence* nil)
  (setq *conj* nil)
  ; make sure the fragment will be filled
  (manip-frag frag)
  (setq newfrag (fill-gap-frag frag t)) 
  newfrag)

(defvar *done-frags* nil)

(defun manip-frag (frag)
  (setq *done-frags* nil)
  (do-manip-frag frag)
  )

; manip frag makes sure that the negations and tense information are on the
; appropriate fragment part

(defun do-manip-frag (frag &aux rr plist pp as vp asp)
  (when (and frag (not (member frag *done-frags*)))
	(push frag *done-frags*)
  (case (fragment-fragtype frag)
	(S (when (fragment-parts frag) 
		 (setp frag 'parts (check-Sconj (fragment-parts frag))))
	   (when (setq pp (fragment-ADVP frag))
		 (setq plist (cond ((fragment-parts pp)) (t (list pp))))
		 ;check to see if should passivize
		 (dolist (f plist)
			 (when (fragment-S f)
			       (setq rr (try-passive frag (fragment-S f)))
			       (when rr
				    (setq frag (car rr))
				    (setp f 'S (cadr rr)))))
		 (when (fragment-negated (car plist))
		       (setp (car plist) 'negated nil)
		       (setp frag 'negated t))
		 )
	   (when (and (fragment-VP frag) (fragment-adv frag))
		 (setp (fragment-VP frag) 'adv  (combine-l (fragment-adv frag)
					(fragment-adv (fragment-VP frag))))
		 (setp frag 'adv nil) )
	   (when (and (setq vp (fragment-VP frag)) 
		      (setq asp (fragment-aspect frag))
		      (if (setq as (fragment-aspect vp))
			  (not (equal as 'prespart))
			  t))
		 (if as (setp vp 'aspect (if (listp as)
					     (if (listp asp) (append asp as)
						 (append (list asp) as))
					     (if (listp asp) (append asp (list as))
						 (list asp as))))
		     (setp vp 'aspect (fragment-aspect frag)))
		 (if (equal (fragment-aspect frag) 'prespart)
		     (push vp  *nomed-verbs*))
		 (setp frag 'aspect nil))
	   (when (and (setq vp (fragment-VP frag)) (fragment-tense frag)
		      (not (fragment-tense vp)))
		 (setp vp 'tense (fragment-tense frag))
		 (setp frag 'tense nil))
	   (when (and (setq vp (fragment-VP frag)) (fragment-mood frag)
		      (not (fragment-mood vp)))
		 (setp vp 'mood (fragment-mood frag))
		 (setp frag 'mood nil))
	   )
	(NP (if (fragment-negated frag)
		(propagate-neg frag (fragment-head frag)))
	 )
	(AP (if (fragment-negated frag)
		(propagate-neg frag (fragment-head frag))))
	((PP ADVP)
	 (if (fragment-negated frag) (propagate-neg frag (fragment-head frag)))
	 )
	(RELC )
	(VP 
	 (if (and (setq as (fragment-aspect frag)) (fragment-parts frag))
	     (dolist (v (fragment-parts frag))
		     (setp v 'aspect as)
		     (if (equal as 'prespart) (push v *nomed-verbs*))) ))
	)
	; beter way to do negation and tense too?
  (setq plist (symbol-plist frag))
  (do ((prop (car plist) (car plist))
       (value (cadr plist) (cadr plist)))
      ((null prop))
      (setq plist (cddr plist))
      (cond ((member prop *fragment-types*) (do-manip-frag value))
	    ((member prop *fragment-groups*) 
	     (mapcar 'do-manip-frag value)))
      ) )
  frag)

(defun combine-l (a b)
  (if (listp a)
      (if (listp b)
	  (append a b)
	  (append a (list b)))
      (if (listp b)
	  (append (list a) b)
	  (list a b)))
  )


(defvar *filled-frags* nil)

(defun fill-gap-frag (item &optional s)
  (setq *filled-frags* nil)
  (do-fill-gap-frag item s)
  )

; fill-gap-frag calls the appropriate filling routine depending on the type of
; fragment involved

(defun do-fill-gap-frag (item &optional s &aux plist
			   (fragtype (fragment-FRAGTYPE item)))
  (unless (member item *filled-frags*)
	  (push item *filled-frags*)
  (when (and s (not (equal fragtype 'S))) 
	(setq item (make-into-S item))
	(setq fragtype (fragment-fragtype item)))
  (setq item
  (case fragtype
	(S (fill-gap-S item))
	(NP (fill-gap-NP item))
	(PP (fill-gap-PP item))
	(ADVP (fill-gap-ADVP item))
	((VP RelC) (fill-gap-VP item))
	(AP item)
	(otherwise item) 
	))
  (setq plist (symbol-plist item))
  (do ((prop (car plist) (car plist))
       (value (cadr plist) (cadr plist)))
      ((null prop))
      (setq plist (cddr plist))
      (cond ((member prop *fragment-types*) 
	     (setp item prop (do-fill-gap-frag value)))
	    ((member prop *fragment-groups*) 
	     (setp item prop (mapcar 'do-fill-gap-frag value))))
      ))
  item)

(defun S-advp-into-S-PP (frag &aux av ns np)
  ; change into NP with PP
  (setq av (fragment-advp frag))
  (setp av 'name (fragment-name frag))
  (setp av 'frag-type 'PP)
  (when (setq ns (fragment-S av) )
	(setp ns 'frag-type 'NP)
	(setp av 'NP ns)
	(setp av 'S nil))
  (dolist (p (fragment-parts av))
	  (setp p 'frag-type 'PP)
	  (when (setq ns (fragment-S p) )
		(setp p 'name (fragment-name frag))
		(setp p 'NP ns)
		(setp ns 'frag-type 'NP)
		(setp p 'S nil))
	  )
  (setq np (cond ((fragment-NP frag))
		 (t (make-into-NP (fragment-name frag)))))
  (setp np 'number (fragment-number frag))
  (setp np 'head frag)
  (setp frag 'NP np)
  (setp frag 'name nil)
  (setp frag 'ADVP nil)
  (setp frag 'VP (make-frag 'VP (cons 'name (fragment-name np))
			    (cons 'verb 'be)
			    (list 'objects av)))
  (setp (fragment-VP frag) 'head frag)
  ;(setq vp1 (fragment-vp frag))
  frag)

; fill-gap-S fills in gaps for a sentence fragment

(defun fill-gap-S (frag &aux info
			(save-neg *already-neg*))
  (when (and (not (fragment-VP frag) )
	     (not (getp frag 'copy))
	     (setq info (find-entity frag))
	     (equal (fragment-fragtype info) (fragment-fragtype frag))
	     (if (equal (fragment-fragtype info) 'S)
		 (equal (fragment-aspect info) (fragment-aspect frag))
		 t)
	     (not (fragment-negated info))
	     (or (epi-variable (fragment-name frag))
		 (skolem-constant-p (fragment-name frag))))
	(when (and (parent-of frag info) (null (noun-info frag)))
	      (unless (equal frag info) (setp frag 'required nil) )
	      (return-from fill-gap-S frag))
	(copy-info info frag)
	(manip-frag frag)
	)

  (when (and (fragment-ADVP frag)
	     (not (or (fragment-parts frag) (fragment-VP frag)))
	     (not (fragment-VP (fragment-S (fragment-ADVP frag)))))
	; change into NP with PP
	(setq frag (S-advp-into-S-pp frag))
	)
	   
  (when (or (fragment-negated frag) 
	    (fragment-negated (fragment-vp frag)))
	(setq frag (move-negation frag)))

  (set-subject t (fragment-NP frag))

  (if (and (fragment-NP frag) (fragment-number (fragment-NP frag)))
      (add-number (fragment-number (fragment-NP frag)) (fragment-ADVP frag)) )
	
  (unless (or (null (fragment-name frag))
	      (equal (fragment-name frag) 'removed))
	  (pushnew frag *current-sentence*))

  (setq *already-neg* save-neg)
  frag)

(defun check-there (frag &aux np vp objs)
  (when (and (setq np (fragment-np frag))
	     (setq vp (fragment-vp frag))
	     (equal (getp vp 'verb) 'be)
	     (setq objs (fragment-objects vp))
	     (not (fragment-relc np))
	     (same-name np (car objs))
	     (or (null (fragment-noun np))
		 (equal (fragment-noun np) (fragment-noun (car objs))))
	     (equal (fragment-fragtype (car objs)) 'NP)
	     (not (member (find-sort (fragment-name np)) 
			  '(episode propos)))
	     (or (skolem-constant-p (fragment-name np))
		 (null (fragment-name np))
		 (epi-variable (fragment-name np)))
	     (not (member np *said* :test 'same-name)))
	(setp np 'pronoun "there"))
  )

; These routines are responsible for filling in gaps in the fragments -
; determiners in NP's, etc., as well as ensureing that every sentence
; fragment has a verb.

(defun quant-det-plur (quant &optional arg)
  (case quant
	(A (if arg "all" nil))
	(most "most")
	(many "many")
	(some "some") ;any
	(few "few")
	(none "no")
	(wh "what")
	(t (if (and quant (symbolp quant))
	       (if (equal (getp quant 'quant-type) 'modified)
		   (let ((q (quant-det-plur (getp quant 'operator)))
			 (a (mapcar #'(lambda (a) (quant-det-plur a t))
				    (getp quant 'arguments))))
			(if (listp q) (squish (append q (list " ") a))
			    (squish (append (list q " ") a))))
		   (string-downcase (princ-to-string quant)))
	       ""))
	)
  )

; returns t if name is a variable in a positive context
(defun plus-context (name)
  (not (and (epi-variable name) (getp name 'matchable)))
  )

(defun find-verb (item &aux
		       (ftype (fragment-fragtype item)))
  (when item
  (case ftype
	(VP (fragment-verb item))
	((PP AP NP) (find-verb (fragment-head item)))
	))
  )

; decide-determiner decidew what the determiner for a variable should be,
; using its number, whehter it is negated, whether it is the subject, etc.
; neg is either argfrag's negation status, or argfrag's head negation status

(defun decide-determiner (argfrag &aux pp q
				  (nnoun (fragment-noun argfrag))
				  (number (fragment-number argfrag)))

  (when (and nnoun (not (fragment-print argfrag))
	     (not (fragment-determiner argfrag)))
	       
	(cond ; mass nouns
	      ((check-noun-property nnoun 'MASS)
	       (when (equal (fragment-objtype argfrag) 'constant)
		     (add-determiner "some" argfrag)))

	      ; constants
	      ((or (equal (fragment-objtype argfrag) 'constant)
		   (and (equal (fragment-objtype argfrag) 'function)
			(not (non-constant-arg (list (fragment-name argfrag))))))
	       (unless (equal number 'plural)
		       (if (and (member (fragment-noun argfrag)
					*basic-nouns*)
				(null (fragment-AP argfrag)))
			   (add-determiner "some" argfrag)
			   (if (and (not (getp (fragment-name argfrag) 
					       'unique))
				    (or (getp (fragment-name argfrag) 
					      'indef)
					(and (not (fragment-subject argfrag))
					     (equal (find-verb (fragment-head argfrag)) 'be)
					     )))
			       (add-determiner "a" argfrag)
			       (add-determiner "the" argfrag)))) )

	      ;; negations
	      ((fragment-negated argfrag) 
	       (cond ((setq q (fragment-quantifier argfrag))
		      (case q
		      ((A E) (add-determiner "no" argfrag))
		      (t (add-determiner 
			  (squish (list "not " (quant-det-plur q))) argfrag))))
		     (t (add-determiner "no" argfrag)))
	       (setp argfrag 'negated nil) 
	       (setq *already-neg* t)
	       )

	      ;; negation elsewhere in sentence
	      ((and *already-neg*
		  (not (equal (fragment-fragtype (fragment-head argfrag)) 'PP)))
	       (if (equal number 'plural)
		   (if (plus-context (fragment-name argfrag))
		       (add-determiner "all" argfrag)
		       (add-determiner "any" argfrag))
		   (if (plus-context (fragment-name argfrag))
		       (add-determiner "every" argfrag)
		       (if (and (member (fragment-noun argfrag)
					*basic-nouns*)
				(null (fragment-AP argfrag)))
			   (add-determiner "any" argfrag)
			   (add-determiner "a" argfrag))) 
		   ))

	      ;; first variable
	      (*first* 
	       (setq *first* nil)
	       ; for the first variable, use "all" or "every" 
	       (if (equal number 'plural)
		   (add-determiner (quant-det-plur 
				    (fragment-quantifier argfrag)) argfrag)
		   (if (equal (fragment-quantifier argfrag) 'none)
		       (add-determiner "no" argfrag)
		   (if (fragment-relc argfrag)
		       (add-determiner "any" argfrag)
		       (if (and (member nnoun *basic-nouns*)
				(null (fragment-AP argfrag)))
			   (if (and (not (plus-context (fragment-name argfrag)))
				    (not *conj*))
			       (add-determiner "every" argfrag)
			       (add-determiner "some" argfrag))
			   (add-determiner "a" argfrag))
		       ) )
		   ))

	      ;; check for other args with same noun, adj
	      ((ambiguous-np argfrag))

	      ; otherwise a variable other than the first
	      (t ;otherwise
		 (if (and (member nnoun *basic-nouns*)
			  (null (fragment-AP argfrag))
			  (equal number 'singular))
		     ; want someone or something instead of a person or a thing
			 (add-determiner "some" argfrag)

		     ; for mass and plural, say nothing - people like food,
		     ;  wolves liek to eat creatures
		     (if (equal number 'plural)
			 (add-determiner "" argfrag)

			 ; otherwise, singular, use a
			 (add-determiner "a" argfrag)))
		 )
	      )
	; match determiner for any "of x" PP's beneath this NP
	(when (setq pp (fragment-PP argfrag))
	      (if (fragment-parts pp)
		  (mapcar #'(lambda (p) 
			  (if (and (null (fragment-adv p))
				   (equal (fragment-objtype p) 'variable)
				   (equal (fragment-PREP p) 'of))
			      (decide-determiner (fragment-NP p) )))
			  (fragment-parts pp))
		  (if (and (null (fragment-adv pp))
			   (equal (fragment-objtype pp) 'variable)
			   (equal (fragment-PREP pp) 'of))
		      (decide-determiner (fragment-NP pp) ))) )

	; ensure no determiners for NPor or NPand beneath this NP
	(if (fragment-parts argfrag)
	    (dolist (l (fragment-parts argfrag)) 
		    (add-determiner "" l)))
	)
  (fragment-determiner argfrag)
  )

;parent-of returns t if kid has parent anywhere in its ancestry( including
;itself

(defun parent-of (kid parent)
  (cond ((null parent) nil)
	((null kid) nil)
	((equal kid parent) t)
	(t (parent-of (fragment-head kid) parent)))
  )

; decide-number decides whether an NP should be singular or plural.
; If it is the first variable, it will be plural, unless it has noun
; person or thing (everyone and everything sound better than all people and
; all things)
; Other variables will have the same number as the first.
; Negation reverses the number.

(defun decide-number (argfrag &aux pp
			(nnoun (fragment-noun argfrag))
			(num (fragment-number argfrag))
			)
  (when (and (not num) (fragment-parts argfrag))
	(if (equal (fragment-conj argfrag) 'and)
	    (setq num 'plural)
	    (setq num (fragment-number (car (fragment-parts argfrag)))))
	(setp argfrag 'number num) )
  (when (and num (not *already-universal*))
	(setq *already-universal* num))
  (unless (or num (not (equal (Fragment-objtype argfrag) 'variable)))
  (if *already-universal*
      ; if not the first variable, set the number to the previous  number
      (setq num *already-universal*)
      ;otherwise first, so set the number to plural unless 
      ; person or thing
      (let ()
	   (if (or (equal (fragment-quantifier argfrag) 'E)
		   (and (member nnoun *basic-nouns*)
			(member (fragment-quantifier argfrag) '(A E none))
		;	(not (member (fragment-quantifier argfrag) 
		;		'(few many most some)))
			)
		   )
	       (setq num 'singular)
	       (setq num 'plural))
	   (setq *already-universal* num))
      )

  ; if something has been previously negated, reverse number unless 
  ; thing or person
  (when (and (or *already-neg* (fragment-negated argfrag) )
	     (equal (fragment-objtype argfrag) 'variable)
	     (not (plus-context (fragment-name argfrag))))
	(case num 
	      (plural 
	       (when (member (fragment-quantifier argfrag) '(A E none))
	       ;(unless (member (fragment-quantifier argfrag)
	;		       '(few many most some)) )
		       (setq num 'singular)))
	      (singular (if (and (member nnoun *basic-nouns*)
				 (member (fragment-quantifier argfrag) '(A E none))
				 ;(not (member (fragment-quantifier argfrag) 
				;	      '(few many most some)))
				 (not (fragment-AP argfrag)))
			    (setq num 'singular)
			    (setq num 'plural))
			    ))
	))

  ; if a mass noun, set number to singular so verbs come out right
  (if (setq nnoun (fragment-NOUN argfrag))
      (cond ((check-noun-property nnoun 'MASS)
	     (if (getp argfrag 'number)
		 (setp argfrag 'number 'singular)
		 (add-number 'singular argfrag) ))
	    ((and (member nnoun *basic-nouns*)
		  (member (fragment-quantifier argfrag) '(A E none))
		  ;(not (member (fragment-quantifier argfrag) 
		;	       '(few many most some)))
		  (null (fragment-AP argfrag)))
	     (add-number 'singular argfrag) )
	     
	    (t (add-number num argfrag) ) )
      (add-number num argfrag))

  ; ensure that all PP of's beneath this have thee same number
  (when (setq pp (fragment-PP argfrag))
	(if (fragment-parts pp)
	    (mapcar #'(lambda (p) 
		  (if (and (null (fragment-adv p))
			   (equal (fragment-objtype p) 'variable)
			   (equal (fragment-PREP p) 'of))
		      (add-number num (fragment-NP p))))
		    (fragment-parts pp))
	    (if (and (null (fragment-adv pp))
		     (equal (fragment-objtype pp) 'variable)
		     (equal (fragment-PREP pp) 'of))
		(add-number num (fragment-NP pp)))) )
  (fragment-number argfrag)
  )

; decide-pronoun decidew what pronoun should be used when an NP has been
; prviously referred to.

(defun decide-pronoun (argfrag samefrag &aux h hnp
			       (name (fragment-name argfrag))
			       (subj (fragment-subject argfrag))
			       pronoun pro)
  (unless (not (member (fragment-fragtype argfrag) '(NP S)))
  (setq pronoun
	;; personal
	(cond ((and (fragment-proper-noun samefrag)
		    (setq pro (get-pronoun (fragment-name samefrag) subj)))
	       pro)

	      ; if plural, use they or them
	      ((or (equal (fragment-number samefrag) 'plural)
		   (check-noun-property (fragment-noun samefrag) 'group)
		   (check-noun-property (fragment-name samefrag) 'group)
		   )
	       (setp argfrag 'number 'plural)
	       (if (and (setq h (fragment-head argfrag))
			(equal (fragment-fragtype h) 'PP)
			(setq hnp (fragment-head h))
			(equal (fragment-fragtype hnp) 'NP)
			(null (fragment-prep h)))
		   (setp hnp 'number 'plural))
	       (if subj "they" "them"))

	      ((and (epi-variable samefrag)
		    (check-noun-property (fragment-noun samefrag) 'member))
	       nil)

	      ; default mechanisms - if noun person, man or boy, use he, him
	      ((or (check-noun-property (fragment-noun samefrag) 'masculine)
		   (check-noun-property (fragment-name samefrag) 'masculine))
	       (if subj "he" "him"))
	       
	      ; if noun girl or woman, use she, her
	      ((or (check-noun-property (fragment-noun samefrag) 'feminine)
		   (check-noun-property (fragment-name samefrag) 'feminine))
	       (if subj "she" "her"))
	       
	      ; otherwise use it
	      (t (unless (and (equal (fragment-fragtype argfrag) 'S)
			      (equal (fragment-fragtype (fragment-head argfrag))
				     'ADVP))
			 "it"))
	      )
	 ) 

  ;decide if reflexive should be used - if subject and object same,
  ; then use reflexive
  (if (and (not subj) (fragment-head argfrag)
	   (or (equal name (fragment-name (fragment-head argfrag)))
	       ; pp with this as object, with verb be, subject same
	       (check-self-pp (fragment-name argfrag) argfrag)
	       ))
      (if (equal (fragment-number samefrag) 'plural)
	  (setq pronoun (squish (list pronoun "selves")))
	  (setq pronoun (squish (list pronoun "self")))))
	  
  pronoun))

(defun check-self-pp (name argfrag) 
  (declare (ignore name argfrag))
  nil)

; default-noun determines what noun should be associated with an entity based
; on its sort

(defun default-noun (sort)
  (case sort
	(episode 'event)
	(propos 'thing)
	(otherwise (if sort sort 'thing))
	) )

(defun number-string (nstr)
  (if (null nstr) (return-from number-string nil))
  (do ((n 0 (+ n 1)))
      ((= n (length nstr)))
      (if (not (digit-char-p (char nstr n)))
	  (return-from number-string nil))
      )
  t)

(defun copy-to-new-NP (frag &aux np vp newvp)
  (setq np (fragment-NP frag))
  (setq vp (fragment-VP frag))
  (setq newvp (make-frag 'VP (cons 'verb (fragment-verb vp))
                              (cons 'number (fragment-number vp))
			      (cons 'aspect 'prespart)
			      (cons 'particles (fragment-particles vp))
			      (cons 'ADVP (fragment-ADVP vp))
			      (cons 'objects (fragment-objects vp))))
  (push newvp *nomed-verbs*)
  ; create an identical S with same info, but not the same pointers
  (make-frag 'NP (cons 'name (fragment-name frag))
	     (cons 'S (make-frag 'S (cons 'name (fragment-name frag))
		 (cons 'NP (make-frag 'NP 
			      (cons 'name (fragment-name np))
			      (cons 'number (fragment-number np))
			      ))
                 (cons 'VP newvp))))
  )

;if fragment is member of NPcon, remove member of part and just say NPconj

(defun remove-member-of (argfrag &aux ppp pp np
				 (hnp (fragment-head argfrag)))
  (when (setq ppp (fragment-PP argfrag))
	(if (and (setq np (fragment-NP ppp)) (equal (fragment-name np) 
						    (fragment-name argfrag)))
	    (setp np 'name nil))
	(dolist (f (fragment-parts ppp))
		(if (and (setq np (fragment-NP f)) (equal (fragment-name np) 
						    (fragment-name argfrag)))
		    (setp np 'name nil))))
  (when (and (setq ppp (fragment-PP argfrag))
	     (setq pp (cond ((car (fragment-parts ppp))) (t ppp)))
	     (equal (fragment-PREP pp) 'of)
	     (setq np (fragment-NP pp))
	     (or (and (check-noun-property (fragment-noun argfrag) 'member)
		      (or (fragment-parts np)
			  (epi-variable (fragment-name np))
			  (quoted-expression-p (fragment-name np))
			  (quasi-quoted-expression-p (fragment-name np))
			  (member (fragment-pronoun np) '("them"))) )
		 ; remove group of if the phrase is
		 ;   full of a group of ...   or  group of # ...
		 (and (check-noun-property (fragment-noun argfrag) 'group) 
		      (or (and hnp 
			       (equal (fragment-fragtype hnp) 'PP)
			       (equal (fragment-prep hnp) 'of)
			       (equal (fragment-pre hnp) 'full))
			  (and (if hnp 
				   (and (not (equal (fragment-verb hnp) 'be))
					(not (fragment-subject hnp))
					(not (equal (fragment-prep hnp) 'of)))
				   t)
			       (number-string (fragment-determiner np))))))
	     )
	   ; set subject indicator
	   (unless (getp argfrag 'save-slist)
		   (setp argfrag 'save-slist 
			 (list (list 'noun (fragment-noun argfrag))
			       (list 'number (fragment-number argfrag))
			       (list 'determiner (fragment-determiner argfrag))
			       (list 'subject (fragment-subject argfrag))))
		   (pushnew argfrag *mod-frags*))
	   (unless (getp pp 'save-slist)
		   (setp pp 'save-slist 
			 (list (list 'prep (fragment-prep pp))))
		   (pushnew pp *mod-frags*))
	   (if (and (fragment-head pp) (fragment-subject (fragment-head pp)))
	       (set-subject (fragment-subject argfrag) np))
	   (setp argfrag 'noun nil)
	   (setp argfrag 'determiner "")
	   (setp argfrag 'number 'plural)
	   (setp pp 'prep nil)
	   (setp argfrag 'number (fragment-number np))
	   )
  )

; fill-variable fills in gaps on a variable NP. If there is no noun, one is
; added if necessary, and the number and determiner are also added.

(defun fill-variable (argfrag &aux
			    (req (fragment-REQUIRED argfrag))
			    )
  ; if there isn't a noun on it, and there should be (required, subject)
  ; add a noun - thing for entities with no set, the sort for others
  (when (and (or req (fragment-subject argfrag))
	     (not (noun-info argfrag))
	     (if (fragment-PP argfrag) (fragment-subject argfrag) t)
	     (if (fragment-subject argfrag) t 
		 (and (null (fragment-AP argfrag)))))
	(add-noun (default-noun (find-sort (fragment-name argfrag))) argfrag))

  argfrag)

(defun head-somewhere (head item)
  (cond ((null item) nil)
	((equal head item) t)
	(t (head-somewhere head (fragment-head item)))))

(defun copy-info (info argfrag)
  (if (or (equal info argfrag) (head-somewhere info argfrag)) 
      (return-from copy-info nil))
  (cond ((fragment-S info) (unless (fragment-S argfrag)
				   (setp argfrag 'S (fragment-S info))))
	(t (unless (fragment-quantifier argfrag)
		   (setp argfrag 'quantifier (fragment-quantifier info)))
	   (setp argfrag 'noun (fragment-noun info))
	   (add-number (fragment-number info) argfrag)
	   (setp argfrag 'proper-noun (fragment-proper-noun info))
	   (unless (fragment-AP argfrag) (setp argfrag 'AP (fragment-AP info)))
	   (unless (fragment-PP argfrag) (setp argfrag 'PP (fragment-PP info)))
	   (unless (fragment-Relc argfrag)
		   (setp (fragment-relc info) 'head argfrag)
		   (setp argfrag 'RelC (fragment-RelC info)))
	   ))
  (setp argfrag 'print (fragment-print info))
  (setp argfrag 'parts (fragment-parts info))
  ;(dolist (p (fragment-parts argfrag)) (setp p 'head argfrag))
  (setp argfrag 'conj (fragment-conj info))
  (setp argfrag 'pre (fragment-pre info))
  (setp argfrag 'NP (fragment-NP info))
  ;(setp (fragment-NP argfrag) 'head argfrag)
  (setp argfrag 'VP (fragment-VP info))
  ;(setp (fragment-VP argfrag) 'head argfrag)
  (setp argfrag 'tense (fragment-tense info))
  (setp argfrag 'copy t)
  (setp info 'copy t)
  argfrag)

(defun has-parent-name (name item &aux
			     (head (fragment-head item)))
  (cond ((null head) nil)
	((equal (fragment-name head) name) t)
	(t (has-parent-name name head)))
  )

; fill-gap-NP tries to fill any gaps in the NP fragment argfrag.
; This includes filling in any missing determiners, or NP descriptions if
; nothing is there.

(defun fill-gap-NP (argfrag &aux 
			     (objtype (fragment-OBJTYPE argfrag))
			     tt info h vp)
  (when (and (setq h (fragment-head argfrag))
	     (or (and (equal (fragment-fragtype h) 'ADVP)
		      (has-parent-name (fragment-name argfrag) argfrag))
		 (and (equal (fragment-fragtype h) 'S)
		      (fragment-subject argfrag)
		      (setq vp (fragment-VP h))
		      (or (equal (fragment-aspect vp) 'prespart)
			  (equal (fragment-aspect h) 'prespart))
		      (equal (fragment-objtype argfrag) 'variable)
		      (not (noun-info argfrag)))))
	(setp argfrag 'required nil) 
	(return-from fill-gap-NP argfrag))

  (when (and (not (or (noun-info argfrag)
		      (unless (or (fragment-subject argfrag)
				  (not (equal (find-verb (fragment-head argfrag)) 'be)))
			      (or (fragment-AP argfrag) (fragment-PP argfrag)))
		      ))
	     (not (getp argfrag 'copy))
	     (setq info (find-entity argfrag))
	     (equal (fragment-fragtype info) (fragment-fragtype argfrag))
	     (or (epi-variable (fragment-name argfrag))
		 (skolem-constant-p (fragment-name argfrag))
		 (fragment-subject info)))
	(copy-info info argfrag)
	(manip-frag argfrag))

  (unless (and (fragment-S argfrag) 
	       (if (fragment-NP argfrag)
		   (fragment-VP (fragment-S argfrag))
		   t))

  ; fill in according to object type
  (case objtype
	(variable (fill-variable argfrag))
	(function (unless (member (fragment-name argfrag) *functions*)
			          (setq argfrag (build-function argfrag))))
	(otherwise 
	 (if (or (null (setq tt 
		  (getp (fragment-name argfrag) 'term-type) ))
		 (equal tt 'constant))
	     (setq argfrag (build-NP argfrag))))
	))
  (remove-member-of argfrag)
  (when (not (equal (fragment-name argfrag) 'removed))
	(pushnew argfrag *current-sentence*))
  (fill-sub-NPs argfrag)

  argfrag)

(defun handle-poss (argfrag pp &aux np
			    (noun (fragment-noun argfrag)))
  (when (and (equal (fragment-prep pp) 'of)
	     (null (fragment-adv pp))
	     (null (fragment-pre pp))
	     (setq np (fragment-NP pp))
	     (not (check-noun-property noun 'group))
	     (not (check-noun-property noun 'member))
	     (not (equal noun 'part))
	     (not (member noun '(start end instance duration cardinality)))
	     (null (fragment-PP np))
	     (null (fragment-relc np)))
	(make-possessive argfrag np)
	)
  )

; fill-sub-NPs fills in gaps in any subparts of the NO argfrag including
; S's, PP's, NPor's, and NPand's.

(defun fill-sub-NPs (argfrag)
  ; fill in subparts of NP if there are any
  (if (and (null (fragment-number argfrag))
	   (fragment-parts argfrag))
      (if (equal (fragment-conj argfrag) 'and)
	  (setp argfrag 'number 'plural)
	  (setp argfrag 'number (fragment-number (car (fragment-parts argfrag)))
		)))
  argfrag)

; make-possessive changes an NP like noun1 of noun2 to noun2's noun1

(defun make-possessive (argfrag np &aux res ns nn pro ap-n)
  (decide-number np)
  (decide-determiner np)
  (setq res 
	; noun1 of name => name's noun1
  (cond ((fragment-proper-noun np)
	 (setp argfrag 'determiner 
		(squish (list (fragment-proper-noun np)
			      (if (ends-in-s (fragment-proper-noun np))
				  "'" "'s"))))
	 t)
	; noun1 of noun2 => noun2's name or noun2s' name (sing, plural)
	((fragment-noun np)
	 ; det noun's
	 (setq ns (determine-noun (fragment-noun np) (fragment-number np)))
	 (setq nn (squish (list ns (if (ends-in-s ns) "'" "'s"))))
	 (setq ap-n (squish (append (if (fragment-AP np)
				 (append (prepare-print (fragment-AP np))
						 (list " ")))
				    (list nn))))
	 (setp argfrag 'determiner 
	       (squish (append (if (and (fragment-determiner np)
				(not (equal (fragment-determiner np) "")))
				   (append (prepare-lex (fragment-determiner np) )
				   (if (and (equal (fragment-determiner np) "a")
						    (vowel ap-n))
					       (list "n ")
					       (list " "))))
			       (list ap-n))))
	 t)
	; noun1 of pro => poss pro noun1
	((setq pro (fragment-pronoun np))
	 (cond ((setq nn 
		      (cond ((or (string-equal pro "they")
				 (string-equal pro "them"))
			     "their")
			    ((or (string-equal pro "he")
				 (string-equal pro "him"))
			     "his")
			    ((or (string-equal pro "she")
				 (string-equal pro "her"))
			     "her")
			    ((string-equal pro "it")
			     "its")))
		(setp argfrag 'determiner nn)
		t)))
	))
  res)



; same-adjectives returns t if f1 and f2 have identical adjective phrases

(defun same-adjectives (f1 f2 &aux
			   (a1 (fragment-AP f1))
			   (a2 (fragment-AP f2)))
  (cond ((null a1)
	 (null a2) )
	((null a2)
	 nil)
	(t (same-AP a1 a2))
	)
  )

; same-adjectives returns t if f1 and f2 are identical adjective phrases

(defun same-AP (f1 f2)
  (and (equal (fragment-adv f1) (fragment-adv f2))
       (equal (fragment-adj f1) (fragment-adj f2))
       (null (fragment-parts f1))
       (null (fragment-parts f2)))
  )

; same-pro returns t if a and b would be ambiguous pronouns if they referred
; to different objects

(defun same-pro (a b)
  (cond ((null a) nil)
	((equal a b) t)
	((and (member a '("they" "them") :test 'equal)
	      (member b '("they" "them") :test 'equal))
	 t)
	((and (member a '("he" "him") :test 'equal)
	      (member b '("he" "him") :test 'equal))
	 t)
	((and (member a '("she" "her") :test 'equal)
	      (member b '("she" "her") :test 'equal))
	 t)
	(t nil))
  )

; this looks through the mentioned np list - if nothing is ambiguous,
; can use pronoun, othewise have to devise a short noun phrase to prevent
; ambiguity
(defun ambiguous-pronoun (argfrag pro &optional (s argfrag))
	; if there is another NP with same noun and AP info, then their
	; pronouns will also be the same, and so ambiguous
  (cond ((member (fragment-determiner s) 
		 (list "other" "another" "yet another" "even more") 
		 :test 'equal)
	 t)

	; propositional arguments are a special case - let them use pronouns
	; regardless
	((equal (find-sort (fragment-name argfrag)) 'propos)
	 nil)

	; if the original usage of this item was as a subject, in this current
	; sentence (not previous ones), and there are no more than 2 entities,
	; use pronoun anyway
	((and (fragment-subject argfrag)  
	      (member argfrag *current-sentence* :test 'same-name)
	      (or (<= (list-length (remove-if-not #'(lambda (a) 
		  (equal (fragment-fragtype a) 'NP)) *current-sentence*)) 2)
		  (not (member-if #'(lambda (a) 
			    (and (not (same-name a argfrag)) 
				 (fragment-subject a))) *current-sentence*))
		    ))
	 nil)

	; check all other NPs in *entity-info* - if this ones pronoun would
	; be the same as any of those - ambiguous
	((and (> (list-length *said*) 1) 
	      (not (and (member (fragment-noun argfrag) '(thing))
			)))
	      (member argfrag *said* :test 
		      #'(lambda (a b)
			(if (not (equal (fragment-fragtype b) 'S))
			    (if (not (same-name a b))
				; same pronoun but different np names
				(same-pro (decide-pronoun b b) pro)
				))))
	      )
	)
  )

; ambiguous-np goes through the list of already mentioned nps to see if this
; one is the same as any of them - if so, it adds determiner "other" "another"
; etc. It can handle up to three nps in the same sentence which have the
; same noun-AP descriptions, although it gets a bit confusing.

(defun ambiguous-np (argfrag &aux done h h2)
  (dolist (f *said*) 
	  (when (and (not (same-name argfrag f))
		     (equal (fragment-objtype f) (fragment-objtype argfrag))
		   (fragment-noun argfrag) (fragment-noun f)
		   (equal (fragment-noun argfrag) (fragment-noun f))
		   (not (and (setq h (fragment-head f))
			     (equal (fragment-fragtype h) 'PP)
			     (setq h2 (fragment-head h))
			     (equal (fragment-fragtype h2) 'NP)
			     (or (check-noun-property h2 'member)
				 (check-noun-property h2 'group))))
		   (same-adjectives argfrag f)
		   )
		(if (equal (fragment-number argfrag) 'plural)
		    (if (equal (fragment-determiner f) "other")
			(setp argfrag 'determiner "even more")
			(setp argfrag 'determiner "other"))
		    (if (equal (fragment-determiner f) "another")
			(setp argfrag 'determiner "yet another")
			(setp argfrag 'determiner "another")))
		(setq done t))
	  (when done (return))
	  )
  done)



; fill-gap-PP fills any gaps in a PP fragment. If the PP fragment is required,
; then the NP frag under it is required, and vice versa.
; The NP beneath is filled.

(defun fill-gap-PP (argfrag &aux  
			    (np (fragment-NP argfrag)))
  (when np
  ; check required
  (if (fragment-required argfrag)
      (setp np 'required t)
      (if (fragment-required np)
	  (setp argfrag 'required t)))

  ; if NP has determiner only, it got it from () => (), and was intended
  ; for things like "only people ..." or "People only eat ..", and doesnt
  ; belong on a  PP so erase it.
  (if (equal (fragment-DETERMINER np) "only")
      (setp np 'DETERMINER nil))
	)
  argfrag)

(defun fill-gap-ADVP (argfrag &aux 
			    (s (cond ((fragment-S argfrag))
				     ((fragment-NP argfrag)))))
  (when s
  ; check required
  (if (fragment-required argfrag)
      (setp s 'required t)
      (if (fragment-required s)
	  (setp argfrag 'required t)))

  ; if NP should be filled, fill it, otherwise erase prep so that this PP
  ; doesn't print
  (if (or (fragment-required argfrag)
	  (fragment-required s)
	  (noun-info s)
	  (fragment-NP s)
	  (fragment-VP s)
	  (fragment-ADVP s)
	  (fragment-parts s))
      nil
      (setp argfrag 'prep nil)
      ))
  argfrag)


; agree-number ensures that the subject and object of a be verb agree in number
; Also, if number is on S frag, but not on NP frag, it is moved

(defun agree-number (npfrag vpfrag &aux n
			  (obj (if vpfrag (car (fragment-objects vpfrag))))
			  (nobj (if obj
				    (if (equal (fragment-fragtype obj) 'NP)
					obj))))
  ; make sure objects of verb agree in number with subject (for be)
  (when (and nobj npfrag (equal (fragment-verb vpfrag) 'be)
	     (null (fragment-aspect vpfrag)))
	    (add-number (fragment-NUMBER npfrag) nobj)
	)
  (when (and npfrag vpfrag)
        (setp vpfrag 'NUMBER (fragment-NUMBER npfrag)))
  (when (and npfrag (fragment-relc npfrag)
	     (not (equal vpfrag (fragment-relc npfrag))))
	(agree-number npfrag (fragment-relc npfrag)))
  (if (and (fragment-parts vpfrag) (setq n (fragment-number vpfrag)))
      (dolist (v (fragment-parts vpfrag))
	      (setp v 'NUMBER n)))
  )

(defun get-first-object (s &aux o)
  (cond ((fragment-VP s) 
	 (if (setq o (car (fragment-objects (fragment-VP s))))
	     (if (equal (fragment-fragtype o) 'NP)
		 (fragment-name o))))
	(t nil))
  )

(defun get-subject (s)
  (cond ((fragment-NP s) (fragment-name (fragment-NP s)))
	(t nil))
  )

(defun do-passive (s)
  (cond ((fragment-parts s) nil)
	(t (passivize s))
	))

; try-passive tests to see if sentences a and b can be passivized so that 
; they have the same subject
(defun try-passive (a b &aux
		      (subj1 (get-subject a))
		      (obj1 (get-first-object a))
		      (subj2 (get-subject b))
		      (obj2 (get-first-object b))
		      newa newb
		      )
  (when (and subj1 subj2 (not (equal subj1 subj2)))
	(if (and obj2 (equal subj1 obj2)
		 (setq newb (do-passive b)))
	    (list a newb)
	    (if (and obj1 (equal subj2 obj1)
		     (setq newa (do-passive a)))
		(list newa b)
		))
	)
  )


(defun check-Sconj (slist &aux
			  first newlist subj2 obj2 new
			  (subj1 (get-subject (car slist)))
			  (obj1 (get-first-object (car slist)))
			  )
  (setq *conj* t)
  ; check to see if any should be passivized, then fill them
  (if subj1
      (let ()
	(dolist (s (cdr slist))
		(if (and (setq subj2 (get-subject s))
			 (not (equal subj1 subj2))
			 (setq obj2 (get-first-object s))
			 (equal subj1 obj2)
			 (setq new (do-passive s)))
		    (setq newlist (append newlist (list new)))
		    (setq newlist (append newlist (list s)))
		    )
		; this technique will be a problem if there are many
		; combined - really only works well for 2
		(if (and (equal (list-length slist) 2)
			 (null new) subj2 obj1 (equal subj2 obj1)
			 (setq new (do-passive (car slist))))
		    (setq first new)
		    ))
	   (if (null first) (setq first (car slist)))
	   (setq newlist (append (list first) newlist))
	   )
      (setq newlist slist)
      )
  newlist
  )



; move-neg-to-S is used when a verb removed has a negation on it.
; If the NP in front has an S beneath (and it should, because otherwise we
; wouldn't have a verb removed), the negation is moved to its verb.
; If it already had a negation, the negation is changed to adverb only

(defun move-neg-to-S (sfrag &aux
			    (vpfrag (fragment-VP sfrag))
			    )
  (when vpfrag
	(if (fragment-negated vpfrag)
	    (let ()
		 (setp vpfrag 'negated nil)
		 (setp vpfrag 'adv (combine-l 'only 
					      (fragment-adv vpfrag)))
		 )
	    (setp vpfrag 'negated t)
	    )
	)
  )

; set-subject sets the subject indicator on a fragment, and propagates that
; to all NPand and NPor fragments beneath it

(defun set-subject (subj frag)
  (setp frag 'subject subj)
  (dolist (f (fragment-parts frag)) (set-subject subj f))
  )

(defun fill-gap-VP (v) v)

; switch-or-and switches NPor and NPand fragments. If a fragment is negated,
; or's becomes and's and vice versa
(defun switch-or-and (frag)
  (case (fragment-conj frag)
	(or (setp frag 'conj 'and)
	    (if (equal (fragment-pre frag) 'either)
		(setp frag 'pre nil))) ;both
	(and (setp frag 'conj 'or)
	     (if (equal (fragment-pre frag) 'both)
		 (setp frag 'pre 'either)))
	)
  )

; move-negation checks to see if it can move a negation to a mor natural place

(defun move-negation (frag &aux
			  (vp (fragment-VP frag))
			  (np (fragment-NP frag))
			  h obj useobj)
  (when (and (fragment-negated frag) vp)
	(if (fragment-negated vp)
	    (let ()
		 (setp vp 'adv 'only)
		 (setp vp 'negated nil))
	    (setp vp 'negated t))
	(setp frag 'negated nil)
	)
  (when (and vp np)
	(when (fragment-negated vp) 
	      ; we have a negated verb
	      (cond ((and (equal (fragment-objtype np) 'variable)
			  (not (and (setq h (fragment-head frag))
				    (equal (fragment-fragtype h) 'S)
				    (fragment-parts h)
				    (not (member (fragment-conj h) '(and or)))))
			  (null (fragment-determiner np))
			  (member (fragment-quantifier np) '(A E))
			  ;(not (member (fragment-quantifier np) 
				       ;'(none few many most some)))
			  )
		     ; if we have a variable subject 
		     ; negate it and handle or's and and's beneath it
		     (setp vp 'negated nil)
		     (switch-or-and np)
		     (if (fragment-negated np)
			 (let () ;; CHANGE
			      (setp np 'negated nil)
			      (add-number 'plural np)
			      (add-determiner "only" np) )
			 (setp np 'negated t))
		     )

		    ((and (setq obj (car (fragment-OBJECTS vp)))
			  (equal (fragment-objtype obj) 'variable)
			  (member (fragment-fragtype obj) '(NP ))) ;PP
		     ; otherwise, if the first object is a variable, move
		     ; the negation there, and handle or's and and's
		     (if (equal (fragment-fragtype obj) 'PP)
			 (setq useobj (fragment-NP obj))
			 (setq useobj obj))
		     (when (or (fragment-required useobj)
			       (noun-info useobj))
		           (switch-or-and useobj)
			   (setp vp 'negated nil)
		           (setp useobj 'negated t))
		     )
		    )
	      )
	)
  frag)



; noun-info returns t if there is important infomation on the NP fragment -
; this information indicates tht no lookup needs to be done on the entity
(defun noun-info (nfrag &optional morethantrans)
  (or (fragment-NOUN nfrag)
      (fragment-PRONOUN nfrag)
      (if morethantrans nil (fragment-proper-noun nfrag))
      (fragment-parts nfrag)
      (if morethantrans nil (fragment-print nfrag))
      (fragment-S nfrag)
      )
  )
