;       Last change    August 1993
(in-package response)

(defvar *fragment-types* '(AP PP RELC ADVP ADVS S NP VP)
  "Indicates legal fragment types")
(defvar *fragment-groups* '(parts objects)
  "Indicates which fields of a fragment indicate a group of fragments")
(defvar *multiple-fields* '(adv pre adj aspect)
  "Indicates which fields of a fragment may have multiple values - a list")
(defvar *fragment-number* 0)
(defvar *fragment-list* nil)
(defvar *entity-list* nil)
(defvar *entity-info* nil)

(defun default-fragtype (item &aux
                              (s (find-sort item)))
  (if (equal s 'episode) 'S 'NP)
  )

(defun make-frag (fragtype &rest properties &aux fnew)
  (setq fnew (build-symbol (list 'frag (incf *fragment-number*)) 'epilog))
  (setf (symbol-plist fnew) nil)
  (setp fnew 'frag-type fragtype)
  (setp fnew 'required t)
  (dolist (p properties) 
	  (setp fnew (car p) (cdr p))
	  (if (listp (cdr p))
	      (dolist (i (cdr p))
		      (if (and (symbolp i) (getp i 'frag-type))
			  (setp i 'head fnew)))
	      (if (and (symbolp p) (getp p 'frag-type)) (setp p 'head fnew)))
	  )
  (if (and (fragment-name fnew) 
	   (not (equal (fragment-name fnew) 'removed)))
      (pushnew (fragment-name fnew) *entity-list*))
  (set-objtype fnew)
  (if (equal fragtype 'S) (set-subject t (fragment-NP fnew)))
  fnew)

(defun highest-type (tlist)
  (cond ((member 'S tlist) 'S)
	((not (member-if #'(lambda (a) (not (member a '(NP PP AP)))) tlist)) 
	 'NP)
	((not (member-if #'(lambda (a) (not (member a '(ADVP ADVS VP)))) tlist))
	      'S)
	(t 'S)))

(defun combine-objects (conj objlist head)
  (make-conj nil conj (mapcar 'car objlist) nil head t)
  )

(defun similar-objects (list &aux 
			     (olist (mapcar #'(lambda (a) (fragment-objects a))
					    list))
			      (item (car olist))
			      (type (fragment-fragtype (car item))))
  (if (> (list-length item) 1) (return-from similar-objects nil))
  (dolist (o (cdr olist))
	  (if (or (> (list-length o) 1)
		  (not (equal (fragment-fragtype (car o)) type)))
	      (return-from similar-objects nil)))
  t)

(defun all-same (rtn list &aux item)
  (setq item (eval (list rtn (append (list 'quote) (list (car list))))))
  (dolist (l (cdr list))
	  (if (not (equal (eval (list rtn (append (list 'quote) (list l)))) item)) 
	      (return-from all-same nil)))
  t)

(defun make-conj (pre conj parts &optional samename head fromobject &aux type name p1 p2 neg objects
		      (heads (remove-if 'null (mapcar #'(lambda (a) (fragment-head a)) 
				     parts)))
		      (types (mapcar #'(lambda (a) (fragment-fragtype a)) 
				     parts))
		      (names (mapcar #'(lambda (a) (fragment-name a)) parts)))
  (unless head (setq head (car heads)))
  (setq types (remove-duplicates types))
  (setq type (cond ((and types (= (list-length types) 1))
		    (car types))
		   (t (highest-type types))))
  (setq names (remove-duplicates names))
  (setq name (cond (samename)
		   ((and names (= (list-length names) 1))
		    (car names))
		   (t nil)))
  (unless (or name fromobject) (setq type 'S))
  (setq p1 (car parts))
  (setq p2 (cadr parts))
  (if (and (equal pre 'if) (fragment-parts p1) (equal (fragment-pre p1) 'if))
      (setp p1 'pre 'when))
  (if (equal type 'S)
      (setq parts (mapcar #'(lambda (a) (coerce-to 'S a)) parts)))
  (if (equal type 'ADVP)  ; order so that temporal relations are last
      (setq parts (sort parts #'(lambda (a b) 
		(declare (ignore b))
		(not (member (fragment-prep a) '(while because 
		       before after overlaps overlapped-by when as ) ) ) ))))

  (if (> (list-length types) 1)
      (setq parts (mapcar #'(lambda (a) (coerce-to type a)) parts)) )

  ; check to see if one or both are already conjunctions
  (cond ((or (null conj) (and (null name) (null fromobject)))
	 (make-frag type (cons 'pre pre) (cons 'conj conj)
						  (cons 'name name) 
						  (cons 'head head) 
						  (cons 'parts parts)))
	((and (null (fragment-parts p1)) (all-same 'fragment-parts parts))
	 (cond ((and (equal type 'VP)
		     (all-same 'fragment-negated parts)
		     (all-same 'fragment-adv parts)
		     (all-same 'fragment-aspect parts)
		     (all-same 'fragment-tense parts)
		     (all-same 'fragment-mood parts)
		     (all-same 'fragment-verb parts)
		     (similar-objects parts)
		     )
		(setq neg (fragment-negated (car parts)))
		(dolist (p parts) (setp p 'negated nil))
		(when neg (if (equal conj 'and) 
			      (setq conj 'or) (setq conj 'and)))
		(setq objects (combine-objects conj 
                      (mapcar #'(lambda (a) (fragment-objects a)) parts)
		      (car parts) ))
		(cond (objects 
		       (setp (car parts) 'objects (list objects))
		       (car parts))
		      (t (make-frag type (cons 'pre pre) 
						  (cons 'conj conj)
						  (cons 'name name) 
						  (cons 'head head) 
						  (cons 'parts parts))))
		)
	       ((all-same 'fragment-negated parts)
		(setq neg (fragment-negated p1))
		(dolist (p parts) (setp p 'negated nil))
		(when neg 
		      (if (equal conj 'and) (setq conj 'or) (setq conj 'and)))
		(make-frag type (cons 'pre pre) (cons 'conj conj)
			   (cons 'name name) (append (list 'parts) parts)
			   (cons 'head head) (cons 'negated neg)) )
	       ((> (list-length parts) 2)
		(setq parts (sort parts #'(lambda (a b) 
					          (declare (ignore b))
						  (not (fragment-negated a)))))
		(make-frag type (cons 'pre pre) (cons 'conj conj)
						  (cons 'name name) 
						  (cons 'head head) 
						  (cons 'parts parts)))
	       ((and (fragment-negated p1) (not (fragment-negated p2)))
		(if (equal conj 'and) (setq conj 'but))
		(make-frag type (cons 'pre pre) (cons 'conj conj)
			   (cons 'head head) (cons 'name name) (list 'parts p2 p1)))
	       (t (if (equal conj 'and) (setq conj 'but))
		(make-frag type (cons 'pre pre) (cons 'conj conj)
			   (cons 'head head) (cons 'name name) (list 'parts p1 p2)))
	       ))
	((> (list-length parts) 2)
	 (make-frag type (cons 'pre pre) (cons 'conj conj)
					   (cons 'name name) 
					   (cons 'head head) 
					   (cons 'parts parts)))
        ((null (fragment-parts p1))
	 ; p2 must have be a conjunction
	 (add-to-conj name type pre conj p1 p2))
	((null (fragment-parts p2))
	 (add-to-conj name type pre conj p2 p1))
	(t (merge-conj name type pre conj p1 p2 head) ) ; both are conjunctions
	)
  )

(defun add-to-conj (name type pre conj item cfrag &aux parts)
  ; if negation value same, just add item to parts of cfrag
  (cond ((or (not (equal (fragment-pre cfrag) pre))
	     (not (case (fragment-conj cfrag)
			((and but) (equal conj 'and))
			(or (equal conj 'or)))))
	 (make-frag type (cons 'pre pre) (cons 'conj conj)
		    (cons 'name name) (list 'parts item cfrag)))
	((equal (fragment-negated item) (fragment-negated cfrag))
	 (setp item 'negated nil)
	 (setp item 'head cfrag)
	 (setp cfrag 'parts (append (fragment-parts cfrag) (list item)))
	 cfrag)
	((fragment-negated cfrag)
	 (make-frag type (cons 'pre pre) (cons 'conj conj)
		    (cons 'name name) (cons 'parts parts)))
	(t ; item is negated, cfrag isn't, but cfrag might have
	   ; a single negated item in its parts
	   (let ((ps (fragment-parts cfrag))
		 negs pos p1)
		(dolist (p ps)
			(if (fragment-negated p)
			    (setq negs (append negs (list p)))
			    (setq pos (append pos (list p)))))
		(cond (negs
		       (setq p1 (make-conj pre (if (equal conj 'and) 'or 'and)
					   (append negs (list item))))
		       ; should only be 1 pos heres
		       (if (equal conj 'and) (setq conj 'but))
		       (make-frag type (cons 'pre pre) (cons 'conj conj)
				  (cons 'name name) (list 'parts (car pos) p1)))
		      (t (make-frag type (cons 'pre pre) (cons 'conj conj)
				  (cons 'name name) (list 'parts cfrag item)))
		      )))
	)
  )

(defun merge-conj (name type pre conj cfrag1 cfrag2 &optional head &aux p1 c1 ofrag
			mfrag nfrag pieces neg-pieces pos-pieces )
  ; if two conjunctions - if identical, and equal to pre,conj given, combine
  ; parts
  (cond ((or (not (equal (fragment-pre cfrag1) pre))
	      (not (case (fragment-conj cfrag1)
			 ((and but) (equal conj 'and))
			 (or (equal conj 'or)))))
	 (add-to-conj name type pre conj cfrag1 cfrag2))
	((or (not (equal (fragment-pre cfrag2) pre))
	      (not (case (fragment-conj cfrag2)
			 ((and but) (equal conj 'and))
			 (or (equal conj 'or)))))
	 (add-to-conj name type pre conj cfrag2 cfrag1))
	((equal (fragment-negated cfrag1) (fragment-negated cfrag2))
	 (setq p1 (fragment-parts cfrag2))
	 (dolist (p p1) (setp p 'head cfrag1))
	 (setp cfrag1 'parts (append (fragment-parts cfrag1) p1))
	 cfrag1)
	; one neg, one pos, take any negs off the + one and put on the neg
	((if (fragment-negated cfrag1)
	     (setq c1 (member-if 'fragment-negated (fragment-parts cfrag2))))
	 (cond (c1 (setq mfrag cfrag1)
		   (setq pieces (fragment-parts cfrag2))
		   (setq ofrag cfrag2)
		   )
	       (t (setq mfrag cfrag2)
		   (setq pieces (fragment-parts cfrag1))
		   (setq ofrag cfrag1)
		   ))
	 (dolist (p pieces)
		 (if (fragment-negated p)
		     (setq neg-pieces (append neg-pieces (list p)))
		     (setq pos-pieces (append pos-pieces (list p))))
		 )
	 (setq ofrag (car pos-pieces))
	 (setq nfrag (add-to-conj name type pre conj (car neg-pieces) mfrag))
	 (if (equal conj 'and) (setq conj 'but))
	 (make-frag type (cons 'pre pre) (cons 'conj conj)
		    (cons 'name name) (list 'parts ofrag nfrag))
	 )
	(t (if (equal conj 'and) (setq conj 'but))
	   (if (and (fragment-negated cfrag1) (not (fragment-negated cfrag2)))
	       (make-frag type (cons 'pre pre) (cons 'conj conj)
		      (cons 'head head) (cons 'name name) (list 'parts cfrag2 cfrag1))
	       (make-frag type (cons 'pre pre) (cons 'conj conj)
		      (cons 'head head) (cons 'name name) (list 'parts cfrag1 cfrag2))) )
	)
  )

(defun fragment (item)
  (and (symbolp item) (getp item 'frag-type)))

(defun initial-fragment (name &aux frag tt
                              (def (default-fragtype name)))
  (unless (or (null name) (equal name 'removed) (not (symbolp name)))
	  (pushnew name *entity-list*))
  (cond ((and (symbolp name) (setq tt (getp name 'term-type)))
         (case tt
               (record
                ; figure out how to say record and use as print property
                (setq frag (make-frag def (cons 'name name))) 
		(setq frag (build-NP frag))
		(update-entity-info name frag)
                )
               ((quasi-quoted-expression quoted-expression)
                (setq frag (make-frag def (cons 'name name))) 
		(if (symbolp name)
		    (setp frag 'print (princ-to-string 
				       (getp name 'print)))
		    (setp frag 'print (princ-to-string name))) 
		(update-entity-info name frag)
                )
               (otherwise ; modified handled elsewhere
                (setq frag (make-frag def (cons 'name name))))
               ))
	((numberp name)
	 (setq frag (make-frag 'NP (cons 'print (princ-to-string name)))))
	((stringp name)
	 (setq frag (make-frag 'NP (cons 'print name))))
        (t (setq frag (make-frag def (cons 'name name))))
        )
  frag)

(defun coerce-to (type item)
  (cond ((not (fragment item))
         (pushnew item *entity-list*)
         (make-frag type (cons 'name item)))
        ((and (not (equal type 'S))
	      (equal (fragment-fragtype item) type) item))
        (t (case type
                 (NP (make-into-NP item))
                 (S (make-into-S item))
                 (VP (make-into-VP item))
                 (otherwise item)))
        )
  )

; some general useful routines

; passivize changes the given fragment into a passive fragment if possible -
; the object becomes the subject, the verb is passivized, and the subject is
; added in a by PP as a verb object.

(defun passivize (p &aux
                    (subj (fragment-NP p))
                    (verbp (fragment-VP p))
                    (verb (fragment-verb verbp))
                    (objs (fragment-objects verbp))
                    (obj (car objs))
                    passvp newv newvp)
  ; cannot passivize if no object, if we have a complex VP with infinitives
  ; (actually we probably can but it is messy), or if we have the verb be
  (when (or (and obj (member (fragment-aspect verbp) '(infinitive prespart)) )
            (equal verb 'be) (null verbp)
            (equal (fragment-aspect verbp) 'passive) ;already passive
            (null obj)
            (not (setq newv (passive verb))))
        (return-from passivize nil))
 
  (if (traced-p 'verbalization)
      (print-line "  Response: Passivizing "
                  (fragment-name subj) " " verb " "
                  (cond ((fragment-particles verbp)) (t " ")) " "
                  (fragment-name obj) "    to " (fragment-name obj) " "
                  newv " " (cond ((fragment-particles verbp)) (t " "))
                  " by " (fragment-name subj)))
  (unless (and (symbolp obj) (getp obj 'frag-type)) 
	  (setq obj (make-into-NP obj)))
  (set-subject t obj)
  (set-subject nil subj)
  (setq passvp
	(make-frag 'VP (cons 'name (fragment-name obj))
				 (cons 'aspect 'passive) (cons 'verb newv)
                       (append (list 'objects)
			       (append (cdr objs) (list (make-frag 'PP
                                         (cons 'prep 'by)
                                         (cons 'NP subj)))))) )
  (setq newvp (make-frag 'VP (cons 'name (fragment-name obj))
                         (cons 'negated (fragment-negated verbp))
                         (cons 'verb 'be)
                         (cons 'aspect (fragment-aspect verbp))
                         (cons 'tense (fragment-tense verbp))
                         (cons 'mood (fragment-mood verbp))
                         (cons 'particles (fragment-particles verbp))
                         (cons 'adv (fragment-adv verbp))
                         (list 'objects passvp)))

  (make-frag 'S (cons 'name (fragment-name p))
             (cons 'head (fragment-head p))
             (cons 'advs (fragment-advs p))
             (cons 'tense (fragment-tense p))
             (cons 'aspect (fragment-aspect p))
             (cons 'mood (fragment-mood p))
             (cons 'negated (fragment-negated p))
             (cons 'advp (fragment-advp p))
             (cons 'NP obj) (cons 'adv (fragment-adv p))
             (cons 'VP newvp)
           ) 
  )

; passive determines whether there is a passive form for a verb
; if so, it returns the part of the new VP (be (Vpass verb)), and updates
; the tense appropriately

(defun passive (verb &aux
                     nv) 
  (setq nv (determine-verb nil verb nil nil 'passive nil))
  (if (equal nv 'no-passive)
      nil
      nv)
  )

; maybe test properties here instead? This really should be improved.
(defun better-noun (n1 n2)
  (cond ((member n1 '(location member thing part one)) nil)
	((member n2 '(location member thing part one)) t)
	(t t)))

; returns t if new is better than old 
(defun better-item (new old &aux n1 n2)
  (cond ((not (equal (fragment-name new) (fragment-name old))) t)
        ((and (equal (fragment-fragtype new) 'S)
	      (not (equal (fragment-fragtype old) 'S))) t)
	((and (equal (fragment-fragtype old) 'S)
	      (not (equal (fragment-fragtype new) 'S))) nil)
	((getp (fragment-name old) 'trans) nil)
	((getp (fragment-name new) 'trans) t)
	((negation-above old) t)
	((negation-above new) nil)
	((and (setq n1 (fragment-noun new)) (setq n2 (fragment-noun old)))
	 (better-noun n1 n2))
	((and (fragment-subject new) (not (fragment-subject old))) t)
	((or (noun-info new) (fragment-vp new)) t)
	((or (noun-info old) (fragment-vp old)) nil)
	(t nil))
  )

(defun negation-above (frag)
  (cond ((null frag) nil)
	((fragment-negated frag) t)
	((equal (fragment-head frag) frag) 
	 (print-line "head equals frag " frag)
	 (setp frag 'head nil)
	 nil)
	(t (negation-above (fragment-head frag))))
  )

(defun copy-s-info (new save)
  ; fix heads too!!!!!
  (unless (or (equal new save) (null (fragment-name save)))
	  (unless (fragment-AP save) 
		  (setp save 'AP (fragment-AP new))
		  (setp (fragment-AP save) 'head save)
		 ; (setp new 'AP nil)
		  )
	  (unless (fragment-PP save) 
		  (setp save 'PP (fragment-PP new))
		  (setp (fragment-PP save) 'head save)
		 ; (setp new 'PP nil)
		  )
	  (unless (fragment-Relc save) 
		  (setp save 'Relc (fragment-Relc new))
		  (setp (fragment-Relc save) 'head save)
		 ; (setp new 'Relc nil)
		  )
	  ))


(defun update-said (name item &aux olditem)
  (unless (or (null name) (equal name 'removed) )
  (setq olditem (car (member item *said* :test 'same-name)))
  (cond (olditem
	 (if (not (or (noun-info item) (fragment-VP item)))
	     nil 
	     (progn 
	      (unless (or (equal olditem item) 
			  (not (better-item item olditem)))
		 (setq *said* (remove olditem *said* :test 'equal))
		 (push item *said*)) )  ) )
	(t (push item *said*))))
  )

(defun update-entity-info (name item &optional oldi &aux old olditem saved)
  (unless (or (null name) (equal name 'removed) )
  (setq old (assoc name *entity-info*))
  (cond (old
	 (if (not (or (noun-info item) (fragment-VP item)))
	     nil 
	     (progn 
	 (setq olditem (cadr old))
	 (unless (or (equal olditem item) 
		     (if oldi (not (equal oldi olditem)))
		     (not (better-item item olditem)))
		 (setq *entity-info* (remove old *entity-info* :test 'equal))
		 (push (list name item) *entity-info*)) )  ) )
	(t (push (list name item) *entity-info*))))
  (unless (or (null name) (equal name 'removed)
	      (epi-variable name) (null item)
	      (null (setq saved (cadr (assoc name *entity-info*))))
	      )
	  (copy-s-info item saved))
  )

(defun find-entity (frag &aux 
			 (name (fragment-name frag)))
  (cadr (assoc name *entity-info*)))

(defun propagate-neg (item head &aux h h2)
  (when head
	(case (fragment-fragtype head)
	      (AP ; if head is AP, so is fragment below it - do not propagate
		  )
	      (PP
	       (unless (equal (fragment-type item) 'PP)
	       (setp head 'negated t)
	       (setp item 'negated nil)
	       (if (equal (fragment-fragtype (fragment-head head)) 'VP)
	       (propagate-neg head (fragment-head head))) ))

	      (NP; must have PP or AP below to get here
		   (when (member (fragment-fragtype item) '(AP)) ;PP
			 (setp item 'negated nil)
			 (setp head 'negated t)
			 (propagate-neg head (fragment-head head))))
	      ((VP RELC)
	       ; place negation on verb - if already negated, remove
	       ; the negation and add adverb only
	       ; gives sentence "People only like to eat people food"
	       ; instead of people do not like to eat not people food"
	       (if (fragment-negated head)
		   (let ()
			(setp head 'adv 
			      (pushnew 'only (fragment-adv head)))
			(setp head 'negated nil)
			(setp item 'negated nil))
		   (let ()
			(setp head 'negated t)
			(setp item 'negated nil)))
	       (if (and (fragment-negated head) 
			(member (getp head 'aspect) 
				'(passive infinitive prespart)))
		   (propagate-neg head (fragment-head head)))
	       (if (and (equal (fragment-fragtype head) 'relc)
			(setq h (fragment-head head))
			(setq h2 (fragment-head h))
			(equal (fragment-fragtype h2) 'VP))
		   (propagate-neg head (fragment-head head)))
	       )
	      )
	)
  )

(defun set-objtype (frag &aux fname)
  (cond ((epi-variable (fragment-name frag))
	 (setp frag 'objtype 'variable))
	((equal (getp (fragment-name frag) 'term-type) 'function)
	 (setp frag 'objtype 'function))
	((equal (getp (fragment-name frag) 'term-type) 'modified)
	 (setp frag 'objtype 'function))
	((and (symbolp (setq fname (fragment-name frag)))
	      (getp fname 'entity-type)
	      (not (equal (getp fname 'term-type) 'constant)))
	 (setp frag 'objtype 'variable))
	(t (setp frag 'objtype 'constant)))
  )
