;       Last change    August 1993
(in-package response)

(tweakable '*default-trans* t 
  "Indicates whether to guess the translation (t), or ask the user (nil)"
  'boolean)

(tweakable '*response-warn* t 
  "Indicates whether to warn the user about missing translation info"
  'boolean)

(tweakable '*active-topics* nil
  "Topics which indicate that a predicate should be said as a verb, not adj"
  'symbol-list)

(tweakable '*fragment-routines* '(make-adverb force-adverb make-passive) 
  "List of routines which may be invoked through translation info"
  'symbol-list)

(defvar *already-frags* nil)
(defvar *already-translated* nil)

(defun make-adverb (&rest args &aux arg adj adv)
  ; should be one argument - an AP - make adverb out of it
  (setq arg (car args))
  (setq adj (fragment-adj arg))
  (setq adv (cond ((getp adj 'adverb))
		  (t (calc-word-form adj 'ly))))
  adv)

(defun force-adverb (&rest args &aux arg adj adv)
  ; should be one argument - an AP - make adverb out of it
  (setq arg (car args))
  (setq adj (cond ((fragment-adj arg))
		  ((fragment-noun arg))
		  ((fragment-verb arg))))
  (setq adv (cond ((getp adj 'adverb))
		  (t (string-downcase (princ-to-string adj)))))
  adv)

(defun make-passive (&rest args &aux 
			   (p (car args)))
  (if (fragment p) (passivize p) p)
  )

(defvar *conditional* nil)
(defvar *antecedent-trans* nil)
(defvar *consequent-trans* nil)
(defvar *number* nil)
(defvar *functions* nil)

(defun wff-fragment (wff &aux frag)
  (setq *conditional* nil)
  (setq *already-frags* nil)
  (setq *already-translated* nil)
  (setq frag (wff-trans wff))
  (when (traced-p 'fragment)
	(print-line "  Response: Fragment for " (getp wff 'print))
	(print-frag frag 2))
  frag)


(defun wff-trans (wff &optional other-ante &aux wres 
		      (wtype (getp wff 'wff-type)))
  (setq wres
    (if (member wff *already-translated*)
	wff
      (case wtype
	(constant wff)
	(variable 
	 (make-frag 'S (cons 'name wff)
		    (cons 'NP (make-frag 'NP (cons 'noun 'thing)
					 (cons 'determiner 'some)))
		    (cons 'VP (make-frag 'VP (cons 'verb 'happen)))))
	(episodic
	 (let* ((args (getp wff 'arguments))
		res tense)
	       (setq res (make-into-S (wff-trans (car args))))
	       (setp res 'name (last1 args))
	       (set-objtype res)
	       (update-entity-info (fragment-name res) res)
	       ; add tense info if applicable (maybe move somewhere else?)
	       (when (not (equal (getp wff 'pred) '_))
		     (setq tense (tense-info (last1 args)))
		     (update-entity-info (last1 args) res)
		     (if tense (setp res 'tense tense)) )
	      (when (equal (getp wff 'pred) '_)
		    (setq res (make-frag 'NP (cons 'name (fragment-name res))
					 (cons 'pre "that")
					  (cons 'S res))))
	       (when (neg-p wff) (setp res 'negated t))
	       (if other-ante
		   (conditional-trans other-ante (list res))
		   res)
	       ))
	(modified
	 (let ((op (getp wff 'operator))
	       (opargs (getp wff 'arguments))
	       res)
	      (setq opargs (mapcar #'(lambda (a) 
					     (item-trans a nil nil)) opargs))
	      (setq res (op-trans op (neg-p wff) opargs))
	      res))
	((prefix causal)
	 (let* ((args (mapcar 'item-trans (getp wff 'arguments)))
		(pred (pred-trans (getp wff 'pred) (neg-p wff) args))
		res)
	   (setq res pred)
	   (if other-ante
	       (conditional-trans other-ante (list res))
	     res)
	   )
	 )
	(quantified
	 (let ((r (getp wff 'restrict))
	       (m (getp wff 'main))
	       (var (getp wff 'variable))
	       (quant (getp wff 'quantifier))
	       res rt extra)
	      (setq extra (quant-info quant var))
	      (if r (setq rt (wff-trans r)))
	      (if r 
		  (setq rt (cond ((car (conjunct-frags rt extra))) (t rt)))
		  (setq rt extra))
		     

	      (setq res
	      (if (getp wff 'conditional)
		  (wff-trans m (append other-ante (if r (list rt))))
		  (if (null rt)
		      (wff-trans m other-ante)
		      (let* ((mt (wff-trans m))
			     (cres (conjunct-frags rt mt))
			     ress)
			   (setq ress 
				 (cond ((car cres))
				       (t (make-conj nil 'and (list rt mt)))))
			   (if other-ante 
			       (conditional-trans other-ante (list ress))
			       ress)
			   ))
		  ) )
	       (when (neg-p wff) (setp res 'negated t)) 
	       res)
	 )
	(logical
	 (let ((args (getp wff 'arguments))
	       (pred (getp wff 'pred))
	       res mres)
	      (if (and (listp pred) (numberp (car pred))
		       (<= (car pred) 1.0))
		  (setq pred (car pred)))
	      (setq res
	      (cond ((getp wff 'conditional)
		     (let ((rt (wff-trans (car args)))
			   (mt (wff-trans (cadr args))))
			  (conditional-trans (list rt) (list mt) 
					     (adverb-for pred)) ))
		    (t (let ((res (apply (if (equal pred 'and) 
					   'conjunct-frags 'disjunct-frags)
					 (setq mres (mapcar 'wff-trans args))))
			     ress)
			    (setq ress 
				 (cond ((null res) 
					(make-conj nil pred mres))
				       ((= (list-length res) 1) (car res))
				       (t (make-conj nil pred res))))
			   (if other-ante 
			       (conditional-trans other-ante (list ress))
			       ress)
			   ))
	      ))
	      (when (neg-p wff) (setp res 'negated t)) 
	      res)
	 )
	(quasi-quoted-expression (wff-trans (getp wff 'quote-wff) other-ante))
	(quoted-expression wff)
	(t wff) ;treat as constant
	))
	)
  (when (traced-p 'trans-details)
      (print-line "  Response generator: Translated wff " 
		  (print-info wff) " to ")
      (if (fragment wres) (print-frag wres) (print-it wres))
      )
  (setq *already-translated* (append *already-translated* (list wff)))
  wres)

(defun item-trans (item &optional neg args fromop)
  (cond ((symbolp item)
	 (if (getp item 'pred-type) 
	     (pred-trans item neg args fromop)
	     (case (getp item 'entity-type)
	       (wff (wff-trans item))
	       (pred (pred-trans item neg args fromop))
	       (term (term-trans item))
	       (op (op-trans item neg args))
	       (otherwise (term-trans item))) ) )
	(t (term-trans item))
	)
  )

(defun lambda-var (item)
  (if (fragment item)
      (getp (fragment-name item) 'lambda)
      (getp item 'lambda) )
  )

(defun pred-trans (pred &optional neg args fromop &aux res fr
			(numargs (list-length args))
			(realargs args))
  (when fromop 
	(setq *already-frags* (union *already-frags* (mapcar #'(lambda (a) (if (fragment a) a nil)) args)))
	(setq args (mapcar #'(lambda (a) (if (fragment a) (fragment-name a) a))
			   args))  )
  (setq res
    (cond ((symbolp pred)
	   (case (getp pred 'pred-type)
	     (constant 
	      (if (equal pred 'equal)
		  (let ()
		    (cond ((lambda-var (car args)) (cadr args))
			  (t (op-and-arg-trans neg 
					       (translation-for pred 'pred numargs)
					       args)) )
		    )
		(op-and-arg-trans neg (translation-for pred 'pred numargs)
				  args)) )
	     (variable 
	      (op-and-arg-trans neg (translation-for pred 'pred numargs)
				args))
	     (modified
	      (let ((op (getp pred 'operator))
		    (opargs (getp pred 'arguments)))
					; change here to avoid things like john makes itself
					; instead of john makes some juice
		(setq opargs (mapcar #'(lambda (a) 
					 (item-trans a nil 
						     (if (and (symbolp op) (getp op 'indicators)) 
							 nil args) t))  opargs))
		(setq res (op-trans op neg (append opargs args)))
		(unless (or (fragment-VP res)
			    (fragment-name res) (null args))
		  (setp res 'name 
			(if (fragment (car args))
			    (fragment-name (car args))
			  (car args)))
		  (update-entity-info (fragment-name res) res)
		  (set-objtype res))
		res))
	     (lambda
		 (setq res (wff-trans pred))
	       (if fromop 
		   (setq res (cond ((fragment-VP res))
				   ((fragment-ADVP res))
				   ((fragment-PP res))
				   (t res))) )
	       (unless (or (null res) 
			   (if (fragment-name res) 
			       (not (lambda-var (fragment-name res))) nil) 
			   (null args))
		 (setp res 'name 
		       (if (fragment (last1 args))
			   (fragment-name (last1 args))
			 (last1 args)))
		 (if (equal (fragment-fragtype res) 'VP)
		     (setq res (make-into-S res))
		   (let ()
		     (update-entity-info (fragment-name res) res)
		     (set-objtype res)) ) )
	       res)
	     (otherwise 
	      (op-and-arg-trans neg (translation-for pred 'pred numargs)
				args))
	     ))
	  (t )))
  (when (and (not fromop) *already-frags*)
    (setq fr (remove-if 'null (mapcar #'(lambda (a) 
					  (if (fragment a) a nil)) (intersection realargs *already-frags*))))
    (when fr
      (setq res (car (apply 'conjunct-frags (append fr
						    (if (fragment-adv res)
							(list (coerce-to 'S res))
						      (list res) )
						    ))))
      ))
  (when (traced-p 'trans-details)
    (print-line "  Response generator: Translated predicate " 
		(print-info pred))
    (cond (args 
	   (print-it " with arguments ")
	   (dolist (a args) (if (fragment a) (print-frag a) (print-it " " a)))
	   (print-line "    to"))
	  (t (print-it " to")))
    (if (fragment res) (print-frag res) (print-it res))
    )
  res)

(defun op-trans (op &optional neg args &aux res
		    (numargs (list-length args)))
  (setq res
  (cond ((symbolp op)
	 (case (getp op 'op-type)
	       (constant 
		(op-and-arg-trans neg (translation-for op 'op numargs) args))
	       (variable 
		(op-and-arg-trans neg (translation-for op 'op numargs) args))
	       (modified
		(let ((mop (getp op 'operator))
		      (opargs (getp op 'arguments)))
		     (setq opargs (mapcar #'(lambda (a) 
			(item-trans a nil nil t)) opargs))
		     (setq res (op-trans mop nil (append opargs args)) )
		     (unless (or (fragment-name res) (null args))
			     (setp res 'name 
				   (if (fragment (car args))
				       (fragment-name (car args))
				       (car args)))
			     (update-entity-info (fragment-name res) res)
			     (set-objtype res))
		     (if neg (setp res 'negated t))
		     res))
	       (otherwise 
		(op-and-arg-trans neg (translation-for op 'op numargs) args))
	       ))
	((numberp op) 
	 (if (> op 1) 
	     (op-and-arg-trans neg (list 'NP 1 '(number plural)
			 (list 'AP (list 'adj (princ-to-string op)))) args)
	     (op-and-arg-trans neg (list 'NP 1 
			 (list 'AP (list 'adj (princ-to-string op)))) args)
	     ))
	(t nil)))
  (when (traced-p 'trans-details)
      (print-line "  Response generator: Translated operator " (print-info op))
      (cond (args 
	    (print-it " with arguments ")
	    (dolist (a args) (if (fragment a) (print-frag a) (print-it " " a)))
	    (print-line "    to"))
	    (t (print-it " to")))
      (if (fragment res) (print-frag res) (print-it res))
      )
  res)

(defun term-trans (term &aux res)
  (setq res
	(if (member term *already-translated*)
	    term
  (cond ((symbolp term)
	 (case (getp term 'term-type)
	       (constant term)
	       (variable term)
	       (modified
		(let* ((afrags (mapcar 'item-trans (getp term 'arguments)))
		       (res (op-trans (getp term 'operator) nil afrags) ))
		     (setp res 'name term)
		     (set-objtype res)
		     res
		     )
		)
	       (function
		(if (not (member (getp term 'operator) '(date set-of)))
		(let* ((afrags (mapcar 'item-trans (getp term 'arguments)))
		       (res (op-trans (getp term 'operator) nil afrags) ))
		      (push term *functions*)
		     (setp res 'name term)
		     (setp res 'objtype 'function)
		     res)
		    term)
		)
	       (wff (wff-trans term))
	       (quasi-quoted-expression 
		(if (getp term 'quote-wff) (wff-trans term) term))
	       (otherwise term)
	       ))
	(t (initial-fragment term)))) )
  (when (traced-p 'trans-details)
      (print-line "  Response generator: Translated term " (print-info term) 
		  " to ")
      (if (fragment res) (print-frag res) (print-it res))
      )
  (setq *already-translated* (union *already-translated* (list term)))
  res)

(defun numlist (num &aux llist)
  (do ((a num (- a 1)))
      ((<= a 1))
      (push a llist)
      )
  llist)

(defun trans-warn (item)
  (when (and *response-warn* (not (getp item 'warned)))
	(warn-error "Response: Missing translation information for " item 
		    " - guessing")
	(setp item 'warned t))
  )

(defun translation-for (item ttype &optional numargs &aux df tr)  
  (cond ((not (symbolp item)) nil)
	((getp item 'trans))
	(t (setq df (default-trans item ttype numargs))
	   (when (not *default-trans*)
	     (print-line "Default information for " ttype " " item 
			 " would be: " df)
	     (print-line "Enter translation information for " ttype " " 
			 item " or () to accept default: ")
	     (setq tr (read))
	     (if tr (setq df tr)) )
	   (setp item 'trans df)
	   df)
	)
  )

(defun default-trans (item ttype &optional numargs &aux pp)
  (cond ((epi-variable item)		; should only occur in axiom schemas
	 (case (find-sort item)
	       (pred (if (> numargs 1)
			 '(VP (name 1) (pred do) (objects 
				   (NP (noun thing) (AP (adj some))) 
				   (PP (prep to) (NP 2))))
			 '(VP (name 1) (conj or) (parts (VP (verb be)) 
				 (VP (verb do) (objects (NP (noun thing) (det some))))))
			 ;'(AP (name 1)(adv some) (adj predicate))
			 ))
	       (quantifier '(AP (name 1)(adv some) (adj quantifier)))
	       (operator '(AP (name 1)(adv some) (adj operator)))
	       (otherwise '(AP (name 1)(adv some) (adj thing)))
	       ))
	;also handle "compound" preds with pred-parts property
	((setq pp (getp item 'pred-parts))
	 (trans-warn item)
	 (let ((t1 (translation-for (cadr pp) ttype numargs))
	       aps)
	      (when t1
	      (case (car t1)
		    (NP (if (setq aps (member-if #'(lambda (a) 
			         (and (listp a) (equal (car a) 'AP))) t1))
			    (let* ((p1 (car aps))
				  (p2 (remove p1 t1 :test 'equal))
				  (p3 (car (member-if #'(lambda (a)
                                     (and (listp a) (equal (car a) 'adj))) p1)))
				  (p4 (remove p3 p1 :test 'equal)))
			      (append p2 (list (append p4 (list (list 'adj 
				      (append (list (car pp)) (cdr p3)))))))
			      )
			    (append t1 (list (list 'AP (list 'adj (car pp)))))
			    ))
		    (otherwise
		     (if (member 'adv t1)
			 (let* ((p1 (assoc 'adv t1))
			       (p2 (remove p1 t1 :test 'equal)))
			      (append p2 (list (list 'adv 
				     (append (list (car pp)) (cdr p1) ))))
			      )
			 (append t1 (list (list 'adv (car pp))))
			 ) ) ) )
	      ) )
	((and numargs (equal ttype 'pred))
	 (let ((tt (type-pred-p item)))
	   (if (not tt) (trans-warn item))
	   (if (> numargs 1)
	       (list 'S '(NP 1) (list 'VP 
				      (list 'verb item)
				      (append (list 'objects) (numlist numargs))))
	     (cond (tt
		    (list 'NP '(name 1) (list 'noun item)))
		   ((active-topic item)
		    (list 'S '(NP 1) (list 'VP (list 'verb item))))
		   (t
		    (list 'AP '(name 1) (list 'adj item)) ))
	     ))
	 )
	
	((type-pred-p item)
	 (list 'NP (list 'noun item)))
	((equal ttype 'op) 
	 (trans-warn item)
	 (list 1 (list 'adv item)))
	))

(defun active-topic (item &aux (ind (getp item 'indicators)))
  (setq ind (mapcar #'(lambda (a) (if (listp a) (car a) a)) ind))
  (dolist (i ind)
	  (if (member i *active-topics* :test #'(lambda (a b)
				(member (hier-rel a b) '(equivalent subsumed))))
	      (return-from active-topic t)))
  nil)

(defun quant-info (quant var &aux extra)
  (when (and quant (not (equal (find-sort var) 'episode)) )
	(setq extra (make-frag 'NP (cons 'name var) (cons 'quantifier quant)
			       (cons 'required nil)))
	)
  extra)
			  
(defun tense-info (item)
  (cond ((not (symbolp item)) nil)
	((epi-variable item) nil)
	(t 'past))
  )


(defun combine-trans (trans frags &optional (force-frag t) &aux result frag num
			    alf ftype)
  (if (fragment trans) (return-from combine-trans trans))
  (setq frag
  (cond ((numberp trans)
	 ; just return the argument corresponding to that
	 ; it may be a fragment or a simple symbol
	 (setq result (if (>= (length frags) trans) (nth (- trans 1) frags)))
	 (when (and result (not (fragment result)))
	       (cond ((setq alf (car (member result *already-frags* :test 
			  #'(lambda (a b) 
			    (equal (fragment-name a) (fragment-name b))))))
		      (setq result (make-frag (fragment-fragtype alf) 
				       (cons 'name result)) ))
		     ((or (and (symbolp result)
			       (member (getp result 'term-type ) 
				  '(function record quoted-expression
					     quasi-quoted-expression)))
			  (or (numberp result) (stringp result))
			  force-frag)
		      (setq result (initial-fragment result)))) )
	 result)
	((atom trans) trans)
	((numberp (setq num (car trans)))
	 ; take that fragment, add properties to it
	 (setq result (combine-trans (car trans) frags))
	 ; do the rest of the translations and add as properties
	 (setq result (combine-properties result (cdr trans) frags))
	 result)
	((member (setq ftype (car trans)) *fragment-types*)
	 ; build a fragment of this type, using the info in the rest of the
	 ; trans
	 (cond ((numberp (setq num (cadr trans)))
		;take that item and coerce it into the desired form
		(setq result (combine-trans num frags))
		(setq result (coerce-to ftype result))
		)
	       (t (setq result (make-frag ftype)))
	       )
	 ; do the rest of the translations and add as properties
	 (setq result (combine-properties result 
		(if (numberp num) (cddr trans) (cdr trans)) frags))
	 result)
	((member ftype *fragment-groups*)
	 ; return a list of the desired fragments
	 (mapcar #'(lambda (a) (combine-trans a frags)) (cdr trans))
	 )
	((member ftype *fragment-routines*)
	 (apply ftype (mapcar #'(lambda (a) 
		(if (numberp a)
		    (cond ((if (>= (length frags) a) (nth (- a 1) frags)))
			  (t a))
		    a)) frags))
	 )
	(t ; not a new fragment, just a property
	   (mapcar #'(lambda (a) (combine-trans a frags)) trans)
	   )
	))
  frag)

(defun combine-properties (frag props frags &aux newfrag nfrag vfrag newf newp
				oldp)
  (dolist (p props)
	  (cond ((atom p) ; shouldn't ever be here!
		 (print-line "how did an atom get into the property list? "
			     props))

		((or (member (car p) *fragment-types*)
		     (member (car p) *fragment-groups*))
		 (setq oldp (getp frag (car p)))
		 (setq newf (combine-trans p frags))
		 (setp frag (car p) 
		       (if (and oldp (member (car p) *fragment-types*))
			   (let ((ll (if (listp oldp) (append oldp (list newf))
					  (list oldp newf)))
				 )
				(dolist (l ll) 
					(setp l 'name (fragment-name frag)))
			   (make-conj nil 'and ll (fragment-name frag) frag))
			   newf))
		 (when (and newf (member (car p) '(AP PP RELC))
			    (not (getp newf 'name)))
		       (setp newf 'name (getp frag 'name)))
		 (if (member (car p) *fragment-types*)
		     (when (getp frag (car p))
			   (setp (getp frag (car p)) 'head frag))
		     (dolist (a (getp frag (car p))) (setp a 'head frag)))
		 (when (and (fragment-name frag) (member (car p) '(VP S)))
		       (update-entity-info (fragment-name frag) frag))
		 )
		((equal (car p) 'name)
		 (setq nfrag (combine-trans (cadr p) frags nil))
		 (cond ((fragment nfrag)
		        (setp frag 'name (fragment-name nfrag))
			(setp frag 'objtype (fragment-objtype nfrag))
			(if (and (equal (fragment-fragtype frag) 'AP)
				 (equal (fragment-fragtype nfrag) 'NP)
				 (not (fragment-AP frag)))
			    (let ()
				 (setp frag 'head nfrag)
				 (setp nfrag 'AP frag)
				 (setq newfrag nfrag))
			    (let ()

			(setq nfrag (coerce-to 'NP nfrag))
			(setq vfrag (coerce-to 'VP frag))
			(if (equal (fragment-fragtype frag) 'S)
			    (if vfrag 
				(setq frag (make-frag 'S (cons 'NP nfrag)
				      (cons (fragment-fragtype vfrag) vfrag)))
				(setp frag 'NP nfrag))
			(setq newfrag (make-frag 'S (cons 'NP nfrag)
				  (cons (fragment-fragtype vfrag) vfrag))))
		        ) ))
		       (t (setp frag 'name nfrag)
			  ))
		 (unless (or (null (fragment-name frag)) 
			     (equal (fragment-name frag) 'removed))
			 (pushnew (fragment-name frag) *entity-list*))
		 (set-objtype frag)
		 )
		((member (car p) *fragment-routines*)
		 (setq frag (apply (car p) (mapcar #'(lambda (a) 
				(if (numberp a)
		    (cond ((if (>= (length frags) a) (nth (- a 1) frags)))
			  (t a))
		    a)) frags)))
		 )
		(t ; just a regular property
		   (if (member (car p) *multiple-fields*)
		       (setq oldp (getp frag (car p)))
		       (setq oldp nil))
		   (setq newp (combine-trans (cadr p) frags) )
		   (setq newp
		   (if oldp
		       (if (listp oldp)
			   (if (listp newp) (append newp oldp) (append (list newp) oldp))
			   (if (listp newp) (append newp (listp oldp)) (list newp oldp)))
		       newp))

		   (setp frag (car p) 
			 (if (and (member (car p) '(determiner adv adj prep))
				  (symbolp newp))
			     (cond ((getp newp 'print)) 
				   ((fragment-name newp)) 
				   (t newp))
			     newp))
		   (when (and (fragment-name frag)
			      (member (car p) '(noun pronoun proper-noun)))
			 (update-entity-info (fragment-name frag) frag))
		   )
		))
  (if newfrag newfrag frag)
  )

(defun op-and-arg-trans (neg optrans argtrans &aux newop)
  (setq argtrans (mapcar #'(lambda (a) 
			     (if (fragment a) a (op-and-arg-trans nil a nil))) argtrans))
  (setq newop (combine-trans optrans argtrans))
  (when neg (if (fragment-negated newop) 
		(setp newop 'negated nil)
	      (setp newop 'negated t)))
  newop)

(set-display-function 'trans 'predicate-translation "Shows the translation information for a predicate")

(defun predicate-translation (brief &rest args &aux tr item)
  (declare (ignore brief))
  (dolist (a args)
	  (if (symbolp a) (setq tr (getp a 'trans)) (setq tr nil))
	  (if tr
	      (print-line "Translation for " a ": " tr)
	      (let ()
		   (print-line "No translation information for " a )
		   (print-line "  default translation: ")
		   (cond ((type-pred-p a)
			  (print-it (list 'NP (list 'noun a))))
			 ((equal (entity-type a) 'op)
			  (print-it (list 1 (list 'adv item))))
			 (t (print-it (list 'VP (list 'verb a)))
			    (print-it "  OR  ")
			    (print-it (list 'AP (list 'adj a)))
			    ))
		   )
	      ))
  )

(defun add-word (item trans)
  (when (symbolp item)
	(setp item 'trans trans))
  )

(defun conditional-trans (ante cons &optional adverb &aux pp)
  (setq *conditional* t)

  (setq ante (cond ((apply 'conjunct-frags ante)) (t ante)))
  (setq cons (cond ((apply 'conjunct-frags cons)) (t cons)))

  (setq pp (massage-conditional ante cons adverb))
  (setq ante (car pp))
  (setq cons (cdr pp))

  ; combine if possible, otherwise make into if-then statement
  (if (if-then-test ante cons)
      (if-then ante cons)
      (let* ((res (cond-statement ante cons)))
           (if (> (list-length res) 1)
	       (if *success-combine* 
		   (if-then (list (car res)) (cdr res)) 
		   (if-then ante cons))
               (car res)))
      )
  )


