;       Last change    August 1993
(in-package set-equal)

(tweakable 'epilog::*set-assert* t 
  "Indicates whether or not set specialist should make assertions"
  'boolean)


(defun update-card (net curset setmems fromset &aux ll never-done mems netname)
  (setq ll (list-length (remove '+ setmems)))
  (when (and (numberp (episet-cardinality-max curset))
	     (= (episet-cardinality-max curset) ll))
	(setq setmems (remove '+ setmems))
	(change-field 'episet-finished curset t))
  ; if the set contents for set contain a + (came from a partially known
  ; set), or if any element of setdesc is itself a partially known set,
  ; make sure this one is flagged as partially known
  (unless (episet-finished curset)
	  (if (or (+member setmems) (unfin-set-member net setmems) never-done)
	      (setq setmems (union setmems (list '+)))
	      (change-field 'episet-finished curset t)))
  (if (traced-p 'epilog::set-entry)
      (print-line  (if fromset "    Set Specialist: "
			       "    Equality Specialist: ")
		   " Setting contents of set "
		       (episet-identifiers curset) " to " setmems))
  (change-field 'episet-contents curset setmems)

  ;; call EPILOG to update its equivalence sets
  (when (member 'equivalence (episet-properties curset))
	(setq mems (remove '+ setmems))
	(setq netname (episet-subnet-identifier net)) 
	(dolist (m mems) (save-equivalence m netname mems))
	)

  (when (or (not (numberp (episet-cardinality-min curset)))
	    (< (episet-cardinality-min curset) ll))
	(if (traced-p 'epilog::set-entry)
      (print-line (if fromset "    Set Specialist: " 
		      "    Equality Specialist: ")
		  "Changing cardinality lower bound of set "
		   (episet-identifiers curset) " to " ll))
	(change-field 'episet-cardinality-min curset ll))
  (when (and (episet-finished curset)
	     (or (not (numberp (episet-cardinality-max curset)))
		 (< ll (episet-cardinality-max curset))
		 (> ll (episet-cardinality-max curset))
		 ))
	(if (traced-p 'epilog::set-entry)
	    (print-line (if fromset "    Set Specialist: "
			    "    Equality Specialist: ")
			"Changing cardinality upper bound of set "
		  (episet-identifiers curset) " to " ll))
	(change-field 'episet-cardinality-max curset ll))
  )

; unfin-set-member returns t if any set in setdesc is a partially known set

(defun unfin-set-member (net setdesc)
  (dolist (m setdesc)
	  (if (equal (find-sort m) 'set) ;(compatible-sorts 'set (find-sort m))
	      (let ((s (findset net m)))
		   (when (and s
			      (not (episet-finished s)))
			 (return t))))
	  ))

; sets-equal makes two sets equal. If both exist and have members, the
; intersection is taken as the members for the new set

(defun sets-equal (net setname1 setname2 &optional fromset from-check &aux
		       (sarray (episet-subnet-sets net))
		       (curset1 (findset net setname1))
		       (curset2 (findset net setname2))
		       c1 c2 newset ll newmin newmax never-done)



  (cond (
	 ;; if neither set exists, create one and make both names
	 ;; point to it.
	 (and (null curset1) (null curset2))
	 (if (traced-p 'epilog::set-entry)
	     (print-line (if fromset "    Set Specialist:" 
				     "    Equality Specialist:") " Setting " 
			 setname2 " equal to " setname1))
	 (setq curset1 (getset net setname1))
	 (change-field 'episet-identifiers curset1
		       (cons setname2 (episet-identifiers curset1)))
	 (change-hash setname2 sarray curset1))

	 ;; if 2 exists but 1 doesn't, make 1 point to 2.
	((null curset1)
	 (if (traced-p 'epilog::set-entry)
	     (print-line (if fromset "    Set Specialist:" 
				     "    Equality Specialist:") " Setting " 
			 setname1 " equal to " setname2))
	 (change-field 'episet-identifiers curset2
		       (cons setname1 (episet-identifiers curset2)))
	 (change-hash setname1 sarray curset2))

	;; if 1 exists but 2 doesn't, make 2 point to 1.
	((null curset2)
	 (if (traced-p 'epilog::set-entry)
	     (print-line (if fromset "    Set Specialist:" 
				     "    Equality Specialist:") " Setting " 
			 setname2 " equal to " setname1))
	 (change-field 'episet-identifiers curset1
		       (cons setname2 (episet-identifiers curset1)))
	 (change-hash setname2 sarray curset1))
	
	;; both sets exist, so take the union of those members
	;; (filling in more of partially known sets)
	(t (let ((curmems1 (episet-contents curset1))
		 (curmems2 (episet-contents curset2))
		 (memtype1 (episet-member-type curset1))
		 (memtype2 (episet-member-type curset2))
		 (conn1 (episet-connections curset1))
		 (conn2 (episet-connections curset2)))
	     (setq newset (union curmems1 curmems2))
	     ;; set the partially known flag (finished) to indicate partially known
	     ;; if either set was previously partially known
	     (unless from-check (check-equal-members net curmems1 curmems2))
	     (change-field 'episet-connections curset1 
			   (union conn1 conn2 :test 'equal))
	     (change-field 'episet-properties curset1 
			   (union (episet-properties curset1) (episet-properties curset2)
				  :test 'equal))
	     
	     (change-field 'episet-finished curset1 
			   (and (episet-finished curset1) (episet-finished curset2) 
				(not never-done)))
	     (change-field 'episet-member-type curset1
			   (cond ((null memtype1) memtype2)
				 ((null memtype2) memtype1)
				 (t	; shouldn't occur
				  memtype1)))
	     (change-field 'episet-all curset1
			   (or (episet-all curset1) (episet-all curset2)))
	     (when (not (equal memtype1 (episet-member-type curset1)))
	       (add-type-assert setname1 (episet-member-type curset1) 
				(episet-all curset1)) )
	     (unless (null (set-exclusive-or newset curmems1))
	       ;; change cardinalities also
	       (setq ll (list-length (remove 'epilog::+ newset)))
	       (setq newmin (if (numberp (setq c1 (episet-cardinality-min curset1)))
				(if (numberp (setq c2 (episet-cardinality-min curset2)))
				    (if (> c1 c2) c1 c2)
				  c1)
			      c2))
	       (if (and (numberp newmin) (< newmin ll)) (setq newmin ll))
	       (change-field 'episet-cardinality-min curset1 newmin)
	       (if never-done
		   (setq newmax nil)
		 (let ()
		   (setq newmax (if (numberp (setq c1 (episet-cardinality-max curset1)))
				    (if (numberp (setq c2 (episet-cardinality-max curset2)))
					(if (< c1 c2) c1 c2)
				      c1)
				  c2))
		   (if (and (numberp newmax) (numberp (episet-cardinality-min curset1))
			    (> (episet-cardinality-min curset1) newmax) )
		       (setq newmax (episet-cardinality-min curset1)))
		   (if (and (numberp newmax) (> ll newmax)) (setq newmax ll)) ) )
	       (change-field 'episet-cardinality-max curset1 newmax)
	       (when (and (numberp (episet-cardinality-max curset1))
			  (= (episet-cardinality-max curset1) ll))
	         (change-field 'episet-finished curset1 t)
		 (setq newset (remove 'epilog::+ newset)))
	       (when (traced-p 'epilog::set-entry)
		 (print-line  (if fromset "    Set Specialist: " 
				"    Equality Specialist: ")
			      "Setting contents of set "
			      setname1 " to " newset)
		 (print-line (if fromset "    Set Specialist: "
			       "    Equality Specialist: ")
			     "Setting " setname2
			     " equal to " setname1))
	       (change-field 'episet-contents curset1 newset) )
	     (change-field 'episet-identifiers curset1
			   (append (episet-identifiers curset1)
				   (episet-identifiers curset2)))
	     (mapc #'(lambda (name)
		       (change-hash name sarray curset1))
		   (episet-identifiers curset2))
	     )
	   )
	)
  )

(defun add-type-assert (setname type set-of-all &aux a1)
  (when (and type epilog::*set-assert*)
	(if set-of-all
	    (setq a1 (list 'epilog::A 'epilog::x 
			   (list (list 'epilog::x 'member-of setname)
				 'epilog::<=>
				 (list 'epilog::x (print-info type)))))
	    (setq a1 (list 'epilog::A 'epilog::x 
			   (list 'epilog::x 'member-of setname)
			   (list 'epilog::x (print-info type)))) )
	(when (traced-p 'epilog::set-assert)
	      (print-line "    Set Specialist: Asserting: " a1))
	(setq epilog::*assertions* (append epilog::*assertions* 
					   (make-spec-inference (list a1))) ))
  )

(defun enter-equivalence (net a1 a2 &optional fromset from-check &aux
			      a1name a2name s1 s2
                              mems1 mems2)
  (when (and (symbolp a1) (symbolp a2))
    (setq a1name (equiv-set-name a1))
    (setq a2name (equiv-set-name a2))
    (setq s1 (getset net a1name))
    (setq s2 (getset net a2name))
    (change-field 'episet-properties s1 (union (episet-properties s1)
					       (list 'equivalence)))
    (change-field 'episet-properties s2 (union (episet-properties s2)
					       (list 'equivalence)))
    (setq mems1 (episet-contents s1))
    (setq mems2 (episet-contents s2))
    (setq mems1 (union mems1 mems2))
    (setq mems1 (union mems1 (list a1 a2)))
    
    ;; update cardinality and members
    (update-card net s1 mems1 fromset)
    (sets-equal net a1name a2name fromset from-check)
    ))

(defun check-equal-members (net mems1 mems2 &aux a1)
  (when (or (and (not (member '+ mems1)) (equal (list-length mems1) 1))
            (and (not (member '+ mems2)) (equal (list-length mems2) 1)))
        (dolist (a mems1)
                (dolist (b mems2)
                        ; add modal embedding if there is any!!
                        (when epilog::*set-assert*
			      (setq a1 (make-spec-inference (list a 'equal b)))
			      (when (traced-p 'epilog::set-assert)
				    (print-line 
				     "    Set Specialist: Asserting: " a1))
                        (setq *assertions* (append *assertions* (list a1))) )
                        (enter-equivalence net a b t t)))
        )
  )

