;       Last change    August 1993
(in-package set)

;                         FUNCTIONS
; These functions are functions the user may have in propositions
; When evaluated, they use the global specialist effort level for effort

; set-of returns a set corresponding to the given argument members

(defun set-of (&rest args &aux
		       res )
  (unless (null args)
	  (setq res (new-record args 'set))
	  (if (or (traced-p 'epilog::function-eval)
		  (traced-p 'epilog::set-function-eval))
	      (print-line "    Set Specialist: SET-OF "
		               args " evaluated to " (print-info res)))
	  )
  res)

(defun Nset-of (&rest args &aux res nset
		       (usenet (get-subnet nil)))
  (unless (null args)
	  (setq res (pack (append (list 'epilog::set-of )
				  (mapcar 'print-info args))))
	  (unless (findset usenet res) 
		  (change-property res 'sort 'set)
		  (setq nset (getset usenet res))
		  ;; don't need to call writeable-copy because this is
		  ;; guaranteed to create a new struct.
		  (change-field 'episet-contents nset args)
		  (change-field 'episet-finished nset t))
          (if (or (traced-p 'epilog::function-eval)
		  (traced-p 'epilog::set-function-eval))
              (print-line "    Set Specialist: SET-OF "
		               args " evaluated to " (print-info res)))
	  )
  res)

; calc-intersect determines the intersection of two lists of members
(defun calc-intersect (lists &aux
			     (list1 (car lists))
			     (list2 (cadr lists))
			     (fin2 (not (+member list2)))
			     )
  (if (null list2)
      list1
      (let (res)
	   (dolist (l (remove '+ list1))
		   (if (equal (is-a-member l nil (remove '+ list2) fin2) t)
		       (setq res (append res (list l)))) )
	   (if (and (+member list1) (+member list2))
	       (pushnew '+ res))
	   (if (cddr lists)
	       (calc-intersect (cons res (cddr lists)))
	       res)
	   ))
  )

; intersect-0 returns a set which consists of all elements in the intersection
; of all arguments - if an argument is a set, it is expanded to its set 
; elements - if  an element is a set, it is expanded to all its set elements

(defun intersect-members-0 (&rest args &aux
			 (usenet (get-subnet nil))
			 res)
  (unless (null args)
	  (setq res (calc-intersect (mapcar #'(lambda (a)
				  (if (not (immed-set a))
				      (if (equal (find-sort a) 'set)
					  (get-contents-0 usenet a)
					  (list a))
				      (if (equal (find-sort a) 'set)
					  (get-set-immed-0 usenet a)
					  (list a))
				      )) args)))
	  (unless (null res)
	          (setq res (new-record res 'set))
          (if (or (traced-p 'epilog::function-eval)
		  (traced-p 'epilog::set-function-eval))
              (print-line "    Set Specialist: INTERSECT-MEMBERS-0 "
		               args " evaluated to " (print-info res)))
	  ))
  res)

; intersect returns the set corresponding to the intersection of all
; set arguments

(defun intersect-members (&rest args &aux
			(usenet (get-subnet nil))
			res)
  (unless (or (null args)
	      (eval (append (list 'or) (mapcar #'(lambda (a) 
				    (not (equal (find-sort a) 'set)))
		      args))))
	  (setq res (calc-intersect (mapcar #'(lambda (a)
					      (if (not (immed-set a))
						  (get-contents usenet a)
						  (get-set-immed a)))
					    args)))
	  (unless (null res)
	          (setq res (new-record res 'set))
          (if (or (traced-p 'epilog::function-eval)
		  (traced-p 'epilog::set-function-eval))
              (print-line "    Set Specialist: INTERSECT-MEMBERS "
		               args " evaluated to " (print-info res)))
	  ))
  res)

; calc-union determines the union of two lists of members

(defun calc-union (lists &aux
			 (list1 (car lists))
			 (list2 (cadr lists))
			 )
  (if (null list2)
      list1
      (let ((res (union list1 list2)))
	   (if (cddr lists)
	       (calc-union (cons res (cddr lists)))
	       res)
	   ))
  )

; union-of returns the set corresponding to the union of all
; set arguments

(defun union-of-members (&rest args &aux
		       (usenet (get-subnet nil))
		       res)
  (unless (or (null args)
	      (eval (append (list 'or) (mapcar #'(lambda (a) 
				    (not (equal (find-sort a) 'set)))
		      args))))
	  (setq res (calc-union (mapcar #'(lambda (a)
					      (if (not (immed-set a))
						  (get-contents usenet a)
						  (get-set-immed a)))
					    args)))
	  (unless (null res)
	          (setq res (new-record res 'set))
          (if (or (traced-p 'epilog::function-eval)
		  (traced-p 'epilog::set-function-eval))
              (print-line  "    Set Specialist: UNION-OF-MEMBERS "
		               args " evaluated to " (print-info res)))
	  ))
  res)

; union-of-0 returns a set which consists of all elements in the union of all
; arguments - if an argument is a set, it is expanded to its set elements - if
; an element is a set, it is expanded to all its set elements

(defun union-of-members-0 (&rest args &aux
			 (usenet (get-subnet nil))
			 res)
  (unless (null args)
	  (setq res (calc-union (mapcar #'(lambda (a)
				  (if (not (immed-set a))
				      (if (equal (find-sort a) 'set)
					  (get-contents-0 usenet a)
					  (list a))
				      (if (equal (find-sort a) 'set)
					  (get-set-immed-0 usenet a)
					  (list a))
				      )) args)))
	  (unless (null res)
	          (setq res (new-record res 'set))
          (if (or (traced-p 'epilog::function-eval)
		  (traced-p 'epilog::set-function-eval))
              (print-line "    Set Specialist: UNION-OF-MEMBERS-0 "
		               args " evaluated to " (print-info res)))
	  ))
  res)

; set-diff returns the set corresponding to the set difference between the 
; two set arguments

(defun set-diff-members (&rest args &aux
		       (usenet (get-subnet nil))
		       (s1 (car args))
		       (s2 (cadr args))
		       res)
  (unless (or (null args)
	      (cddr args)
	      (eval (append (list 'or) (mapcar #'(lambda (a) 
				    (not (equal (find-sort a) 'set)))
		      args))))
	  (setq res (calc-set-diff ;set-difference 
		     (if (not (immed-set s1))
			 (findset usenet s1) 
	                 (get-set-immed s1))
		     (if (not (immed-set s2))
			 (findset usenet s2) 
	                 (get-set-immed s2))
		    ))
	  (unless (null res)
	          (setq res (new-record res 'set))
          (if (or (traced-p 'epilog::function-eval)
		  (traced-p 'epilog::set-function-eval))
              (print-line "    Set Specialist: SET-DIFF-MEMBERS "
		               args " evaluated to " (print-info res)))
	  ))
  res)

; set-diff-0 returns the set corresponding to the set difference between the 
; two set arguments (using lowest level members)

(defun set-diff-members-0 (&rest args &aux
		       (usenet (get-subnet nil))
		       (s1 (car args))
		       (s2 (cadr args))
		       res)
  (unless (or (null args)
	      (cddr args)
	      (eval (append (list 'or) (mapcar #'(lambda (a) 
				    (not (equal (find-sort a) 'set)))
		      args))))
	  (setq res (calc-set-diff ;set-difference 
		     (if (not (immed-set s1))
			 (get-contents-0 usenet s1)
	                 (get-set-immed-0 usenet s1))
		     (if (not (immed-set s2))
			 (get-contents-0 usenet s2)
	                 (get-set-immed-0 usenet s2))
		    ;:test 'equal 
		    ))
	  (unless (null res)
	          (setq res (new-record res 'set))
          (if (or (traced-p 'epilog::function-eval)
		  (traced-p 'epilog::set-function-eval))
              (print-line  "    Set Specialist: SET-DIFF-MEMBERS-0 "
		               args " evaluated to " (print-info res)))
	  ))
  res)

(defun cardinality-of (&rest args &aux
			     (usenet (get-subnet nil))
			     (s1 (car args))
			     res)
  (setq res
    (if (not (immed-set s1))
	(let* ((setitem (getset usenet s1))
	       (cardmin (episet-cardinality-min setitem))
	       (cardmax (episet-cardinality-max setitem)))
	  (if (equal cardmin cardmax)
	      cardmin
	    nil
	    )
	  )
      (list-length (get-set-immed s1))
      ))
  (if (or (traced-p 'epilog::function-eval)
	  (traced-p 'epilog::set-function-eval))
      (print-line "    Set Specialist: CARDINALITY-OF "
		               args " evaluated to " (print-info res)))
  res
  )

; set-members returns a quoted set expression consisting of the currently
; known members of the set given to it

(defun set-members (&rest args &aux
		       (usenet (get-subnet nil))
		       (s1 (car args))
		       res)
  (setq res
    (if (immed-set s1)
	(get-set-immed s1)
      (get-contents usenet s1)
      )
    )
  (setq res (remove-if #'+equal res))
  (unless (null res)
    (setq res (new-record res 'set))
    (if (or (traced-p 'epilog::function-eval)
	    (traced-p 'epilog::set-function-eval))
	(print-line "    Set Specialist: SET-MEMBERS "
		    args " evaluated to " (print-info res)))
    )
  res)

; set-members returns a quoted set expression consisting of the currently
; known lowest level (recursively) members of the set given to it

(defun set-members-0 (&rest args &aux
		       (usenet (get-subnet nil))
		       (s1 (car args))
		       res)
  (setq res
	(if (not (immed-set s1))
	    (get-contents-0 usenet s1)
	    (get-set-immed-0 usenet s1)))
  (setq res (remove-if #'+equal res))
  (unless (null res)
	  (setq res (new-record res 'set))
  (if (or (traced-p 'epilog::function-eval)
	  (traced-p 'epilog::set-function-eval))
      (print-line "    Set Specialist: SET-MEMBERS-0 "
		               args " evaluated to " (print-info res)))
	  )
  res)

; number-members returns the number of members currently known in the
; given set

(defun number-members (&rest args &aux
			     (s1 (car args))
			     (usenet (get-subnet nil))
			     res)
  (when (and (null (cdr args))
	     (equal (find-sort s1) 'set))
	(if (not (immed-set s1))
	    (setq res (list-length (remove '+ (get-contents usenet s1))))
	    (setq res (list-length (remove '+ (get-set-immed s1)))))
	(if (or (traced-p 'epilog::function-eval)
		(traced-p 'epilog::set-function-eval))
              (print-line "    Set Specialist: NUMBER-MEMBERS "
		               args " evaluated to " (print-info res)))
	)
  res)

; number-members-0 is the same as number-members except that it returns the
; number of lowest level members by expanding each set member to its
; members, recursively.

(defun number-members-0 (&rest args &aux
			     (s1 (car args))
			     (usenet (get-subnet nil))
			     res)
  (when (and (null (cdr args))
	     (equal (find-sort s1) 'set))
	(if (not (immed-set s1))
	    (setq res (list-length (remove '+ (get-contents-0 usenet s1))))
	    (setq res (list-length (remove '+ (get-set-immed-0 usenet s1)))))
	(if (or (traced-p 'epilog::function-eval)
		(traced-p 'epilog::set-function-eval))
              (print-line "    Set Specialist: NUMBER-MEMBERS-0 "
		               args " evaluated to " (print-info res)))
	)
  res)

(defun set-of-all (&rest args &aux
			 (pr (car args))
			 (usenet (get-subnet nil))
			 res setn)
  (unless (or (null args)
	      (cdr args))
	  (setq res (pack (list 'epilog::set-of-all (print-info pr))))
	  (change-property res 'sort 'set)
	  (setq setn (findset usenet res))
	  (unless setn
		  (setq setn (getset usenet res))
		  ;; don't need to call writeable-copy because this is
		  ;; guaranteed to create a new struct.

		  ; how to get this in? this doesn't work!
		  (add-type-assert res pr t)
		  (change-field 'episet-member-type setn pr)
		  (change-field 'episet-all setn t)
		  )
	(if (or (traced-p 'epilog::function-eval)
		(traced-p 'epilog::set-function-eval))
              (print-line "    Set Specialist: SET-OF-ALL "
		               args " evaluated to " (print-info res)))
	  )
  res)


(defun intersect (&rest args &aux
		       (usenet (get-subnet nil))
		       sclist res sname ntype type1 type2 cp argsets)
  ; must only be two sets here
  (unless (or (null args)
	      (eval (append (list 'or) (mapcar #'(lambda (a) 
				    (not (equal (find-sort a) 'set)))
		      args))))
	  ; build name for this set, create it
	  (setq res (pack (append (list 'epilog::intersect) 
				  (mapcar 'print-info args))))
	  (cond ((setq sname (findset usenet res))) 
		(t (setq sname (getset usenet res))
		   ;; don't need to call writeable-copy because this is
		   ;; guaranteed to create a new struct.

		   (change-property res 'sort 'set)
		   ; find current contents and add them also member type, all
		   ; what if some of these sets are "immed sets"?
		   (setq argsets (mapcar #'(lambda (a) (getset usenet a)) args))
		   (setq type1 (episet-member-type (car argsets)))
		   (setq type2 (episet-member-type (cadr argsets)))
		   (cond ((and type1 type2 
			       (setq cp (spec-compare-preds type1 type2))
			       (member cp '(equivalent subsumed)))
			  (setq ntype type1))
			 ((and cp (member cp '(subsumes)))
			  (setq ntype type2))
			 ((null type1) (setq ntype type2))
			 ((null type2) (setq ntype type1))
			 (t (setq ntype
			      (normalize-pred (list 'L 'epilog::l-x 
				       (list (list 'epilog::l-x (print-info type1)) 
					     'and 
					     (list 'epilog::l-x (print-info type2)))) '+)))
			 )
		   (when ntype 
			 (change-field 'episet-member-type sname ntype)
			 (add-type-assert res ntype nil))

		   (change-field 'episet-contents sname
				    (calc-intersect (mapcar 'episet-contents argsets)))
		   (change-field 'episet-connections sname
				    (setq sclist (list 
						 (make-episet-connect :main-set res
							    :how 'intersect
							    :sets args))))
		   ; add this connection to all the argument sets as well
		   (dolist (a argsets)
			   (change-field 'episet-connections-from a
					    (append (episet-connections-from a)
						    sclist)))
		   ))
	(if (or (traced-p 'epilog::function-eval)
		(traced-p 'epilog::set-function-eval))
              (print-line "    Set Specialist: INTERSECT "
		               args " evaluated to " (print-info res)))
	  )
  res)

(defun calc-set-diff (s1 s2 &aux
			 res set2 mems1 mems2 fin2)
  (cond ((listp s1) (setq mems1 s1))
	(t (setq mems1 (episet-contents s1))))
  (cond ((listp s2) 
	 (setq mems2 s2)
	 (setq fin2 (not (+member s2))))
	(t (setq mems2 (episet-contents s2))
	   (setq set2 s2)
	   (setq fin2 (episet-finished s2))))
  (dolist (m (remove '+ mems1))
	  (if (null (is-a-member m set2 mems2 fin2))
	      (setq res (append res (list m))))
	  )
  (if (+member mems1) (pushnew '+ res))
  res)

(defun set-diff (&rest args &aux
		       (usenet (get-subnet nil))
		       sclist res sname ntype type1 type2 argsets)
  ; must only be two sets here
  (unless (or (null args)
	      (eval (append (list 'or) (mapcar #'(lambda (a) 
				    (not (equal (find-sort a) 'set)))
		      args))))
	  ; build name for this set, create it
	  (setq res (pack (append (list 'epilog::set-diff) 
				  (mapcar 'print-info args))))
	  (cond ((setq sname (findset usenet res))) 
		(t (setq sname (getset usenet res))
		   ;; don't need to call writeable-copy because this is
		   ;; guaranteed to create a new struct.

		   (change-property res 'sort 'set)
		   ; find current contents and add them also member type, all
		   ; what if some of these sets are "immed sets"?
		   (setq argsets (mapcar #'(lambda (a) (getset usenet a)) args))
		   (setq type1 (episet-member-type (car argsets)))
		   (setq type2 (episet-member-type (cadr argsets)))
		   (cond ((null type2) (setq ntype type1))
			 ((and type1 type2) 
			  (setq ntype
			    (normalize-pred (list 'L 'epilog::l-x 
				  (list (list 'epilog::l-x (print-info type1)) 
					'and 
					(list 'not 'epilog::l-x (print-info type2))))
					       '+)))
			 )
		   (when ntype 
			 (change-field 'episet-member-type sname ntype)
			 (add-type-assert res ntype nil))

		   (change-field 'episet-contents sname 
				 (calc-set-diff (car argsets) (cadr argsets)))
		   (change-field 'episet-connections sname
				    (setq sclist (list 
						 (make-episet-connect :main-set res
							    :how 'set-diff
							    :sets args))))
		   ; add this connection to all the argument sets as well
		   (dolist (a argsets)
			   (change-field 'episet-connections-from a
					    (append (episet-connections-from a)
						    sclist)))
		   ))
	(if (or (traced-p 'epilog::function-eval)
		(traced-p 'epilog::set-function-eval))
              (print-line "    Set Specialist: SET-DIFF "
		               args " evaluated to " (print-info res)))
	  )
  res)

(defun union-of (&rest args &aux
		       (usenet (get-subnet nil))
		       sclist res sname ntype tlist type1 type2 cp argsets)
  (unless (or (null args)
	      (eval (append (list 'or) (mapcar #'(lambda (a) 
				    (not (equal (find-sort a) 'set)))
		      args))))
	  ; build name for this set, create it
	  (setq res (pack (append (list 'epilog::union-of) 
				  (mapcar 'print-info args))))
	  (cond ((setq sname (findset usenet res))) 
		(t (setq sname (getset usenet res))
		   ;; don't need to call writeable-copy because this is
		   ;; guaranteed to create a new struct.

		   (change-property res 'sort 'set)
		   ; find current contents and add them also member type, all
		   ; what if some of these sets are "immed sets"?
		   (setq argsets (mapcar #'(lambda (a) (getset usenet a)) args))
		   (cond ((equal (list-length args) 2)
			  (setq type1 (episet-member-type (car argsets)))
			  (setq type2 (episet-member-type (cadr argsets)))
			  (cond ((and type1 type2 
				      (setq cp (spec-compare-preds type1 type2))
				      (member cp '(equivalent subsumes)))
				 (setq ntype type1))
				((and cp (member cp '(subsumed)))
				 (setq ntype type2))
				((and type1 type2) (setq ntype 
				(normalize-pred (list 'L 'epilog::l-x
				    (list (list 'epilog::l-x (print-info type1)) 
					  'or 
					  (list 'epilog::l-x (print-info type2)))) '+)) )
				))
			 (t (setq tlist (mapcar 'episet-member-type argsets))
			    (unless (member 'nil tlist) (setq ntype
				(normalize-pred (list 'L 'epilog::l-x
				  (append (list (list 'epilog::l-x (print-info (car tlist)))
						'or )
				  (mapcar #'(lambda (b) (list 'epilog::l-x (print-info b)) )
					  (cdr tlist)))) '+)))
			    ))
		   (when ntype 
			 (change-field 'episet-member-type sname ntype)
			 (add-type-assert res ntype nil))

		   (change-field 'episet-contents sname
				    (calc-union (mapcar 'episet-contents argsets)))
		   (change-field 'episet-connections sname
				    (setq sclist (list 
						 (make-episet-connect :main-set res
							    :how 'union-of
							    :sets args))))
		   ; add this connection to all the argument sets as well
		   (dolist (a argsets)
			   (change-field 'episet-connections-from a
					    (append (episet-connections-from a)
						    sclist)))
		   ))
	(if (or (traced-p 'epilog::function-eval)
		(traced-p 'epilog::set-function-eval))
              (print-line "    Set Specialist: UNION-OF "
		               args " evaluated to " (print-info res)))
	  )
  res)

(defun member-type-of (&rest args &aux
			  (usenet (get-subnet nil))
			  (a1 (car args))
			  (s1 (findset usenet a1))
			  res)
  (unless (or (null args) (cdr args) (null s1))
	  (setq res (episet-member-type s1))
	(if (or (traced-p 'epilog::function-eval)
		(traced-p 'epilog::set-function-eval))
	      (print-line "    Set Specialist: MEMBER-TYPE-OF " args 
			  " evaluated to " (print-info res)))
	  )
  res)

