;       Last change    August 1993
(in-package set)


(set-display-function 'epilog::set-info 'set-info 
  "Displays information about sets")

(defun print-s-info (item)
  (if (null item) "unknown" item))

(defun print-bounds (min max)
  (cond ((equal min max) (print-s-info min))
	(t (list (if min min '-infinity) (if max max '+infinity))) )
  )

; print set contents

(defun print-set (brief key node)
  (declare (ignore brief))
  (unless (or (null node)
              (member 'epilog::empty-set (episet-identifiers node)))
  (print-line key 
             #\tab "Set      " #\tab (episet-identifiers node) #\tab
	     (if (episet-finished node) " fully known " " partially known "))
  (if (episet-properties node)
      (print-line #\tab "Set Properties: " #\tab (episet-properties node)))
  (if (episet-member-type node)
      (print-line #\tab "Member Type : " #\tab (print-info (episet-member-type node))
		  #\tab (if (episet-all node) " all "  "")))
   (print-line
             #\tab "Cardinality : " #\tab (print-bounds
					   (episet-cardinality-min node)
					   (episet-cardinality-max node)) *nl*
             #\tab "Contents " #\tab 
	     (print-s-info (remove-if #'+equal (episet-contents node))))
   (when (episet-connections node)
	 (print-line #\tab "Connections: ")
	 (dolist (c (episet-connections node))
		 (print-line #\tab #\tab (episet-connect-how c) " " 
			     (mapcar #'(lambda (a) (print-info a)) 
					 (episet-connect-sets c))))
	 )
   (when (episet-connections-from node)
	 (print-line #\tab "Connections used in: ")
	 (dolist (c (episet-connections-from node))
		 (print-line #\tab #\tab (print-info (episet-connect-main-set c)) 
			     " " (episet-connect-how c) " "
			     (mapcar #'(lambda (a) (print-info a)) 
					 (episet-connect-sets c))))
	 )
  ))

(defun print-sets (brief st)
  (maphash #'(lambda (key node) (print-set brief key node)) st)
)

(defun print-net (context &aux
			  (usec context))
  (if (null context)
      (setq usec '/))
  (let ((context-array (retrieve-specialist-info *environment*
						 'set-equal::set-equal)))
    (if (null context-array)
	nil
      (gethash usec context-array))))

(defun set-info (brief &rest args &aux 
		       (setn (car args))
		       (context (cadr args))
		       (net (print-net context)))
  (when (not (null net))
	(if (or (null setn)
		(equal setn 'epilog::all))
	    (let ()
	         (print-line "Sets for Context" #\tab 
			    (if (null context) '/ context))
                 (print-sets brief (episet-subnet-sets net))
		 )
	    (print-set brief setn (findset net setn))
	    ))
  )

