;; 8 March 2005 - Alexander S. Yeh has found a problem with the
;; combination of evidences (see email to epilog@cs:
;;  Date: Tue, 01 Mar 2005 18:44:25 -0500
;;  From: Alexander S. Yeh <asy@mitre.org>
;;  To: EPILOG <epilog@cs.rochester.edu>
;;  Cc: Samuel L. Bayer <sam@mitre.org>
;;  Subject: Re: Combining probabilities?)
;; the problem has been found to be in the function consistency-check
;; (see the function for see the changes). FM.

;; 8 February 2005 - added the inference catching mechanism
;; (primordial version).
;; -To trace use (trace 'inference-catching).
;; -The catching tests are in the global variable:
;; *inference-catchers*. This is an association list where the key is
;; the name and the value is the cather. The name identifies the
;; catcher and indentifies also the destination where the caught wffs
;; will be stored.
;; -To add a catching test call the macro define-catcher with
;; parameter the name of the catcher and the catching predicate.  An
;; example of use of this function is: (define-catcher 'catcher1
;; #'listp)
;; -The function run-catchers-on-wff runs the catching tests on the
;; particular wff passed as parameter. The keyword parameter
;; :stop-after defines after how many succeded tests we stop trying
;; other catchers. The default is to try all the catchers (value: nil)
;; -The function clean-caught-wffs-for empties from the list of caught
;; wffs in the catcher passed as parameter.
;; -The functions activate/deactivate enable and disable the catcher
;; with name the name passed as parameter to these two functions.

;;       Last change    August 1993
(in-package epilog)

;     Entry routines for story, knowledge, mp's, meta info

(tweakable '*say-knowledge* t 
  "Indicates whether or not to say input knowledge" 'boolean)
(tweakable '*say-mp* nil "Indicates whether or not to say meaning postulates"
  'boolean)
(tweakable '*say-meta* nil "Indicates whether or not to say meta facts" 
  'boolean)
(tweakable '*say-story* t "Indicates whether or not to say input  story wffs"
  'boolean)
(tweakable '*say-infer* t "Indicates whether or not to say inferred wffs"
  'boolean)
(tweakable '*say-immediate* nil "Indicates whether to say things before loaded"
  'boolean) 
(tweakable '*consistency-effort* 3
    "How hard to test consistency of assertions 0=none, 1=lookup only, 2=verify with probs of 1 only, 3=verify, 7=question"
  '(integer 0 7))
(tweakable '*consistency-action* 4 ;1
    "What to do if input is inconsistent 0=enter anyway, 1=warn+enter, 2=dont enter, 3=warn+dont enter, 4=combine"
  '(integer 0 4))

(tweakable '*rule-forward* t 
 "Indicates whether or not forward inference should be attempted for new rules"
  'boolean)
(tweakable '*story-forward* t 
 "Indicates whether or not forward inference should be attempted for new facts"
  'boolean)

(traceable 'entry-eval "Traces verifications done during input")
(traceable 'forward-details "Displays parents formulas for inferred wffs")
(traceable 'inference-catching "Displays information regarding the inference catching mechanism.")

(defvar *save-class* nil)
(defvar *save-intermediate* nil)
(defvar *intermediate-class* nil)

(defvar *meaning-postulate* nil)
(defvar *simplification-schema* nil)
(defvar *meta* nil)
(defvar *goal-only* nil)
(defvar *pieces* nil)
(defvar *ordered-pieces* nil)
(defvar *exists-q* nil)
(defvar *do-not-enter-spec* nil)

(defvar *inference-catchers* nil)

(defstruct catcher
  (name nil) ;; the name (useless only for additional reference)
  (predicate nil) ;; the catching predicate
  (active t) ;; by default the catcher is active
  )

(defun stuff-it (formula &key (prob 1) (skolemized nil) (parents nil)
			 (rank 0) (input-index nil) (inherit nil)
			 assumption assumptions 
			 (proof-prop nil) (rules-so-far nil) 
			 &aux r name tp cl res cres ceff spec-assert 
			 new save-indent save-spaces oldpieces op)
  (setq *simplification-schemas* nil)

  (if (and (> rank 0) (listp formula) (equal (car formula) 'nec))
      (setq formula (cadr formula)))
  (unless (or (splittable formula) (and (= rank 0) (skolemizable formula)))
    (setq name (normalize formula '+ t))
    (if (not name) (return-from stuff-it nil))
    (if (member name parents) (return-from stuff-it nil))
    (if (member name '(YES NO INCONSISTENT)) 
	(return-from stuff-it (if (equal name 'NO) (list 'NO) nil)))
    (setq prob (combine-probs *normed-prob* prob)) )
  (if (and name (> rank 0) (null input-index) (member name parents))
      (return-from stuff-it nil))
  ; Try for simplification mps
  (unless (or (null name) *meaning-postulate*)
    (setq *simplification-schemas* t)
    (classify name)
    (when (setq new (mp-inferences name rules-so-far assumptions proof-prop))
      (when (traced-p 'simplification-schema)
	(print-line "Replacing " formula " by ")
	(print-line "     " (car new)) )
      (return-from stuff-it
	(stuff-it (car new)
		  :prob (combine-probs prob (caddr new))
		  :parents (union parents (cadr new))
		  :rank rank
		  :assumption assumption
		  :assumptions assumptions
		  :inherit inherit
		  :input-index input-index
		  :proof-prop proof-prop
		  :rules-so-far (union rules-so-far (remove-if-not #'(lambda (a) (getp a 'meaning-postulate)) (cadr new))))))
    (setq *simplification-schemas* nil)
    )
  
  (setq res (split-n-skol formula name
			  :prob prob
			  :parents parents
			  :rank rank
			  :assumption assumption
			  :assumptions assumptions
			  :inherit inherit
			  :input-index input-index
			  :proof-prop proof-prop
			  :rules-so-far rules-so-far
			  :skolemized skolemized))
  (if res (return-from stuff-it res))
  
  (if (not name) (return-from stuff-it nil))

  (dolist (p parents)
    (setp p 'descendants (union (getp p 'descendants) (list name)))
    (if (or (getp p 'support) (getp p 'contradict))
	(print-line "  Probability of " name ": " (getp name 'print)
		    " may be inaccurate " (calc-prob-list prob)))
    )
  ; other consistency tests and simplifications
  ; test consistency, handle as directed - simplify using 
  ; wffs with probability 1
  ; simplify, and load simplification in addition to original

  (setq *assertions* nil)
  (when input-index (setq *spec-simplify* t))
  ;; if this is being entered as a result of a question, this step not nec.
  (if (and (> rank 0) (null input-index)) 
      (setq ceff 1) 
    (setq ceff *consistency-effort*))
  (catch 'entry-catcher 
    (setq cres
      (consistency-check name 
			 :proof-prop proof-prop 
			 :assumptions assumptions
			 :parents parents
			 :rank rank
			 :effort ceff
			 :prob prob)
      )
    )
  (setq *spec-simplify* nil)
  (cond ((null cres) (return-from stuff-it cres))
	((equal cres t) nil)
	((member (car cres) '(YES NO INCONSISTENT)) 
	 (return-from stuff-it (list (car cres))))
	(t (setq *save-class* nil)
	   (setq *save-intermediate* nil)
	   (setq name (car cres))
	   (setq prob (caddr cres))
	   (setq parents (cadr cres))
	   ;; Try again for simplification mps
	   (unless (or (null name) *meaning-postulate*)
	     (setq *simplification-schemas* t)
	     (classify name)
	     (when (setq new (mp-inferences name rules-so-far assumptions proof-prop))
	       (when (traced-p 'simplification-schema)
		 (print-line "Replacing " formula " by ")
		 (print-line "     " (car new)) )
	       (return-from stuff-it
		 (stuff-it (car new)
			   :prob (combine-probs prob (caddr new))
			   :parents (union parents (cadr new))
			   :rank rank
			   :assumption assumption
			   :assumptions assumptions
			   :inherit inherit
			   :input-index input-index
			   :proof-prop proof-prop
			   :rules-so-far (union rules-so-far (remove-if-not #'(lambda (a) (getp a 'meaning-postulate)) (cadr new))))))
		   (setq *simplification-schemas* nil) )
	   ) 
	)

  ;; if consistency test returns t, ok to continue else stop
  ;; recheck in case normalization cleaned up some YES's 
  (setq res (split-n-skol (getp name 'print) name
			  :prob prob
			  :parents parents
			  :rank rank
			  :assumption assumption
			  :assumptions assumptions
			  :inherit inherit
			  :input-index input-index
			  :proof-prop proof-prop
			  :rules-so-far rules-so-far
			  :skolemized skolemized))
  (when res (return-from stuff-it res))

  (when (and *forward-full* (> rank 0) (getp name 'conditional))
	(return-from stuff-it nil))

  ;;stuff properties into new name
  (change-property name 'c-prob (calc-prob-list prob))
  (change-property name 'prob prob)
  (change-property name 'rank rank)
  (change-property name 'assumption assumption)
  (when inherit (setp name 'inherit-int inherit))
  (change-property name 'parents parents)
  (change-property name 'input-index input-index)
  (if proof-prop (change-property name proof-prop t))
  (when (and input-index (not *qa-mode*) (= rank 0))
	(add-to-input-array input-index name))
  (set-interest name)
  (set-complexity name)
  (setq cl (classify-and-store name))
  (unless *meaning-postulate* 
	  (setq tp (topical-class name *intermediate-class*)))
  (when *meta* (setp name 'META t))
  (when *meaning-postulate* (setp name 'MEANING-POSTULATE t))
  (when *goal-only* (setp name 'GOAL-ONLY t))

  (when (and *print* (if (or *qa-mode* proof-prop) (traced-p 'qa-success) t))
	(print-wff name cl proof-prop nil tp) )

  (if (and *print* *say-immediate* (not proof-prop)) (say-it (list name)))
  (unless proof-prop (setq *pieces* (append *pieces* (list name))))

  ;specialist entry
  (unless (or proof-prop (not *spec-enter*)) 
	  ; Do NOT enter anything into specialists which does not have
	  ; probability 1!!
	  (when (and (equal (calc-prob-list prob) 1) 
		     (not *do-not-enter-spec*))
		(spec-enter name))
	  (setq spec-assert *assertions*)
	  (when (and *spec-assert* *assertions*)
		(setq save-indent *indent*)
		(setq save-spaces *spaces*)
		(set-indent (+ rank 1))
		(dolist (s spec-assert)
			(when (traced-p 'spec-assert)
			      (print-line "Asserting " s " from specialists"))
			(setq oldpieces *pieces*)
			(setq op *ordered-pieces*)
			(setq *pieces* nil)
			(setq *ordered-pieces* nil)
			(setq r (stuff-it s
					  :rank (+ rank 1)
					  :prob prob
					  :parents (list name) ;parents
					  :input-index input-index
					  :assumption assumption
					  :assumptions assumptions
					  :proof-prop proof-prop
					  :rules-so-far rules-so-far))
			(unless proof-prop (order-pieces))
			(setq *pieces* oldpieces)
			(setq *ordered-pieces* (append op *ordered-pieces*))
			(setq res (append res r)) )
		(reset-indent save-spaces save-indent)
		)
	  )

  (setq res (append (list name) res
		    ;; Change by ANK: if assumption is t, add name
		    ;; to the list of assumptions.
		    ; (use-formula name proof-prop rules-so-far assumptions)))
		    (use-formula name proof-prop
				 rules-so-far
				 (if assumption
				     (cons name assumptions)
				   assumptions))))
  res)

(defun pr-order ()
  (dolist (p *ordered-pieces*)
	  (dolist (w p)
		  (print-line w " " (getp w 'rank) " " (print-info w)))
	  (print-line)))

(defun order-pieces ()
  (when (and *pieces* (not *qa-mode*))
	(setq *ordered-pieces* (append (list *pieces*) *ordered-pieces*))
	(setq *pieces* nil)
	)
  )

(defun start-pieces () 
  (setq *pieces* nil)
  (setq *ordered-pieces* nil))

(defun use-formula (name proof-prop rules-so-far assumptions &optional qa &aux res)
  ; try input driven inferencing with the formula
  ; don't do on mp's or meta information
  (unless (or *meaning-postulate* *meta* *goal-only* 
	      (getp name 'meaning-postulate)
	      (getp name 'meta)
	      (and proof-prop (if qa nil (not *goal-forward*))))
	  (setq res (make-inferences name rules-so-far assumptions qa proof-prop))
	  )
  (remprop name 'lit-classes)
  res)

(defun make-inferences (name rules-so-far assumptions &optional qa proof-prop &aux res
										   (save-indent *indent*)
										   (save-spaces *spaces*)
										   (rank (getp name 'rank))
										   )
  ;; meaning postulates are applied regardless of interest or other 
  (set-indent (+ rank 1))
  (unless (or *meaning-postulate* *meta* ) 
    (setq res (mp-inferences name rules-so-far assumptions proof-prop))
    )

  ;; knowledge rules are only applied if interest is high enough, and
  ;; for input story sentences regardless of interest
  (when (and (or qa (if (getp name 'conditional) *rule-forward* *story-forward*))
	     (or (= (getp name 'rank) 0) (check-interest name)))
    (setq res (append res (input-driven-inferences name rules-so-far 
						   assumptions proof-prop)))
    )
  ;; also meaning postulate inferences
  (reset-indent save-spaces save-indent)
  (dolist (inf-formula res)
    (if (traced-p 'inference-catching)
	(print-line "Inference catching mechanism: applying catchers to " inf-formula)
      )
    (run-catchers-on-wff inf-formula)
    )
  res)

(defun split-n-skol (formula name &key prob parents rank input-index proof-prop
			     assumption assumptions
			     rules-so-far skolemized inherit 
			     &aux change res)
  ; if skolemizable, skolemize
  ; if result of that, or original, is splittable, split as well
  (when (skolemizable formula)
	(when (or (> rank 0) (member (car formula) '(E! THE)))
	      (unless name (setq name (normalize formula '+)))
	      (if (exists-already name) 
		  (return-from split-n-skol '(YES))))
        (setq formula (skolemize formula (= rank 0)))
	(when (null formula) (return-from split-n-skol '(YES)))
	(setq change t)
	(setq skolemized t))
  ; then split
  ; splits - probability is based on prob before split
  ; also, if split is because of skolemization, first item gets prob 1
  (if (splittable formula)
      (let* ((fs (split formula))
	     (first (> (length fs) 1))
	     (newp (if first (new-symbol 'PROB))))
	    (when newp
		  (setp newp 'assumption assumption)
		  (setp newp 'prob prob)
		  (setp newp 'c-prob (calc-prob-list prob)))

         ;; Note: A conjunction may have prob. < 1, and this
         ;; raises the question what the probabilities of the
         ;; conjuncts are. To handle this we store (atomic) NAMES
         ;; of such probabilities, rather than the probabilities
         ;; themselves, under the PROB indicators. (The numerical 
         ;; values are stored under the names.) Then, when computing 
	 ;; the probability of a rule antecedent, we multiply together only
         ;; probabilities WITH DISTINCT NAMES.

	 (mapcar #'(lambda (w)
	     (let* ((p (if (and first skolemized) 1 (list newp))) 
		    pres)
		  (setq first nil)
		  (setq pres (stuff-it w 
			    :parents parents
			    :rank rank
			    :assumption assumption
			    :assumptions assumptions
			    :inherit inherit
			    :input-index input-index
			    :proof-prop proof-prop
			    :rules-so-far rules-so-far
			    :prob p))
		   (if assumption
		       (setq assumptions (append assumptions pres)))
		   (setq res (append res pres))
		   )) fs)
	    )
      (let ()
	(if change
	  (setq res (stuff-it formula
			    :parents parents
			    :rank rank
			    :assumption assumption
			    :assumptions assumptions
			    :inherit inherit
			    :input-index input-index
			    :proof-prop proof-prop
			    :rules-so-far rules-so-far
			    :prob prob
			    :skolemized skolemized))) )
      )
  res)

(defun print-wff (name cl proof-prop reassert tp &aux i 
		       (rank (getp name 'rank)))
  (unless proof-prop (print-line))
  (cond (reassert 
	 (if proof-prop
	     (if (traced-p 'qa-success)
		 (print-line "Reasserting formula for " proof-prop " attempt ..."))
	     (print-line "Reasserting ... ")) )
	((> rank 0) 
	 (when (traced-p 'forward) 
	       (if proof-prop 
		   (print-line "** INFERRED " proof-prop " FORMULA **")
		   (print-line "** INFERRED FORMULA **"))) )
	(proof-prop
	 (print-line "Loading " proof-prop " assumption ..."))
	((getp name 'input-index) (print-line "Loading story sentence ..."))
	(*simplification-schema* (print-line "Loading simplification schema ...")) 
	(*meaning-postulate* (print-line "Loading meaning postulate ...")) 
	(*meta* (print-line "Loading meta information ...")) 
	(t (print-line "Loading knowledge ...")) )

  (print-line "     " name ": " (getp name 'print))
  (unless proof-prop
	  (when (and (not reassert) (= rank 0) 
		     (setq i (getp name 'input-index)))
		(print-line "     Stored in Input Array [" i "]") )
	  (when (and (not reassert) (getp name 'parents)) 
		(print-line "     Rank:  " rank)
		(if (traced-p 'forward-details)
		    (let ((par (getp name 'parents)))
			 (cond (par
			 (print-line "     Parents:  " (car par) ": "
				     (print-info (car par)))
			   (dolist (p (cdr par))
				   (print-line "               " p ": "
					       (print-info p)))) 
			       (t (print-line "     Parents: specialist"))))
		    (if (getp name 'parents)
			(print-line "     Parents:  " (getp name 'parents) )
			(print-line "     Parents: specialist")
			)) )
	  (unless reassert
		(print-line "     Probability:  " (getp name 'c-prob) )
		(when (getp name 'controlled-vars)
		      (print-line "     Controlled variables: "
				  (getp name 'controlled-vars)))
		(when (traced-p 'entry-class)
		      (print-line "     Interest:  " (getp name 'interest)  )
		      (when (getp name 'conditional)
			    (print-line (if (getp name 'equivalence)
					    "     -/+Trigger(s):  " 
					    "     -Trigger(s):  " )
				  (mapcar 'print-info (getp name '-trigger)  ))
			    (print-line (if (getp name 'equivalence)
					    "     +/-Trigger(s):  " 
					    "     +Trigger(s):  " )
				  (mapcar 'print-info (getp name '+trigger)  ))
			    )
		      (print-line "     Classified under:  " )
		      (dolist (c cl) (print-line "           " c))
		      (unless *meaning-postulate* 
		      (print-line "     Topically classified under:  " )
		      (dolist (c tp) (print-line "           " c)))
		      )) 
	  )
  )


 ;**********************************
 ;*                                *
 ;*   M A I N   R O U T I N E S    *
 ;*                                *
 ;**********************************

 ;;
 ;; Load the formula whose name is PHI-NAME into the knowledge
 ;; base. PHI-NAME is assumed already to have certain properties,
 ;; such as WFF (the formula itself), PROB (the lower bound on its
 ;; subjective probability -- default 1), RANK (the number of inf-
 ;; erence steps that led to this formula, starting with "givens"),
 ;; PARENTS (for rank > 0 formulas), and INPUT-INDEX -- "where we
 ;; are" in the surface narrative, in terms of successively stored
 ;; minimal wffs. (This is NIL for non-narrative, "background"
 ;; knowledge.)
 ;;
 ;; First do some normalizations: distribute negations, and for
 ;; non-conditional formulas (at the top level, or following a
 ;; necessity operator) distribute modal operators (other than POSS)
 ;; and universal quantifiers over conjunctions. Also, skolemize
 ;; top-level existentials and split up top-level conjunctions,
 ;; LOADing the conjuncts separately. Also, standardize any remain-
 ;; ing variables.
 ;;
 ;; If the formula to be loaded is a conditional, identify the
 ;; "key literals" (or "key wffs") of its antecedent, and the "key
 ;; predicate" for each such key literal. Associate the list of
 ;; (key-pred key-lit) pairs with PHI-NAME (under indicator KEYS),
 ;; and make PHI-NAME accessible from each key predicate (under
 ;; indicator RULES). (Note that there is no separation of key
 ;; literals on the basis of sign; the assumption is that the wffs
 ;; making up the antecedent of a conditional will usually be pos-
 ;; itive and that non-conditional story facts will usually be pos-
 ;; itive as well; cases of sign mismatch will be weeded out by
 ;; unification failure.)
 ;;
 ;; If the formula to be loaded is non-conditional (typically, a
 ;; story fact), store it in the hash table (using the wff itself
 ;; as a hash key); also store it under "double indices"
 ;; consisting of its key predicate, together with an atomic
 ;; term occurring within it. At this point INFERENCE MERGING
 ;; of the new formula with previously loaded ones should
 ;; be attempted, though this is not yet implemented. It seems
 ;; that one needs to categorize Skolem constants into rank-0
 ;; constants (from explicit existentials in the input) and rank>0
 ;; ones (inferred entities). Rank-0 constants cannot be merged
 ;; with other rank-0 constants (i.e., explicit indefinites in the
 ;; input generally introduce distinct entities), while all other
 ;; merges are possible. The requirement is that they be of compat-
 ;; ible types (e.g., wolf and animal) -- and also that other 
 ;; available information about them not be blatantly contradictory.
 ;; In the case of episodic constants, their overall descriptions
 ;; should be compatible, i.e., one should entail the other.
 ;; 
 ;; NOW, (1) IF PHI IS A RANK-0 EPISODIC FORMULA, TRY NARRATIVE
 ;; INFERENCING BASED ON THE FORMULA AND ITS PREDECESSOR, I.E., THE
 ;; EPISODIC FORMULA WITH THE GREATEST INPUT-INDEX WHICH IS SMALLER
 ;; THAN THAT OF THE CURRENT FORMULA; (2) ACCESS POSSIBLY RELEVANT
 ;; RULES VIA THE KEY PREDICATE, AND TRY FORWARD-CHAINING WITH THEM.
 ;; (These are the most important steps, since it is here that the
 ;; "story-inferencing" actually takes place.  Note, however, that
 ;; step (1) has not been implemented in this program.)  For the
 ;; resulting conclusions, if any, decide which ones to LOAD, and
 ;; then LOAD them recursively. (Formulas whose probability is low
 ;; while their rank is high are not loaded.)
 ;;

(defun not-same-path (wffs-a wffs-b &optional allowed &aux 
			     p lefta leftb int smint)
  (cond ((or (null wffs-a) (null wffs-b)) t)
	((null (set-exclusive-or wffs-a wffs-b)) nil)
	(t (setq lefta (set-difference wffs-a wffs-b))
	   (setq leftb (set-difference wffs-b wffs-a))
	   (setq int (union (setq smint (intersection wffs-a wffs-b)) allowed))
	   (cond ((and (= (list-length lefta) 1) 
		       (setq p (set-difference (getp (car lefta) 'parents) int)))
		  (not-same-path p leftb int))
		 ((and (= (list-length leftb) 1)
		       (setq p (set-difference (getp (car leftb) 'parents) int) ))
		  (not-same-path lefta p int))
		 ((null smint) t)
		 (t (not-same-path lefta leftb int)))
	   ))
  )

(defun original-parent (par wffs)
  (cond ((null wffs) nil)
	((member par wffs) t)
	(t (dolist (w wffs)
		   (if (original-parent par (getp w 'parents))
		       (return-from original-parent t))
		   )
	   nil))
  )

(defun Old-not-same-path (wffs-a wffs-b &aux lefta leftb int)
  (cond ((or (null wffs-a) (null wffs-b)) t)
	((null (set-exclusive-or wffs-a wffs-b)) t)
	(t (setq lefta (set-difference wffs-a wffs-b))
	   (setq leftb (set-difference wffs-b wffs-a))
	   (setq int (intersection wffs-a wffs-b))
	   (cond ((and (= (list-length lefta) 1) 
		       (null (set-exclusive-or 
			      (set-difference (getp (car lefta) 'parents) int) 
			      leftb))) nil)
		 ((and (= (list-length leftb) 1)
		       (null (set-exclusive-or lefta 
			      (set-difference (getp (car leftb) 'parents) int)))
		       ) nil)
		 (t t)) )
	))

		    
(defun consistency-check (name &key proof-prop parents rank prob assumptions 
			       (effort *consistency-effort*)
			       &aux vans newpar newprob 
			       (phi (getp name 'print))
			       contwff o-c-prob c-prob savever ver o-prob wffs)
  (if (traced-p 'entry-eval)
      (print-line "Entry: Verifying new formula " (getp name 'print)))
  (setq *do-not-enter-spec* nil)
  (setq *insertion* t)
  (setq *fn-simplified* nil)
  (setq ver (consistency-verify name proof-prop assumptions effort))
  (when (and ver *fn-simplified* 
	     (not (member (car ver) '(YES NO inconsistent))))
    (setq savever ver)
    (setq ver (consistency-verify (car ver) proof-prop assumptions effort))
    (if ver 
	(let () 
	  (setq prob (combine-probs prob (caddr savever)))
	  (setq parents (union parents (cadr savever))))
      (setq ver savever))
    )
  (setq *insertion* nil)
  (unless ver (return-from consistency-check t))
  (when (and ver (traced-p 'entry-eval))
    (print-it "        --> " (print-info (car ver))))
  (setq vans (car ver))
  (setq o-prob (caddr ver))
  (setq wffs (cadr ver))
  (setq newpar (union parents wffs))
  (setq newprob (combine-probs prob o-prob))
  (setq o-c-prob (calc-prob-list o-prob))
  (setq c-prob (calc-prob-list prob))
  
;  (print-line name " " phi  " " prob " " parents " " ver)

  (cond ((equal vans 'UNKNOWN) (return-from consistency-check t))
	((equal vans 'NO)
	 (cond (proof-prop
		(print-contrad name phi ver proof-prop parents)
		(return-from consistency-check (list vans newpar newprob)) )
	       (t  (if (member *consistency-action* '(1 3))
		       (print-contrad name phi ver proof-prop parents))
		   (cond ((equal *consistency-action* 4)
			  ; combine results
			  ; add prob to contradicting for name
			  ; recalc prob from scratch
			  (when (traced-p 'FORWARD)
				(print-combine name phi parents proof-prop 
					       c-prob o-c-prob ver t))
			  (if (null (getp name 'support))
			      (setp name 'support (list prob)))
			  (setp name 'contradict
				(append (getp name 'contradict) (list o-prob)))
			  (setq newprob (recalc-prob name phi))
			  (setq newpar parents)
			  (when (and (= (list-length wffs) 1)
				     (setq contwff (car wffs))
				     (not (getp contwff 'SPEC-EVAL)))
				(if (null (getp contwff 'support))
				    (setp contwff 'support (list (getp contwff 'prob))))
				(setp contwff 'contradict (append (getp contwff 'contradict) (list prob)))
			        (recalc-probs (car wffs)) )
			  (list name newpar newprob)
			  )
			 ((> *consistency-action* 1)
			  ; do not enter
			  (return-from consistency-check nil))
			 (t (if (or (null (cadr ver)) 
				    (getp (caadr ver) 'spec-eval))
				(setq *do-not-enter-spec* t))
			    (return-from consistency-check t)) )
		   )) )
	((and (equal vans 'YES) 
	      (= (list-length wffs) 1))
	 (cond ((getp (car wffs) 'SPEC-EVAL)
		(return-from consistency-check t))
	       ((not (not-same-path wffs parents))
		(cond ((> c-prob o-c-prob)
		       (when (traced-p 'forward)
			 (print-line " Replacing probability of " name ": " 
				     phi
				     " with new higher probability " c-prob) )
		       (setp (car wffs) 'prob prob)
		       (recalc-prob (car wffs)) )
		      (t (if (if proof-prop (traced-p 'qa-success) nil) 
			     (print-duplicate name phi proof-prop ver parents)))
		      )
		nil)
	       
	       ((original-parent (car wffs) parents)
		nil)

	       ;; changed on March 8th 2005 to correct a bug in the
	       ;; evidence combination discovered by Alexander S. Yeh.
	       ;; Instead of returning nil even in the case of rank
	       ;; greater than 0 (because the condition (= 0 rank) was
	       ;; inserted as a nested if, instead that directly in
	       ;; the condition of the cond), we consider the new
	       ;; evidence a duplicate only if the rank is 0. FM.
; 	       ((equal (car wffs) name)
; 		(if (= 0 rank)
; 		    (print-duplicate name phi proof-prop ver parents)))
 	       ((and (equal (car wffs) name) (= 0 rank))
 		(print-duplicate name phi proof-prop ver parents))

	       ((and (equal *consistency-action* 4)
		     (equal (getp (car wffs) 'wff-type) (getp name 'wff-type)))
		(when (traced-p 'FORWARD)
		  (print-combine name phi parents proof-prop c-prob o-c-prob 
				 ver nil))
					;combine results
		(if (null (getp name 'support))
		    (setp name 'support (list (getp name 'prob))))
		(setp name 'support
		      (append (getp name 'support) (list prob)))
		(if (and (>= c-prob o-c-prob) (not (null parents)))
		    (setq newpar parents)
		  (setq newpar (getp name 'parents)))
		(if prob (setp name 'PROB prob))
		(setq newprob (recalc-probs name))
		(list name newpar newprob)
		)
	       ((and (if proof-prop (getp (car wffs) proof-prop) t)
		     (<= c-prob o-c-prob))
		(if (if proof-prop (traced-p 'qa-success) (traced-p 'FORWARD))
		    (print-duplicate name phi proof-prop ver parents))
		(return-from consistency-check (if proof-prop nil '(YES))))
	       (t			;must have c-prob > o-c-prob to get here or 
		(return-from consistency-check t) )
	       )
	 )
	((and (equal vans 'YES) 
	      (<= c-prob o-c-prob)
	      (or proof-prop (> rank 0)))
	 (if proof-prop
	     (if (traced-p 'qa-success)
		 (return-from consistency-check (print-line 
                       " %% Inferred " proof-prop " formula " NAME ": " PHI 
		       " already true - ignoring" *nl* *spaces* 
		       "    NOT LOADed!" ))
		 (return-from consistency-check nil) )
	     (if (traced-p 'FORWARD)
		 (return-from consistency-check (print-line
                      " %% Inferred formula " NAME ": " PHI 
		      " already true " *nl* *spaces* "    NOT LOADed!" ))
		 (return-from consistency-check nil) )) 
	   )
	((and (equal vans 'YES) (> c-prob o-c-prob))
	 (return-from consistency-check t))
	(t ;simplified
	   (list vans newpar newprob))
	)
  )

(defun print-combine (name phi parents proof-prop c-prob o-c-prob res contradict)
  (print-line " Combining " (if contradict "contradictory" "supporting")
	      " evidence for " NAME ": " PHI )
  (if (and parents (not proof-prop)) 
      (print-inf name parents c-prob)
      (print-line "        with probability " c-prob))
  (print-line "      Existing evidence: ")
  (if (and (caadr res) (not (getp (caadr res) 'spec-eval)))
      (print-it (caadr res) ": " (getp (caadr res) 'print))
      (print-it "from specialist evaluation"))
  (unless (or proof-prop (equal (cadr res) (caddr res)))
	  (print-inf name (getp (caadr res) 'parents) o-c-prob))
  )

(defun consistency-verify (phi-name proof-prop assumptions con &aux r)
  (setq *save-class* nil)
  (setq *save-intermediate* nil)
  (cond (proof-prop (setq con 3))
	((and (setq r (getp phi-name 'rank)) (> r 0))
	 (if (> *consistency-effort* 3) (setq con 3)))
	(*meaning-postulate* (setq con 1))
	(*meta* (setq con 2))
	)
  (case con
	(0 nil)
	(1 (lookup phi-name proof-prop t assumptions))
	(2 (verify phi-name proof-prop 1 t assumptions))
	(3 (verify phi-name proof-prop 2 t assumptions))
	(7 (let ((res (question (print-info phi-name) nil nil "Questioning consistency of" nil)))
		(qa-cleanup nil)
		res))
	(otherwise (verify phi-name proof-prop 1 t assumptions))
	)
  )
       
(defun print-duplicate (phi-name phi proof-prop res parents)
  (print-line "%% Duplicate " (if proof-prop "subgoal " "formula ")
		 PHI-NAME ": " PHI *nl* *spaces* "    NOT LOADed!" )
  (when (and parents (not proof-prop)) (print-inf phi-name parents))
  (print-line "      Existing formula: ")
  (if (and (caadr res) (not (getp (caadr res) 'spec-eval)))
      (print-it (caadr res) ": " (getp (caadr res) 'print))
      (print-it "from specialist evaluation"))
  (unless (or proof-prop (equal (cadr res) (caddr res)))
	  (print-inf (caadr res)))
  nil)

(defun print-inf (item &optional (parents (getp item 'parents)) prob &aux i)
  (when parents
  (print-line "       Inferred ")
  (if prob (print-it "with probability " prob " "))
  (print-it "from: ")
  (dolist (i parents)
	  (unless (numberp i)
		  (print-line "           " i ": " (getp i 'print))
		  )))
  (unless parents
  (when (and (getp item 'rank) (= (getp item 'rank) 0) 
	   (setq i (getp item 'input-index)))
      (print-line "       Stored in Input Array [" i "]")
      (if prob (print-it "   with probability " prob))) )
  )

(defun print-contrad (name phi res proof-prop parents)
  (print-line "%% Contradiction found for " name ": " phi )
  (when (and parents (not proof-prop)) (print-inf name parents))
		  
  (cond (proof-prop (print-it " in " proof-prop " attempt "))
	((equal *consistency-action* 1)
	 (print-line "    LOADed anyway!" ))
	((equal *consistency-action* 3)
	 (print-line "    Not LOADed!" ))
	)
  (if (or (null (cadr res)) (getp (caadr res) 'spec-eval)) 
      (print-line "      Specialist evaluation")
      (print-line "      Existing formula: " (caadr res) ": "
		 (getp (caadr res) 'print)))
  (unless (or proof-prop (null (cadr res)) (getp (caadr res) 'spec-eval) 
	      (equal (cadr res) (caddr res)))
	  (print-inf (caadr res)))
  nil)


 ;*****************************************************
 ;*                                                   *
 ;*   U S E R   I N T E R F A C E   R O U T I N E S   *
 ;*                                                   *
 ;*****************************************************

 ;;;
 ;;; Functions to load world knowledge.
 ;
 ; don't need to print time here - this should only be used for rules
(defun knowledge (&rest formula &aux result)
    (qa-cleanup)
    (dolist (f formula)
	    (anon-checkpoint)
	    (setq result (append result (load-knowl1 f))))
    result)

(defun kn (&rest formula)          ; Abbrev. name for KNOWLEDGE
	(apply 'knowledge formula))

(defun goal-knowledge (&rest formula &aux result)
    (qa-cleanup)
    (setq *goal-only* t)
    (dolist (f formula)
	    (anon-checkpoint)
	    (setq result (append result (load-knowl1 f))))
    (setq *goal-only* nil)
    result)

(defun goal-kn (&rest formula)          ; Abbrev. name for GOAL-KNOWLEDGE
	(apply 'goal-knowledge formula))

(defun gkn (&rest formula)          ; Abbrev. name for GOAL-KNOWLEDGE
	(apply 'goal-knowledge formula))

(defun load-knowl1 (formula &aux res save-filter)
  (reset-indent "" 0)
  (start-pieces) 
  (setq res (remove-if #'(lambda (a) (member a '(YES NO UNKNOWN INCONSISTENT)))
		       (stuff-it formula)))
  (order-pieces)
  (when (and *print* (if *meaning-postulate* *say-mp*
			 (if *meta* *say-meta*
			 (or *say-knowledge* *say-infer*))))
	(setq save-filter *filter-threshold*)
	(tweak '*filter-threshold* 100)
      (if *say-knowledge*
	  (if *say-infer*
	      (say-them *ordered-pieces*)
	      (say-it (car *ordered-pieces*)))
	  (if *say-infer*
	      (say-them (cdr *ordered-pieces*))) )
	(tweak '*filter-threshold* save-filter)
	)
  res)

(defun meaning-postulate (&rest formula &aux result)
    (setq *meaning-postulate* t)
    (qa-cleanup)
    (dolist (f formula)
	    (anon-checkpoint)
	    (setq result (append result (load-knowl1 f))))
    (setq *meaning-postulate* nil)
    (dolist (r result)
	    (setp r 'MEANING-POSTULATE t))
    result)

(defun mp (&rest formula)  ;Abbrev for meaning-postulate
  (apply 'meaning-postulate formula))

(defun simplification-schema (&rest formula &aux result)
    (setq *meaning-postulate* t)
    (setq *simplification-schema* t)
    (qa-cleanup)
    (dolist (f formula)
	    (anon-checkpoint)
	    (setq result (append result (load-knowl1 f))))
    (setq *simplification-schema* nil)
    (setq *meaning-postulate* nil)
    (dolist (r result)
	    (setp r 'SIMPLIFICATION-SCHEMA t)
	    (setp r 'MEANING-POSTULATE t))
    result)

(defun simp-schema (&rest formula) 
  ;Abbr for simplification-simp
  (apply 'simplification-schema formula))

(defun ss (&rest formula) 
  ;Abbr for simplification-simp
  (apply 'simplification-schema formula))

(defun meta (&rest formula &aux result)
    (setq *meta* t)
    (qa-cleanup)
    (dolist (f formula)
	    (anon-checkpoint)
	    (setq result (append result (load-knowl1 f))))
    (setq *meta* nil)
    (dolist (r result) (setp r 'META t))
    result)

 ;;;
 ;;; Functions to load story sentences.
 ;
(defun story (&rest formula &aux result)
    (qa-cleanup)
    (dolist (f formula)
      (anon-checkpoint)
      (setq result (append result (load-story-sentence1 f))))
    (if (traced-p 'entry-time) 
	(print-line "    Used "
		    (/ (float (- (get-internal-run-time) *qa-start-time*))
		       internal-time-units-per-second) " seconds CPU time"
		       #\tab
		       (/ (float (- (get-internal-real-time) *qa-start-real-time*))
			  internal-time-units-per-second) " seconds real time" *nl*)
      )
    result)

(defun load-story-sentence1 (formula &aux res save-filter)
  (reset-indent "" 0)
  (start-pieces)
  (setq res (remove-if #'(lambda (a)
			   (member a '(YES NO UNKNOWN INCONSISTENT)))
		       (stuff-it formula :input-index (new-index))))
  (order-pieces)
  (when (and *print* (or *say-story* *say-infer*))
    (setq save-filter *filter-threshold*)
    (tweak '*filter-threshold* 100)
    (if *say-story*
	(if *say-infer*
	    (say-them *ordered-pieces*)
	  (say-it (car *ordered-pieces*)))
      (if *say-infer*
	  (say-them (cdr *ordered-pieces*))) )
    (tweak '*filter-threshold* save-filter)
    )
  res)

; reassert re-loads an existing formula by re-doing the forward inferencing
; storage and classification are not changed

(defun reassert (phi-name &key proof-prop assumptions qa &aux res save-filter)
  (unless qa (qa-cleanup))
  (when (and (symbolp phi-name) (getp phi-name 'print))
	  (start-pieces)
	  (when (and *print* (if (or *qa-mode* proof-prop) 
				 (traced-p 'qa-success) t))
		(print-wff phi-name nil proof-prop t nil) )
	  (setq res (append (list phi-name) 
		    (use-formula phi-name proof-prop nil assumptions qa)))
	  (order-pieces)
	  (when (and *print* (or *say-story* *say-infer*))
		(setq save-filter *filter-threshold*)
		(tweak '*filter-threshold* 100)
		(if *say-story*
		    (if *say-infer*
			(say-them *ordered-pieces*)
			(say-it (car *ordered-pieces*)))
		    (if *say-infer*
			(say-them (cdr *ordered-pieces*))) )
		(tweak '*filter-threshold* save-filter)
		)
	  res))

; should this take into account probabilities?

(defun exists-already (item &aux
                            ans  
			    (qa-iter *qa-iterations*)
			    )
  (when (and (not *qa-mode*) (member (getp item 'quantifier) '(E E! THE)))
	(tweak '*qa-iterations* 10)
	(setq *exists-q* t)
	; also get assumptions involved? should rules-so-far be involved?
	(setq ans (proof-q (print-info item) nil 1 "Verifying existence of" nil))
	(setq *exists-q* nil)
	(qa-cleanup nil)
	(tweak '*qa-iterations* qa-iter)
        (when (and ans (equal (caar ans) 'YES))
	      (when (traced-p 'forward-eval)
                (print-line "Existentially quantified inference already true " )
	        (print-line "   " item ": " (getp item 'print))
                (print-line "      Existing formulas: " )
                (dolist (i (cadar ans))
                        (print-line "        " i ": " (getp i 'print)))
		    )
              (return-from exists-already t)
              )  )
  )

(defun define-catcher (name predicate &key (overwrite t))
  (if (assoc name *inference-catchers*)
      (if overwrite
	  (progn
	    (if (traced-p 'inference-catching)
		(print-line
		 "WARNING: define-catcher: a catcher with this name is already present. RE-DEFINING it.")
	      )
	    (setf (symbol-value name) nil)
	    (rplacd (assoc name *inference-catchers*)
		    (make-catcher :name name :predicate predicate))
	    )
	(if (traced-p 'inference-catching)
	    (print-line
	     "WARNING: define-catcher: a catcher with this name is already present. SKIPPING.")
	  )
	)
    (progn
      (setf (symbol-value name) nil)
      (push (cons name (make-catcher :name name :predicate predicate)) *inference-catchers*)
      )
    )
  )

(defun activate (name)
  (setf (catcher-active (cdr (assoc name *inference-catchers*))) t)
  )

(defun deactivate (name)
  (setf (catcher-active (cdr (assoc name *inference-catchers*))) nil)
  )

(defun clean-caught-wffs-for (name)
  (setf (symbol-value name) nil)
  )

(defun run-catchers-on-wff (wff-name &key (stop-after nil) &aux (wff (getp wff-name 'print)))
  (assert (or (null stop-after) (and (numberp stop-after) (> stop-after 0))))
  (dolist (catcher (remove-if-not #'catcher-active
				  (mapcar #'cdr *inference-catchers*))
	    )
    (if (traced-p 'inference-catching)
	(print-line "Inference catching mechanism: applying catcher " catcher " to " wff-name)
      )
    (and (funcall (catcher-predicate catcher) wff)
	 (push wff (symbol-value (catcher-name catcher)))
	 stop-after
	       (if (<= (setq stop-after (1- stop-after)) 0)
		   (return)
		 )
	       )
	  )
  )
